// jack.c
// LiVES (lives-exe)
// (c) G. Finch 2005 - 2006
// Released under the GPL 2.0
// see file ../COPYING for licensing details

#include "main.h"
#include "callbacks.h"
#include "support.h"

static jack_client_t *jack_transport_client;

int lives_start_ready_callback (jack_transport_state_t state, jack_position_t *pos, void *arg) {
  // we will return FALSE until the first frame is loaded, and the audio seek is ready

  //return mainw->video_seek_ready|mainw->audio_seek_ready;
  return TRUE; // for now
}

void lives_jack_init (void) {
  gchar *jt_client=g_strdup_printf("LiVES-%d",getpid());
  if ((prefs->jack_opts&JACK_OPTS_START_TSERVER)||(prefs->jack_opts&JACK_OPTS_START_ASERVER)) {
    unsetenv ("JACK_NO_START_SERVER");
    setenv ("JACK_START_SERVER","1",0);
  }
  else {
    unsetenv ("JACK_START_SERVER");
    setenv ("JACK_NO_START_SERVER","1",0);
  }

  jack_transport_client=jack_client_new(jt_client);
  g_free(jt_client);

#ifdef ENABLE_JACK_TRANSPORT
  if (jack_transport_client!=NULL) jack_set_sync_callback (jack_transport_client, lives_start_ready_callback, NULL);
  gtk_timeout_add(KEY_RPT_INTERVAL,&lives_jack_poll,NULL);
  g_snprintf(prefs->jack_tserver,256,"~/.jackdrc");
#endif

}

/////////////////////////////////////////////////////////////////
// transport handling

void jack_transport_check_state (void) {
  jack_position_t pos;
  jack_transport_state_t jacktstate;

  if (!(prefs->jack_opts&JACK_OPTS_TRANSPORT_CLIENT)) return;

  if (jack_transport_client==NULL) return;

  jacktstate=jack_transport_query (jack_transport_client, &pos);

  if (mainw->jack_can_start&&(jacktstate==JackTransportRolling||jacktstate==JackTransportStarting)&&mainw->playing_file==-1&&mainw->current_file>0) {
    mainw->jack_can_stop=TRUE;
    on_playall_activate(NULL,NULL);
  }
  if (jacktstate==JackTransportStopped) {
     if (mainw->playing_file>-1&&mainw->jack_can_stop) on_stop_activate (NULL,NULL);
     mainw->jack_can_start=TRUE;
  }
}

gboolean lives_jack_poll(gpointer data) {
  // data is always NULL
  // must return TRUE
  if (!mainw->is_processing) jack_transport_check_state();
  return TRUE;
}

void lives_jack_end (void) {
  jack_client_close (jack_transport_client);
  jack_transport_client=NULL;
}


void jack_pb_start (void) {
  // call this ASAP, then in load_frame_image; we will wait for sync from other clients (and ourself !)
  if (prefs->jack_opts&JACK_OPTS_TRANSPORT_MASTER) jack_transport_start (jack_transport_client);
}

void jack_pb_stop (void) {
  // call this after pb stops
  if (prefs->jack_opts&JACK_OPTS_TRANSPORT_MASTER) jack_transport_stop (jack_transport_client);
}

////////////////////////////////////////////
// audio

static jack_driver_t outdev[JACK_MAX_OUTDEVICES];

static void sample_move_d8_d16(short *dst, guchar *src,
				unsigned long nsamples, gdouble scale, int nDstChannels, int nSrcChannels) {
  int nSrcCount, nDstCount;
  gdouble src_offset_d=0.;
  gint src_offset_i=0;
  gint ccount;
  
  //TRACE("nsamples == %ld, nDstChannels == %d, nSrcChannels == %d\n", nsamples, nDstChannels, nSrcChannels);

  if(!nSrcChannels) return;

  if (scale<0.) {
    src_offset_d=((gdouble)(nsamples)*(-scale)-1.);
    src_offset_i=(gint)src_offset_d*nSrcChannels;
  }

  while(nsamples--) {
    nSrcCount = nSrcChannels;
    nDstCount = nDstChannels;
    ccount=0;

    /* loop until all of our destination channels are filled */
    while(nDstCount) {
      nSrcCount--;
      nDstCount--;

      *(dst++) = *(src+ccount+src_offset_i)<<8; /* copy the data over */
      ccount++;

      /* if we ran out of source channels but not destination channels */
      /* then start the src channels back where we were */
      if(!nSrcCount && nDstCount) {
	ccount=0;
	nSrcCount = nSrcChannels;
      }
    }
    
    /* advance the the position */
    src_offset_i=(gint)(src_offset_d+=scale)*nSrcChannels;
  }
}



/* convert from any number of source channels to any number of destination channels */
static void sample_move_d16_d16(short *dst, short *src,
				unsigned long nsamples, gdouble scale, int nDstChannels, int nSrcChannels, gboolean swap_endian) {
  int nSrcCount, nDstCount;
  register gdouble src_offset_d=0.;
  gint src_offset_i=0;
  gint ccount;

  if(!nSrcChannels) return;

  if (scale<0.) {
    src_offset_d=((gdouble)(nsamples)*(-scale)-1.);
    src_offset_i=(gint)src_offset_d*nSrcChannels;
  }

  while(nsamples--) {
    if ((nSrcCount = nSrcChannels)==(nDstCount = nDstChannels)&&!swap_endian) {
      // same number of channels
      memcpy(dst,src+src_offset_i,nSrcChannels*sizeof(short));
      dst+=nDstCount;
    } 
    else {
      ccount=0;

      /* loop until all of our destination channels are filled */
      while(nDstCount) {
	nSrcCount--;
	nDstCount--;
	
	if (!swap_endian) *(dst++) = *(src+ccount+src_offset_i); /* copy the data over */
	else *(dst++)=((*(src+ccount+src_offset_i)&0x00FF)<<8)+(*(src+ccount+src_offset_i)>>8);
	ccount++;
	
	/* if we ran out of source channels but not destination channels */
	/* then start the src channels back where we were */
	if(!nSrcCount && nDstCount) {
	  ccount=0;
	  nSrcCount = nSrcChannels;
	}
      }
    }
    /* advance the the position */
    src_offset_i=(gint)(src_offset_d+=scale)*nSrcChannels;
  }
}

/* convert from 16 bit to floating point */
/* channels to a buffer that will hold a single channel stream */
/* src_skip is in terms of 16bit samples */
static void sample_move_d16_s16 (sample_t *dst, short *src, unsigned long nsamples, unsigned long src_skip, gboolean is_unsigned, gfloat vol) {
  float svol=SAMPLE_MAX_16BIT/vol;
  register float val;
  
  while (nsamples--) {
    if (!is_unsigned) {
      if ((val = (float)(*src) / svol)>1.0f) val=1.0f;
    }
    else {
      if ((val=((float)((unsigned short)(*src)) / SAMPLE_MAX_16BIT - 1.0)*vol)<-1.0) val=-1.0;
      else if (val>1.0f) val=1.0f;
    }
    *(dst++)=val;
    src += src_skip;
  }
}


inline void sample_silence_dS (sample_t *dst, unsigned long nsamples) {
  memset(dst,0,nsamples*sizeof(sample_t));
}


static int audio_process (nframes_t nframes, void *arg) {
  // JACK calls this periodically to get the next audio buffer
  sample_t* out_buffer[JACK_MAX_OUTPUT_PORTS];
  jack_driver_t* jackd = (jack_driver_t*)arg;
  jack_position_t pos;
  register int i;
  jack_message_t *msg;
  long seek;

#ifdef DEBUG_AJACK
  g_printerr("nframes %ld, sizeof(sample_t) == %d\n", (long)nframes, sizeof(sample_t));
#endif

  if (!mainw->is_ready||jackd==NULL) return 0;

  if(nframes != jackd->chunk_size) jackd->chunk_size = nframes;

  /* retrieve the buffers for the output ports */
  for(i = 0; i < jackd->num_output_channels; i++) out_buffer[i] = (sample_t *) jack_port_get_buffer(jackd->output_port[i], nframes);

  /* process one message */
  if((msg=jackd->msgq)!=NULL) {

    switch (msg->command) {
    case JACK_CMD_FILE_OPEN:
      jackd->fd=open(msg->data,O_RDONLY);
      if (jackd->fd==-1) g_printerr("jack: error opening %s\n",msg->data);
      if (jackd->fd>0) {
	if (jackd->aPlayPtr->data!=NULL) g_free(jackd->aPlayPtr->data);
	if (jackd->buffer_size>0) jackd->aPlayPtr->data=g_malloc(jackd->buffer_size*100);
	else (jackd->aPlayPtr->data)=NULL;
	if (jackd->has_header) jackd->seek_pos=WAV_HEADER_BYTES;
	else jackd->seek_pos=0;
      }
      break;
    case JACK_CMD_FILE_CLOSE:
      if (jackd->fd>0) close(jackd->fd);
      jackd->fd=0;
      if (jackd->aPlayPtr->data!=NULL) g_free(jackd->aPlayPtr->data);
      jackd->aPlayPtr->data=NULL;
      jackd->aPlayPtr->size=0;
      break;
    case JACK_CMD_FILE_SEEK:
      seek=atol(msg->data);
      if (jackd->has_header) seek+=WAV_HEADER_BYTES;
      if (jackd->fd>0&&!jackd->mute) {
	lseek(jackd->fd,seek,SEEK_SET);
      }
      jackd->seek_pos=seek;
      gettimeofday(&tv, NULL);
      jackd->audio_ticks=U_SECL*(tv.tv_sec-mainw->startsecs)+tv.tv_usec*U_SECL/1000000;
      jackd->frames_written=0;
      break;
    }
    if (msg->data!=NULL) g_free(msg->data);
    msg->command=JACK_CMD_PROCESSED;
    jackd->msgq = NULL;    /* take this message off of the queue */
  }  

  jackd->state=jack_transport_query (jackd->client, &pos);

#ifdef DEBUG_AJACK
  g_printerr("STATE is %d %d\n",jackd->state,jackd->play_when_stopped);
#endif

  /* handle playing state */
  if(jackd->state==JackTransportRolling||jackd->play_when_stopped) {
    gulong jackFramesAvailable = nframes; /* frames we have left to write to jack */
    gulong inputFramesAvailable;          /* frames we have available this loop */
    gulong numFramesToWrite;              /* num frames we are writing this loop */
    glong in_frames=0;
    gulong in_bytes;
    gdouble shrink_factor=1.;

    guchar* buffer;

#ifdef DEBUG_AJACK
    g_printerr("playing... jackFramesAvailable = %ld\n", jackFramesAvailable);
#endif

    jackd->num_calls++;

    if (!jackd->in_use||jackd->fd<1||jackd->is_paused) {
      /* output silence if nothing is being outputted */
      if (!jackd->is_silent) for(i = 0; i < jackd->num_output_channels; i++) sample_silence_dS(out_buffer[i], nframes);
      jackd->is_silent=TRUE;
      if (!jackd->is_paused) jackd->frames_written+=nframes;
      return 0;
    }


    if(jackd->buffer_size<(jackFramesAvailable*sizeof(short)*jackd->num_output_channels)) {
      //ERR("our buffer must have changed size\n");
      //ERR("allocated %ld bytes, need %ld bytes\n", jackd->buffer_size,
      // jackFramesAvailable * sizeof(short) * jackd->num_output_channels);
      return 0;
    }

    jackd->is_silent=FALSE;

    if (jackFramesAvailable>0&&jackd->aPlayPtr!=NULL&&jackd->aPlayPtr->data!=NULL&&jackd->num_input_channels>0) {
      /* (bytes of data) / (2 bytes(16 bits) * X input channels) == frames */
      if (jackd->fd>0) {
	jackd->aPlayPtr->size=0;
	
	in_bytes=ABS((in_frames=((gdouble)jackd->sample_in_rate/(gdouble)jackd->sample_out_rate*(gdouble)jackFramesAvailable+(fastrand()/G_MAXUINT32))))*jackd->num_input_channels*jackd->bytes_per_channel;
	if ((shrink_factor=(gdouble)in_frames/(gdouble)jackFramesAvailable)<0.) {
	  if ((jackd->seek_pos-=in_bytes)<(WAV_HEADER_BYTES*jackd->has_header)) {
	    if (*jackd->whentostop==STOP_ON_AUD_END&&jackd->loop==JACK_LOOP_NONE) {
	      *jackd->cancelled=CANCEL_AUD_END;
	    }
	    else {
	      if (jackd->loop==JACK_LOOP_PINGPONG) {
		jackd->sample_in_rate=-jackd->sample_in_rate;
		shrink_factor=-shrink_factor;
		jackd->seek_pos=0+(WAV_HEADER_BYTES*jackd->has_header);
	      }
	      else jackd->seek_pos=jackd->seek_end-in_bytes;
	    }
	  }
	  lseek(jackd->fd,jackd->seek_pos,SEEK_SET);
	}
	if (jackd->mute) {
	  if (shrink_factor>0.) jackd->seek_pos+=in_bytes;
	  if (jackd->seek_pos>=jackd->seek_end) {
	    if (*jackd->whentostop==STOP_ON_AUD_END) {
	      *jackd->cancelled=CANCEL_AUD_END;
	    }
	    else {
	      if (jackd->loop==JACK_LOOP_PINGPONG) {
		jackd->sample_in_rate=-jackd->sample_in_rate;
		jackd->seek_pos-=in_bytes;
	      }
	      else {
		seek=0;
		if (jackd->has_header) seek+=WAV_HEADER_BYTES;
		jackd->seek_pos=seek;
	      }
	    }
	  }
	  for(i = 0; i < jackd->num_output_channels; i++) sample_silence_dS(out_buffer[i], nframes);
	  jackd->frames_written+=nframes;
	  return 0;
	}
	else {
	  gboolean loop_restart;
	  do {
	    loop_restart=FALSE;
	    // playing from a file
	    if (!(*jackd->cancelled)&&ABS(shrink_factor)<=100.) {
	      if ((jackd->aPlayPtr->size=read(jackd->fd,jackd->aPlayPtr->data,in_bytes))==0) {
		if (*jackd->whentostop==STOP_ON_AUD_END) {
		  *jackd->cancelled=CANCEL_AUD_END;
		}
		else {
		  loop_restart=TRUE;
		  if (jackd->loop==JACK_LOOP_PINGPONG) {
		    jackd->sample_in_rate=-jackd->sample_in_rate;
		    lseek(jackd->fd,(jackd->seek_pos-=in_bytes),SEEK_SET);
		  }
		  else {
		    seek=0;
		    if (jackd->has_header) seek+=WAV_HEADER_BYTES;
		    lseek(jackd->fd,(jackd->seek_pos=seek),SEEK_SET);
		  }
		}
	      }
	      else {
		if (shrink_factor<0.) {
		  lseek(jackd->fd,jackd->seek_pos,SEEK_SET);
		}
		else jackd->seek_pos+=jackd->aPlayPtr->size;
	      }
	    }
	  } while (loop_restart);
	}
      }

      inputFramesAvailable = jackd->aPlayPtr->size / (jackd->num_input_channels * jackd->bytes_per_channel);
#ifdef DEBUG_AJACK
      g_printerr("%d inputFramesAvailable == %ld, %ld, %ld %ld,jackFramesAvailable == %ld\n", jackd->aPlayPtr->size, inputFramesAvailable, in_frames,jackd->sample_in_rate,jackd->sample_out_rate,jackFramesAvailable);
#endif
      buffer = jackd->aPlayPtr->data;

      numFramesToWrite = MIN(jackFramesAvailable, inputFramesAvailable/ABS(shrink_factor)); /* write as many bytes as we have space remaining, or as much as we have data to write */

#ifdef DEBUG_AJACK
      g_printerr("inputFramesAvailable after conversion %ld\n", (gulong)((gdouble)inputFramesAvailable/shrink_factor));
      g_printerr("nframes == %d, jackFramesAvailable == %ld,\n\tjackd->num_input_channels == %ld, jackd->num_output_channels == %ld\n",  nframes, jackFramesAvailable, jackd->num_input_channels, jackd->num_output_channels);
#endif

      /* convert from 8 bit to 16 bit and mono to stereo if necessary */
      /* resample as we go */
      if(jackd->bytes_per_channel==1) {
	sample_move_d8_d16 ((short *)(jackd->sound_buffer + (nframes-jackFramesAvailable) / sizeof(short)),
			    (guchar *)buffer, numFramesToWrite, shrink_factor, jackd->num_output_channels, jackd->num_input_channels);
      }
      /* 16 bit input samples */
      /* resample as we go */
      else {
	sample_move_d16_d16((short*)jackd->sound_buffer + ((nframes - jackFramesAvailable) * jackd->bytes_per_channel * jackd->num_output_channels) / sizeof(short), (short*)buffer, numFramesToWrite, shrink_factor, jackd->num_output_channels, jackd->num_input_channels, jackd->reverse_endian);
      }

      jackd->frames_written+=numFramesToWrite;
      jackFramesAvailable-=numFramesToWrite; /* take away what was written */

#ifdef DEBUG_AJACK
      g_printerr("jackFramesAvailable == %ld\n", jackFramesAvailable);
#endif
    }

    for(i=0;i<jackd->num_output_channels;i++) {
      sample_move_d16_s16(out_buffer[i], (short*)jackd->sound_buffer + i, (nframes - jackFramesAvailable), jackd->num_output_channels, jackd->usigned, jackd->volume[i]);
    }


    if(jackFramesAvailable) {
#ifdef DEBUG_AJACK
      g_printerr("buffer underrun of %ld frames\n", jackFramesAvailable);
#endif
      for(i = 0 ; i < jackd->num_output_channels; i++) sample_silence_dS(out_buffer[i] + (nframes - jackFramesAvailable), jackFramesAvailable);
    }
  }


  else if(jackd->state == JackTransportStarting || jackd->state == JackTransportStopped || jackd->state == JackTClosed || jackd->state == JackTReset) {
#ifdef DEBUG_AJACK
    g_printerr("PAUSED or STOPPED or CLOSED, outputting silence\n");
#endif
    
    /* output silence if nothing is being outputted */
    for(i = 0; i < jackd->num_output_channels; i++) sample_silence_dS(out_buffer[i], nframes);
    /* if we were told to reset then zero out some variables */
    /* and transition to STOPPED */
    if(jackd->state == JackTReset) {
      jackd->state = JackTStopped; /* transition to STOPPED */
    }
  }

#ifdef DEBUG_AJACK
  g_printerr("done\n");
#endif

  return 0;
}

static int jack_get_bufsize (nframes_t nframes, void *arg) {
  jack_driver_t* jackd = (jack_driver_t*)arg;
  unsigned long buffer_required;
  //g_printerr("the maximum buffer size is now %lu frames\n", (long)nframes);

  /* make sure the callback routine has adequate memory for the nframes it will get */
  /* ie. Buffer_size < (bytes we already wrote + bytes we are going to write in this loop) */
  /* frames * 2 bytes in 16 bits * X channels of output */
  buffer_required = nframes * sizeof(short) * jackd->num_output_channels;
  if(jackd->buffer_size < buffer_required) {
    //g_printerr("expanding buffer from jackd->buffer_size == %ld, to %ld\n",jackd->buffer_size, buffer_required);
    jackd->buffer_size = buffer_required;
    jackd->sound_buffer = realloc(jackd->sound_buffer, jackd->buffer_size);

    /* if we don't have a buffer then error out */
    if(jackd->sound_buffer==NULL) {
      return 1;
    }
  }
  
  if (jackd->aPlayPtr->data!=NULL) g_free(jackd->aPlayPtr->data);
  if (jackd->fd>0) jackd->aPlayPtr->data=g_malloc(jackd->buffer_size*100);
;

  return 0;
}


int jack_get_srate (nframes_t nframes, void *arg) {
  //g_printerr("the sample rate is now %ld/sec\n", (long)nframes);
  return 0;
}


void jack_shutdown(void* arg) {
  jack_driver_t* jackd = (jack_driver_t*)arg;

  jackd->client = 0; /* reset client */
  jackd->jackd_died = TRUE;

  g_printerr("jack shutdown, setting client to 0 and jackd_died to true\n");
  g_printerr("trying to reconnect right now\n");

  /////////////////////

  jack_audio_init();
  mainw->jackd=jack_get_driver(0);
  mainw->jackd->msgq=NULL;

  audio_seek_bytes(mainw->jackd,mainw->jackd->seek_pos); // at least re-seek to the right place
}

/* Return the difference between two timeval structures in terms of milliseconds */
inline long TimeValDifference(struct timeval *start, struct timeval *end) {
  return (long)((gdouble)(end->tv_sec-start->tv_sec)*(gdouble)1000.+(gdouble)(end->tv_usec-start->tv_usec)/(gdouble)1000.);
}


static void jack_reset_driver(jack_driver_t *jackd) {
  //g_printerr("resetting jackd->dev_idx(%d)\n", jackd->dev_idx);
  /* tell the callback that we are to reset, the callback will transition this to STOPPED */
  jackd->state=JackTReset;
}


void jack_close_device(jack_driver_t* jackd, gboolean close_client) {
  int i;

  if (close_client) {
    //g_printerr("closing the jack client thread\n");
    if (jackd->client) {
      jack_deactivate(jackd->client); /* supposed to help the jack_client_close() to succeed */
      //g_printerr("after jack_deactivate()\n");
      jack_client_close(jackd->client);
    }

    jack_reset_driver(jackd);
    jackd->client=0;
    free(jackd->sound_buffer);
    jackd->sound_buffer=NULL;
    jackd->buffer_size=0;

    /* free up the port strings */
    //g_printerr("freeing up port strings\n");
    if (jackd->jack_port_name_count>1) {
      for (i=0;i<jackd->jack_port_name_count;i++) free(jackd->jack_port_name[i]);
      free(jackd->jack_port_name);
    }
  }
  else {
    if(!jackd->client) {
      g_printerr("critical error, closing a device that has no client\n");
    }
  }
}




static void jack_error_func(const char *desc) {
  g_printerr("Jack audio error %s\n",desc);
}



// create a new client and connect it to jack, connect the ports
int jack_open_device(jack_driver_t *jackd) {
  const char *client_name="LiVES_audio_out";
  const char *server_name="";
  jack_options_t options=JackNullOption;
  jack_status_t status;
  int i;

  /* zero out the buffer pointer and the size of the buffer */
  jackd->sound_buffer=NULL;
  jackd->buffer_size=0;

  /* set up an error handler */
  jack_set_error_function(jack_error_func);
  jackd->client=NULL;
  while (jackd->client==NULL) 
    jackd->client = jack_client_open (client_name, options, &status, server_name);
    
  if (jackd->client==NULL) {
    g_printerr (g_strdup_printf("jack_client_open() failed, status = 0x%2.0x\n", status));
    if (status&JackServerFailed) {
      d_print ("Unable to connect to JACK server\n");
    }
    return 1;
  }
  
  if (status&JackServerStarted) {
    d_print ("JACK server started\n");
  }
  
  if (status&JackNameNotUnique) {
    client_name= jack_get_client_name(jackd->client);
    g_printerr (g_strdup_printf("unique name `%s' assigned\n", client_name));
  }
  
  jackd->sample_out_rate=jack_get_sample_rate(jackd->client);
  
  //g_printerr (g_strdup_printf("engine sample rate: %ld\n",jackd->sample_rate));

  for (i=0;i<jackd->num_output_channels;i++) {
    gchar portname[32];
    g_snprintf(portname, 32, "out_%d", i);

#ifdef DEBUG_JACK_PORTS
    g_printerr("port %d is named '%s'\n", i, portname);
#endif
    jackd->output_port[i] = jack_port_register(jackd->client, portname,
					       JACK_DEFAULT_AUDIO_TYPE,
					       JackPortIsOutput,
					       0);
    if (jackd->output_port[i]==NULL) {
      g_printerr("no more JACK ports available\n");
      return 1;
    }
    jackd->out_chans_available++;
  }
  
  // set process callback and start
  jack_set_process_callback (jackd->client, audio_process, jackd);  

  /* setup a buffer size callback */
  jack_set_buffer_size_callback(jackd->client, jack_get_bufsize, jackd);

  /* tell the JACK server to call `srate()' whenever
     the sample rate of the system changes. */
  jack_set_sample_rate_callback(jackd->client, jack_get_srate, jackd);

  /* tell the JACK server to call `jack_shutdown()' if
     it ever shuts down, either entirely, or if it
     just decides to stop calling us. */
  jack_on_shutdown(jackd->client, jack_shutdown, jackd);

  /* set the initial buffer size */
  jack_get_bufsize(jack_get_buffer_size(jackd->client), jackd);
  return 0;


}


int jack_driver_activate (jack_driver_t *jackd) {
  int i;
  const char** ports;
  gboolean failed=FALSE;

  /* tell the JACK server that we are ready to roll */
  if (jack_activate(jackd->client)) {
    //ERR( "cannot activate client\n");
    return 1;
  }

  // we are looking for input ports to connect to
  jackd->jack_port_flags|=JackPortIsInput;

  /* determine how we are to acquire port names */
  if ((jackd->jack_port_name_count==0)||(jackd->jack_port_name_count==1)) {
    if(jackd->jack_port_name_count==0) {
      //g_printerr("jack_get_ports() passing in NULL/NULL\n");
      ports=jack_get_ports(jackd->client, NULL, NULL, jackd->jack_port_flags);
    }
    else {
      //g_printerr("jack_get_ports() passing in port of '%s'\n", jackd->jack_port_name[0]);
      ports=jack_get_ports(jackd->client, jackd->jack_port_name[0], NULL, jackd->jack_port_flags);
    }

    /* display a trace of the output ports we found */
#ifdef DEBUG_JACK_PORTS
    for (i=0;ports[i];i++) g_printerr("ports[%d] = '%s'\n",i,ports[i]);
#endif

    /* see if we have enough ports */
    if(jackd->out_chans_available<jackd->num_output_channels) {
#ifdef DEBUG_JACK_PORTS
      g_printerr("ERR: jack_get_ports() failed to find ports with jack port flags of 0x%lX'\n", jackd->jack_port_flags);
#endif
      return ERR_PORT_NOT_FOUND;
    }

    /* connect the ports. Note: you can't do this before
       the client is activated (this may change in the future). */
    for(i=0;i<jackd->num_output_channels;i++) {
#ifdef DEBUG_JACK_PORTS
      g_printerr("jack_connect() to port %d('%p')\n", i, jackd->output_port[i]);
#endif
      if(jack_connect(jackd->client, jack_port_name(jackd->output_port[i]), ports[i])) {
#ifdef DEBUG_JACK_PORTS
          g_printerr("cannot connect to output port %d('%s')\n", i, ports[i]);
#endif
          failed=TRUE;
      }
    } 
    free(ports); /* free the returned array of ports */
  }
  else {
    for(i=0;i<jackd->jack_port_name_count;i++) {
#ifdef DEBUG_JACK_PORTS
      g_printerr("jack_get_ports() portname %d of '%s\n", i, jackd->jack_port_name[i]);
#endif
      ports=jack_get_ports(jackd->client, jackd->jack_port_name[i], NULL, jackd->jack_port_flags);
#ifdef DEBUG_JACK_PORTS
      g_printerr ("ports[%d] = '%s'\n", 0, ports[0]);      /* display a trace of the output port we found */
#endif
      if(!ports) {
#ifdef DEBUG_JACK_PORTS
	g_printerr("jack_get_ports() failed to find ports with jack port flags of 0x%lX'\n", jackd->jack_port_flags);
#endif
	return ERR_PORT_NOT_FOUND;
      }

      /* connect the port */
#ifdef DEBUG_JACK_PORTS
      g_printerr ("jack_connect() to port %d('%p')\n", i, jackd->output_port[i]);
#endif
      if(jack_connect(jackd->client, jack_port_name(jackd->output_port[i]), ports[0])) {
	//ERR("cannot connect to output port %d('%s')\n", 0, ports[0]);
	failed=TRUE;
      }
      free(ports); /* free the returned array of ports */
    }
  }

  /* if something failed we need to shut the client down and return 0 */
  if (failed) {
    g_printerr("failed, closing and returning error\n");
    jack_close_device(jackd,TRUE);
    return 1;
  }


  jackd->jackd_died = FALSE;

  jackd->in_use=FALSE;

  jackd->is_paused=FALSE;

  d_print(_("Started jack audio subsystem.\n"));

  return 0;
}



jack_driver_t *jack_get_driver(gint dev_idx) {
  jack_driver_t *jackd = &outdev[dev_idx];
#ifdef TRACE_getReleaseDevice
  g_printerr("dev_idx is %d\n", dev_idx);
#endif
  
  /* should we try to restart the jack server? */
  if (jackd->jackd_died&&jackd->client==0) {
    struct timeval now;
    gettimeofday(&now, 0);
    
    /* wait 250ms before trying again */
    if(TimeValDifference(&jackd->last_reconnect_attempt, &now)>=250) {
      jack_open_device(jackd);
      jackd->last_reconnect_attempt=now;
    }
  }
  
  return jackd;
}



static void jack_reset_dev(gint dev_idx) {
  jack_driver_t *jackd = jack_get_driver(dev_idx);
  //g_printerr("resetting dev %d\n", dev_idx);
  jack_reset_driver(jackd);
}


int jack_audio_init(void) {
  int i,j;
  jack_driver_t *jackd;

  for (i=0;i<JACK_MAX_OUTDEVICES;i++) {
    jackd = &outdev[i];
    jack_reset_dev(i);
    jackd->dev_idx=i;
    jackd->client=NULL;
    jackd->in_use=FALSE;
    for (j=0;j<JACK_MAX_OUTPUT_PORTS;j++) jackd->volume[j]=1.0f;
    jackd->state=JackTClosed;
    jackd->sample_out_rate=jackd->sample_in_rate=0;
    jackd->fd=0;
    jackd->seek_pos=jackd->seek_end=0;
    jackd->msgq=NULL;
    jackd->num_calls=0;
    jackd->chunk_size=0;
    jackd->jackd_died=FALSE;
    jackd->aPlayPtr=(audio_buffer_t *)g_malloc(sizeof(audio_buffer_t));
    jackd->aPlayPtr->data=NULL;
    gettimeofday(&jackd->last_reconnect_attempt, 0);
    jackd->num_output_channels=2;
    jackd->play_when_stopped=FALSE;
    jackd->mute=FALSE;
    jackd->out_chans_available=0;
  }
  return 0;
}


jack_message_t *jack_get_msgq(jack_driver_t *jackd) {
  // force update
  gchar *tmp=g_strdup_printf("%p",jackd->msgq);
  g_free(tmp);
  if (jackd->jackd_died) return NULL;
  return jackd->msgq;
}

gint64 jack_get_time(jack_driver_t *jackd) {
  jack_message_t *msg=jackd->msgq;
  if (msg!=NULL&&msg->command==JACK_CMD_FILE_SEEK) while (jack_get_msgq(jackd)!=NULL); // wait for seek
  return jackd->audio_ticks+(gint64)(((gdouble)jackd->frames_written/(gdouble)jackd->sample_out_rate)*U_SEC);
}


static jack_message_t jack_message;

void audio_seek_frame (jack_driver_t *jackd, gint frame) {
  // seek to frame "frame" in current file
  long seekstart;
  if (frame<1) frame=1;
  if (frame>cfile->frames) frame=cfile->frames;
  seekstart=(long)((gdouble)(frame-1.)/cfile->fps*cfile->arate)*cfile->achans*cfile->asampsize/8;
  audio_seek_bytes(jackd,seekstart);
  cfile->aseek_pos=seekstart;
}


void audio_seek_bytes (jack_driver_t *jackd, long bytes) {
  // seek to relative "secs" in current file
  jack_message_t *jmsg;
  if (jackd->has_header) bytes+=WAV_HEADER_BYTES;
  do {
    jmsg=jack_get_msgq(jackd);
  } while ((jmsg!=NULL)&&jmsg->command!=JACK_CMD_FILE_SEEK);
  jack_message.command=JACK_CMD_FILE_SEEK;
  jack_message.next=NULL;
  jack_message.data=g_strdup_printf("%ld",bytes);
  jackd->msgq=&jack_message;
}


