// multitrack.h
// LiVES
// (c) G. Finch 2005 <salsaman@xs4all.nl>
// released under the GNU GPL 2.0
// see file ../COPYING for licensing details

// multitrack window

#ifndef __HAS_MULTITRACK_H__
#define __HAS_MULTITRACK_H__

#include "events.h"

#define CLIP_THUMB_WIDTH 80
#define CLIP_THUMB_HEIGHT 80

#define CLIP_LABEL_LENGTH 20

#define BLOCK_THUMB_WIDTH 40

#define BLOCK_DRAW_SIMPLE 1
#define BLOCK_DRAW_THUMB 2
#define BLOCK_DRAW_TYPE BLOCK_DRAW_THUMB

#define MT_PLAY_WIDTH 320
#define MT_PLAY_HEIGHT 240

#define TIMECODE_LENGTH 14 // length of timecode text entry, must be >12

#define DEF_TIME 120 // default seconds when there is no event_list

typedef struct _mt lives_mt;

typedef struct _track_rect track_rect;

struct _mt {
  // widgets
  GtkWidget *window;
  GtkWidget *top_vbox;
  GtkWidget *hbox;
  GtkWidget *play_blank;
  GtkWidget *play_box;
  GtkWidget *poly_box;
  GtkWidget *clip_scroll;
  GtkWidget *clip_inner_box;
  GtkWidget *in_out_box;
  GtkWidget *in_hbox;
  GtkWidget *out_hbox;
  GtkWidget *in_image;
  GtkWidget *out_image;
  GtkWidget *context_box;
  GtkWidget *sep_image;
  GtkWidget *timeline_table_header;
  GtkWidget *tl_eventbox;
  GtkWidget *timeline_table;
  GtkWidget *timeline;
  GtkWidget *timeline_eb;
  GtkWidget *infobar;
  GtkWidget *stop;
  GtkWidget *insert;
  GtkWidget *delblock;
  GtkWidget *clipedit;
  GtkWidget *vpaned;
  GtkWidget *scrollbar;
  GtkWidget *playall;
  GtkWidget *jumpto;
  GtkWidget *render;
  GtkWidget *fx_block;
  GtkWidget *fx_region;
  GtkWidget *move_fx;
  GtkWidget *mm_menuitem;
  GtkWidget *mm_move;
  GtkWidget *mm_select;
  GtkWidget *snapt_menuitem;
  GtkWidget *snapt_on;
  GtkWidget *snapt_off;
  GtkWidget *select_track;
  GtkWidget *view_events;
  GtkWidget *view_clips;
  GtkWidget *view_in_out;
  GtkWidget *spinbutton_in;
  GtkWidget *spinbutton_out;
  GtkWidget *checkbutton_start_anchored;
  GtkWidget *checkbutton_end_anchored;
  GtkWidget *timecode;
  GtkWidget *spinbutton_start;
  GtkWidget *spinbutton_end;
  GtkWidget *tl_hbox;
  GtkWidget *fx_box;
  GtkWidget *next_node_button;
  GtkWidget *prev_node_button;
  GtkWidget *del_node_button;
  GtkWidget *node_spinbutton;
  GtkWidget *sel_label;
  GtkWidget *preview_button; // button
  GtkWidget *save_event_list; // menuitem
  GtkWidget *load_event_list; // menuitem
  GtkWidget *undo;
  GtkWidget *redo;
  GtkWidget *remove_gaps;
  GtkWidget *remove_first_gaps;

  GtkAccelGroup *accel_group;
  GList *video_draws; // list of video timeline eventboxes, in layer order
  GtkObject *vadjustment;

  GdkDisplay *display;

  gulong spin_in_func;
  gulong spin_out_func;
  gulong check_start_func;
  gulong check_end_func;

  gulong mm_move_func;
  gulong mm_select_func;
  
  gulong snapt_on_func;
  gulong snapt_off_func;

  weed_plant_t *event_list;

  weed_plant_t *init_event;
  ///////////////////////
  gint num_video_tracks;
  gdouble end_secs;  // max display time of timeline in seconds

  gint clip_selected; // clip in clip window
  gint file_selected; // actual LiVES file struct number which clip_selected matches
  gint current_track; // starts at 0

  GList *selected_tracks;

  gshort poly_state;  // state of polymorph window
#define POLY_NONE 0
#define POLY_CLIPS 1
#define POLY_IN_OUT 2
#define POLY_EFFECT 3


  gshort insert_mode;
#define INSERT_MODE_NORMAL 1  // the default
#define INSERT_MODE_OVERWRITE 2
#define INSERT_MODE_EXPAND 3
  //#define INSERT_MODE_FILL 4

  gint render_file;

  gshort last_direction; // last direction timeline cursor was moved in
#define DIRECTION_NEGATIVE 0
#define DIRECTION_POSITIVE 1

  track_rect *block_selected; // pointer to current selected block, or NULL
  track_rect *putative_block; // putative block to move or copy, or NULL

  gdouble ptr_time; // stored timeline cursor position before playback

  gdouble tl_fixed_length; // length of timeline can be fixed (secs) : TODO

  gdouble fps; // fps of this timeline

  gdouble region_start; // start of time region in seconds (or 0.)
  gdouble region_end; // end of time region in seconds (or 0.)
  gboolean region_updating;

  gboolean is_rendering; // TRUE if we are is the process of rendering to a clip, cf. mainw->is_rendering

#define MAX_DISP_VTRACKS 6
  gint max_disp_vtracks;
  gboolean snap_over;
  gint current_fx;

  gboolean mt_frame_preview; // stops playback and preview contending for play window

  lives_rendered_effect_t *current_rfx;

  gchar layout_name[256];

  // audio details - not used yet
  guint signed_endian; // bitfield

  gint arate; // audio playback rate
  gint achans;
  gint asampsize;

  // cursor warping for mouse move mode
  gint hotspot_x;
  gint hotspot_y;

  gboolean move_effects; // should we move effects attached to a block ?
  gboolean moving_block; // moving block flag

  gshort mouse_mode;
#define MOUSE_MODE_MOVE 1
#define MOUSE_MODE_SELECT 2
#define MOUSE_MODE_COPY 3

  gint sel_x;
  gint sel_y;

  gulong mouse_mot1;
  gulong mouse_mot2;

  gboolean tl_selecting; // for mouse select mode

  gboolean snapt; // block snap

  /* start and end offset overrides for inserting (used during block move) */
  weed_timecode_t insert_start;
  weed_timecode_t insert_end;

  GList *undos;
  size_t undo_buffer_used;
  unsigned char *undo_mem;
  gboolean did_backup;
  gint undo_offset;

  gchar undo_text[32];
  gchar redo_text[32];

  gboolean undoable;
  gboolean redoable;

  gboolean changed;

};


// undo actions
// no event_list
#define MT_UNDO_NONE 0
#define MT_UNDO_INSERT_BLOCK 1

// minimal event_list
#define MT_UNDO_APPLY_FILTER 512
#define MT_UNDO_SPLIT 513
#define MT_UNDO_SPLIT_MULTI 514

// full backups of event_list
#define MT_UNDO_DELETE_BLOCK 1024
#define MT_UNDO_MOVE_BLOCK 1025
#define MT_UNDO_REMOVE_GAPS 1026


typedef struct {
  gint action;
  void *extra;
  size_t data_len; // including this mt_undo
} mt_undo;





// reasons for track invisibility (bitmap)
#define TRACK_I_OUTSCROLLED_PRE (1<<0)
#define TRACK_I_OUTSCROLLED_POST (1<<1)


struct _track_rect {
  // track rectangles (blocks), we translate our event_list FRAME events into these, then when exposed, the eventbox draws them
  // blocks MUST only contain frames from a single clip. They MAY NOT contain blank frames.
  // start and end events MUST be FRAME events
  track_rect *next;
  track_rect *prev;
  weed_plant_t *start_event;
  weed_plant_t *end_event;

  gint64 offset_start; // offset in sourcefile of first frame

  gshort state;
  gboolean start_anchored;
  gboolean end_anchored;
  gboolean ordered; // are frames in sequential order ?
  
  GtkWidget *eventbox; // pointer to eventbox widget which contains this block; we can use its "layer_number" to get the track/layer number

  // if block is bounded either side by another block, we set the limits here : TODO
  gint64 min_tc;
  gint64 max_tc;
};

// block state
#define BLOCK_UNSELECTED 0
#define BLOCK_SELECTED 1


/* translation table for matching event_id to init_event */
typedef struct {
  void *in;
  void *out;
} ttable;


/* clip->layout use mapping, from layout.map file */
typedef struct {
  gchar *handle;
  gint64 unique_id;
  gchar *name;
  GList *list;
} layout_map;





//////////////////////////////////////////////////////////////////

// setup functions
gboolean on_multitrack_activate (GtkMenuItem *, weed_plant_t *); // widget callback to create the multitrack window
lives_mt *multitrack (weed_plant_t *, gint orig_file, gdouble fps); // create and return lives_mt struct
void init_tracks (lives_mt *mt);  // add basic tracks
void init_clips (lives_mt *mt, gint orig_file, gboolean add); // init clip window (or add a new clip)

// delete functions
gboolean on_mt_delete_event (GtkWidget *, GdkEvent *, gpointer mt);
void multitrack_delete (lives_mt *); // free the lives_mt struct
gboolean multitrack_end (GtkMenuItem *, gpointer mt);

// morph the poly window
void polymorph (lives_mt *, gshort poly);

/// sens/desens
void mt_desensitise (lives_mt *);
void mt_sensitise (lives_mt *);

// external control callbacks
void insert_here_cb (GtkMenuItem *menuitem, gpointer mt);
void multitrack_end_cb (GtkMenuItem *, gpointer mt);
void delete_block_cb (GtkMenuItem *, gpointer mt);


//// menuitem callbacks
void on_add_video_track_activate (GtkMenuItem *, gpointer mt);
void multitrack_playall (GtkWidget *, gpointer mt); // menuitem playall
void multitrack_insert (GtkMenuItem *, gpointer mt); // menuitem insert
void multitrack_view_events (GtkMenuItem *, gpointer mt); // menuitem insert
void on_render_activate (GtkMenuItem *, gpointer mt);
void on_jumpto_activate (GtkMenuItem *, gpointer mt);
void on_delblock_activate (GtkMenuItem *, gpointer mt);
void on_seltrack_activate (GtkMenuItem *menuitem, gpointer mt);
void multitrack_view_details (GtkMenuItem *menuitem, gpointer mt);
void mt_add_block_effect (GtkMenuItem *menuitem, gpointer mt);
void on_save_event_list_activate (GtkMenuItem *menuitem, gpointer mt);
void on_load_event_list_activate (GtkMenuItem *menuitem, gpointer mt);
void show_frame_events_activate (GtkMenuItem *menuitem, gpointer);
void on_set_pvals_clicked  (GtkWidget *button, gpointer mt);
void on_snap_over_changed (GtkMenuItem *menuitem, gpointer mt);
void select_all_time (GtkMenuItem *menuitem, gpointer mt);
void select_from_zero_time (GtkMenuItem *menuitem, gpointer mt);
void select_to_end_time (GtkMenuItem *menuitem, gpointer mt);
void select_all_vid (GtkMenuItem *menuitem, gpointer mt);
void select_no_vid (GtkMenuItem *menuitem, gpointer mt);
void on_split_sel_activate (GtkMenuItem *menuitem, gpointer mt);
void multitrack_undo (GtkMenuItem *menuitem, gpointer mt);
void multitrack_redo (GtkMenuItem *menuitem, gpointer mt);

/// track functions
void add_video_track_behind (GtkMenuItem *, gpointer mt);
void add_video_track_front (GtkMenuItem *, gpointer mt);
void delete_video_track(lives_mt *mt, gint layer, gboolean full);
void remove_gaps (GtkMenuItem *menuitem, gpointer mt);
void remove_first_gaps (GtkMenuItem *menuitem, gpointer mt);
void on_split_activate (GtkMenuItem *menuitem, gpointer mt);

// track mouse movement
gboolean on_track_click (GtkWidget *eventbox, GdkEventButton *, gpointer mt);
gboolean on_track_header_click (GtkWidget *eventbox, GdkEventButton *, gpointer mt);
gboolean on_track_between_click (GtkWidget *eventbox, GdkEventButton *, gpointer mt);
gboolean on_track_release (GtkWidget *eventbox, GdkEventButton *event, gpointer mt);
gboolean on_track_header_release (GtkWidget *eventbox, GdkEventButton *, gpointer mt);
gboolean on_track_between_release (GtkWidget *eventbox, GdkEventButton *, gpointer mt);
gboolean on_track_move (GtkWidget *widget, GdkEventMotion *event, gpointer mt);
gboolean on_track_header_move (GtkWidget *widget, GdkEventMotion *event, gpointer mt);

void unselect_all (lives_mt *); // unselect all blocks
void insert_frames (gint filenum, gint64 offset_start, gint64 offset_end, gint64 tc, gshort direction, GtkWidget *eventbox, lives_mt *, track_rect *in_block);
void on_seltrack_toggled (GtkToggleButton *, gpointer mt);
void scroll_track_by_scrollbar (GtkVScrollbar *sbar, gpointer mt);

// block functions
void in_out_start_changed (GtkWidget *, gpointer mt);
void in_out_end_changed (GtkWidget *, gpointer mt);
void in_anchor_toggled (GtkToggleButton *, gpointer mt);
void out_anchor_toggled (GtkToggleButton *, gpointer mt);

// timeline functions
void set_timeline_end_secs (lives_mt *, gdouble secs);
gboolean on_timeline_press (GtkWidget *, GdkEventButton *, gpointer mt);
gboolean on_timeline_release (GtkWidget *, GdkEventButton *, gpointer mt);
gboolean on_timeline_update (GtkWidget *, GdkEventMotion *, gpointer mt);
gint expose_timeline_event (GtkWidget *, GdkEventExpose *, gpointer mt);


// context box text
void clear_context (lives_mt *mt);
void add_context_label (lives_mt *, gchar *text);
void mouse_mode_context(lives_mt *mt);
void do_sel_context (lives_mt *mt);

// playback animation
void animate_multitrack (lives_mt *);


// effect node controls
void on_next_node_clicked  (GtkWidget *, gpointer mt);
void on_prev_node_clicked  (GtkWidget *, gpointer mt);
void on_del_node_clicked  (GtkWidget *, gpointer mt);
void on_node_spin_value_changed (GtkSpinButton *, gpointer mt);

void on_frame_preview_clicked (GtkButton *, gpointer mt);


// utils
guint event_list_get_byte_size(weed_plant_t *event_list, int *num_events);  // returns bytes and sets num_events
gboolean event_list_rectify(weed_plant_t *event_list, gboolean check_clips);
gboolean make_backup_space (lives_mt *mt, size_t space_needed);

// event_list utilities
gboolean compare_filter_maps(weed_plant_t *fm1, weed_plant_t *fm2);
void update_filter_events(lives_mt *mt, weed_plant_t *first_event, weed_timecode_t start_tc, weed_timecode_t end_tc, int track, weed_timecode_t new_start_tc, int new_track);

// layout maps
GList *load_layout_map(void);
void save_layout_map (int *lmap, gchar *file, gchar *dir);
void migrate_layouts (gchar *old_set_name, gchar *new_set_name);
gboolean layout_frame_is_affected(gint clipno, gint frame);
void add_markers(lives_mt *mt, weed_plant_t *event_list);
void remove_markers(weed_plant_t *event_list);

/* default to warn about */
#define LMAP_ERROR_MISSING_CLIP 1
#define LMAP_ERROR_CLOSE_FILE 2
#define LMAP_ERROR_DELETE_FRAMES 3

/*non-default*/
#define LMAP_ERROR_SHIFT_FRAMES 4
#define LMAP_ERROR_ALTER_FRAMES 5

/* info */
#define LMAP_INFO_SETNAME_CHANGED 16


#define LIVES_CURSOR_NORMAL 0
#define LIVES_CURSOR_BLOCK 1


#endif
