// multitrack.c
// LiVES
// (c) G. Finch 2005 <salsaman@xs4all.nl>
// released under the GNU GPL 2.0
// see file ../COPYING for licensing details

// multitrack window

// layout is:

// play | clips/params | ctx menu
// ------------------------------
//            timeline
// ------------------------------
//            messages


#include "main.h"
#include "events.h"
#include "callbacks.h"
#include "effects-weed.h"
#include "resample.h"
#include "support.h"
#include "paramwindow.h"
#include "interface.h"

static int renumbered_clips[MAX_FILES+1]; // used to match clips from the event recorder with renumbered clips (without gaps)
static gdouble lfps[MAX_FILES+1]; // table of layout fps
static void **pchain; // param chain for currently being edited filter

GdkPixbuf *make_thumb (gint file, gint width, gint height, gint frame) {
  // TODO - use layer
  GdkPixbuf *pixbuf,*thumbnail;
  GError *error=NULL;
  gchar *buf;
  gint retries=1;

  if (mainw->playing_file>-1) retries=10000000;

  buf=g_strdup_printf("%s/%s/%08d.%s",prefs->tmpdir,mainw->files[file]->handle,frame,prefs->image_ext);
  pixbuf=gdk_pixbuf_new_from_file(buf,&error);
  g_free(buf);
  if (error==NULL) {
    thumbnail=gdk_pixbuf_scale_simple(pixbuf,width,height,GDK_INTERP_BILINEAR);
    gdk_pixbuf_unref(pixbuf);
  }
  else return NULL;
  return thumbnail;
}

inline gchar *get_track_name(lives_mt *mt, int track_num) {
  GtkWidget *xeventbox=g_list_nth_data(mt->video_draws,track_num);
  return g_object_get_data(G_OBJECT(xeventbox),"track_name");
}



static void draw_block (lives_mt *mt,track_rect *block, gdouble end_secs, gint x1, gint x2) {
  GtkWidget *eventbox=block->eventbox;
  weed_plant_t *event=block->start_event;
  weed_timecode_t tc=get_event_timecode(event);
  gint offset_start=tc/U_SEC/end_secs*eventbox->allocation.width;
  gint offset_end;
  int i,filenum,track;
  GdkPixbuf *thumbnail;
  int framenum;
  gint width=BLOCK_THUMB_WIDTH;

  if ((x1>0||x2>0)&&offset_start>(x1+x2)) return;

  offset_end=(get_event_timecode(block->end_event)/U_SEC+1./cfile->fps)/end_secs*eventbox->allocation.width;

  if (offset_end<x1) return;

  switch (block->state) {
  case BLOCK_UNSELECTED:
    track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"layer_number"));
    filenum=get_frame_event_clip(block->start_event,track);
    if (BLOCK_DRAW_TYPE==BLOCK_DRAW_SIMPLE) {
      gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, 0, offset_start, eventbox->allocation.height);
      gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_end, eventbox->allocation.height, offset_end, 0);
      gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, 0, offset_end, eventbox->allocation.height);
      gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, eventbox->allocation.height, offset_end, 0);
      gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, 0, offset_end, 0);
      gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, eventbox->allocation.height-1, offset_end, eventbox->allocation.height-1);
    }
    else {
      for (i=offset_start;i<offset_end;i+=BLOCK_THUMB_WIDTH) {
	// create a small thumb
	event=get_frame_event_at(mt->event_list,tc,event,FALSE);
	tc+=end_secs/eventbox->allocation.width*width*U_SEC;
	framenum=get_frame_event_frame(event,track);
	
	thumbnail=make_thumb(filenum,width,eventbox->allocation.height-1,framenum);
	if (i+width>offset_end) width=offset_end-i;
	// render it in the eventbox
	if (thumbnail!=NULL) {
	  gdk_pixbuf_render_to_drawable (thumbnail,GDK_DRAWABLE (eventbox->window),mainw->gc,0,0,i,0,width,-1,GDK_RGB_DITHER_NONE,0,0);
	  gdk_pixbuf_unref(thumbnail);
	}
	gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, 0, offset_start, eventbox->allocation.height-1);
	gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_end, eventbox->allocation.height-1, offset_end, 0);
	gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, 0, offset_end, 0);
	gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, eventbox->allocation.height-1, offset_end, eventbox->allocation.height-1);
      }
      if (mainw->playing_file>-1&&mainw->cancelled==CANCEL_NONE) process_one(FALSE);
    }
    break;
  case BLOCK_SELECTED:
    gdk_draw_rectangle (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->bg_gc[0], TRUE, offset_start, 0, offset_end-offset_start, eventbox->allocation.height-1);
    gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, 0, offset_start, eventbox->allocation.height-1);
    gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_end, eventbox->allocation.height-1, offset_end, 0);
    gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, 0, offset_end, eventbox->allocation.height-1);
    gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, eventbox->allocation.height-1, offset_end, 0);
    gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, 0, offset_end, 0);
    gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset_start, eventbox->allocation.height-1, offset_end, eventbox->allocation.height-1);
    break;
  }
  if (mainw->playing_file>-1&&mainw->cancelled==CANCEL_NONE) process_one(FALSE);
}



static gint expose_vidtrack_event (GtkWidget *eventbox, GdkEventExpose *event, gpointer user_data) {
  track_rect *block;
  lives_mt *mt=(lives_mt *)user_data;
  GdkRegion *reg=event->region;
  GdkRectangle rect;
  gdk_region_get_clipbox(reg,&rect);

  gint startx=rect.x,width=rect.width;
  gint hidden;

  if (event!=NULL&&event->count>0) {
    return TRUE;
  }

  hidden=(gint)GPOINTER_TO_INT(g_object_get_data (G_OBJECT(eventbox), "hidden"));
  if (hidden) {
    GtkWidget *label=g_object_get_data (G_OBJECT(eventbox), "label");
    gtk_widget_hide(eventbox);
    gtk_widget_hide(label->parent);
    return FALSE;
  }

  if (width>eventbox->allocation.width-startx) width=eventbox->allocation.width-startx;

  gdk_window_clear_area (eventbox->window, startx, 0, width, eventbox->allocation.height);
  if (mainw->playing_file>-1&&mainw->cancelled==CANCEL_NONE) process_one(FALSE);

  block=(track_rect *)g_object_get_data (G_OBJECT(eventbox), "blocks");

  while (block!=NULL) {
    draw_block(mt,block,mt->end_secs,startx,width);
    block=block->next;
  }
  return FALSE;
}




static void track_select (lives_mt *mt) {
  int i;
  GtkWidget *labelbox,*eventbox,*checkbutton=NULL;
  gint hidden;

  for (i=0;i<mt->num_video_tracks;i++) {
    eventbox=g_list_nth_data(mt->video_draws,i);
    hidden=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"hidden"));
    if (!hidden) {
      labelbox=g_object_get_data(G_OBJECT(eventbox),"labelbox");
      if (i==mt->current_track) {
	gtk_widget_set_state(labelbox,GTK_STATE_SELECTED);
	gtk_widget_set_sensitive (mt->jumpto, g_object_get_data(G_OBJECT(eventbox),"blocks")!=NULL);
	checkbutton=g_object_get_data (G_OBJECT(eventbox), "checkbutton");
	if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton))) {
	  if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt->select_track))) gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->select_track),FALSE);
	  else on_seltrack_activate(GTK_MENU_ITEM(mt->select_track),mt);
	}
	else {
	  if (!gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(mt->select_track))) gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->select_track),TRUE);
	  else on_seltrack_activate(GTK_MENU_ITEM(mt->select_track),mt);
	}
      }
      else {
	gtk_widget_set_state(labelbox,GTK_STATE_NORMAL);
      }
    }
  }
}

static gboolean
track_ebox_pressed (GtkWidget *labelbox, GdkEventButton *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  mt->current_track=GPOINTER_TO_INT(g_object_get_data (G_OBJECT(labelbox),"layer_number"));
  track_select(mt);
  return FALSE;
}


void
scroll_tracks (lives_mt *mt) {
  int rows=0,idx=0;
  GList *vdraws=mt->video_draws;
  GList *table_children;
  gulong seltrack_func;
  GtkWidget *eventbox;
  GtkWidget *label;
  GtkWidget *checkbutton;
  GtkWidget *labelbox;
  GtkWidget *hbox;

  if (mt->timeline_table!=NULL) gtk_widget_destroy(mt->timeline_table);
  
  mt->timeline_table = gtk_table_new (mt->max_disp_vtracks, 40, TRUE);
  gtk_container_add (GTK_CONTAINER (mt->tl_eventbox), mt->timeline_table);

  gtk_table_set_row_spacings (GTK_TABLE(mt->timeline_table),5);
  gtk_table_set_col_spacings (GTK_TABLE(mt->timeline_table),0);

  while (vdraws!=NULL&&rows<mt->max_disp_vtracks) {
    eventbox=(GtkWidget *)vdraws->data;
    if (!GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"hidden"))) {
      label=(GTK_WIDGET(g_object_get_data(G_OBJECT(eventbox),"label")));
      checkbutton=(GTK_WIDGET(g_object_get_data(G_OBJECT(eventbox),"checkbutton")));
      labelbox=gtk_event_box_new();
      hbox=gtk_hbox_new(FALSE,10);

      if (palette->style&STYLE_1) {
	if (!palette->style&STYLE_3) {
	  gtk_widget_modify_bg (labelbox, GTK_STATE_SELECTED, &palette->info_base);
	}
	else {
	  gtk_widget_modify_bg (labelbox, GTK_STATE_SELECTED, &palette->menu_and_bars);
	}
	gtk_widget_modify_fg (label, GTK_STATE_SELECTED, &palette->info_text);
	gtk_widget_modify_fg (checkbutton, GTK_STATE_SELECTED, &palette->info_text);
      }

      g_object_set_data(G_OBJECT(labelbox),"layer_number",GINT_TO_POINTER(GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"layer_number"))));

      gtk_container_add (GTK_CONTAINER (labelbox), hbox);
      gtk_box_pack_start (GTK_BOX (hbox), checkbutton, FALSE, FALSE, 0);
      gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);

      gtk_table_attach (GTK_TABLE (mt->timeline_table), labelbox, 0, 6, rows, rows+1, GTK_FILL, 0, 0, 0);
      
      g_object_set_data (G_OBJECT(eventbox),"labelbox",labelbox);

      gtk_table_attach (GTK_TABLE (mt->timeline_table), eventbox, 6, 40, rows, rows+1,
			(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (GTK_FILL), 0, 0);

      seltrack_func=g_signal_connect_after (GTK_OBJECT (checkbutton), "toggled",
					    G_CALLBACK (on_seltrack_toggled),
					    mt);
      g_signal_connect (GTK_OBJECT (labelbox), "button_press_event",
			G_CALLBACK (track_ebox_pressed),
			(gpointer)mt);

      g_object_set_data(G_OBJECT(checkbutton),"tfunc",(gpointer)seltrack_func);

      g_signal_connect_after (GTK_OBJECT (eventbox), "expose_event",
			      G_CALLBACK (expose_vidtrack_event),
			      (gpointer)mt);

      g_signal_connect (GTK_OBJECT (eventbox), "button_press_event",
			G_CALLBACK (on_track_click),
			(gpointer)mt);
      g_signal_connect (GTK_OBJECT (eventbox), "button_release_event",
			G_CALLBACK (on_track_release),
			(gpointer)mt);

      if (rows==0) GTK_ADJUSTMENT(mt->vadjustment)->value=idx;
      rows++;
    }
    vdraws=vdraws->next;
    idx++;
  }

  
  table_children=GTK_TABLE(mt->timeline_table)->children;

  while (table_children!=NULL) {
    GtkRequisition req;
    GtkTableChild *child=table_children->data;
    req=child->widget->requisition;
    gtk_widget_set_size_request(child->widget,req.width,25);
    table_children=table_children->next;
  }

  gtk_widget_show_all(mt->timeline_table);
  gtk_widget_queue_draw (mt->vpaned);
}

void multitrack_view_clips (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  polymorph (mt,POLY_CLIPS);
}


void multitrack_view_in_out (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  polymorph (mt,POLY_IN_OUT);
}


static void time_to_string (lives_mt *mt, gdouble secs, gint length) {
  gint hours,mins,rest;
  gchar *string;

  hours=secs/3600;
  secs-=hours*3600.;
  mins=secs/60;
  secs-=mins*60.;
  rest=(secs-((gint)secs)*1.)*100.+.5;
  secs=(gint)secs*1.;
  string=g_strdup_printf("   %02d:%02d:%02d.%02d",hours,mins,(gint)secs,rest);
  gtk_entry_set_text (GTK_ENTRY (mt->timecode),string);
  g_free(string);
}


static void renumber_clips(void) {
  // remove gaps from closed files
  // we also ensure a unique_id to help with later id of the clips
  int cclip=1;
  int i;

  renumbered_clips[0]=0;

  for (i=1;i<=MAX_FILES&&mainw->files[i]!=NULL;i++) {
    while (mainw->files[i]->unique_id==0l) mainw->files[i]->unique_id=random();
    renumbered_clips[i]=i;
  }
  cclip=i;

  for (++i;i<=MAX_FILES;i++) {
    if (mainw->files[i]!=NULL) {
      while (mainw->files[i]->unique_id==0l) mainw->files[i]->unique_id=random();
      mainw->files[cclip]=mainw->files[i];
      mainw->files[i]=NULL;
      if (mainw->scrap_file==i) mainw->scrap_file=cclip;
      if (mainw->current_file==i) mainw->current_file=cclip;
      renumbered_clips[i]=cclip;
      mainw->first_free_file=cclip=i;
    }
    else renumbered_clips[i]=i;
  }
}


static void rerenumber_clips(gchar *lfile) {
  // we loaded an event_list, now we match clip numbers in event_list with our current clips
  // this gets used for translations in event_list_rectify/init_tracks, and then it can be discarded
  GList *lmap;
  int i,rnc;
  gchar **array;

  renumbered_clips[0]=0;
  
  for (i=1;i<=MAX_FILES&&mainw->files[i]!=NULL;i++) {
    renumbered_clips[i]=0;
    lmap=mainw->files[i]->layout_map;
    while (lmap!=NULL) {
      if (!strncmp(lmap->data,lfile,strlen(lfile))) {
	array=g_strsplit(lmap->data,"|",-1);
	rnc=atoi(array[1]);
	renumbered_clips[rnc]=i;
	lfps[i]=strtod(array[3],NULL);
	g_strfreev(array);
      }
      lmap=lmap->next;
    }
  }
}



void clip_select (lives_mt *mt) {
  GList *list=GTK_BOX (mt->clip_inner_box)->children;
  GtkBoxChild *clipbox=NULL;
  gint len;
  int i;
  mt->file_selected=-1;

  if (list==NULL) return;
  multitrack_view_clips(NULL,mt);

  if (mt->clip_selected>=(len=g_list_length(list))) mt->clip_selected=0;
  if (mt->clip_selected<0) mt->clip_selected=len-1;
  for (i=0;i<len;i++) {
    clipbox=g_list_nth_data (list,i);
    if (i==mt->clip_selected) {
      GtkAdjustment *adj;
      gint value=(adj=GTK_RANGE(GTK_SCROLLED_WINDOW(mt->clip_scroll)->hscrollbar)->adjustment)->upper*mt->clip_selected/len;
      gtk_adjustment_clamp_page(adj,value-adj->page_size/2,value+adj->page_size/2);
      gtk_widget_set_state(clipbox->widget,GTK_STATE_SELECTED);
      mt->file_selected=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(clipbox->widget),"file"));
      gtk_widget_set_sensitive (mt->insert, TRUE);
    }
    else gtk_widget_set_state(clipbox->widget,GTK_STATE_NORMAL);
  }
}


gboolean mt_prevclip (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  mt->clip_selected--;
  multitrack_view_clips(NULL,mt);
  clip_select(mt);
  return TRUE;
}

gboolean mt_nextclip (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  mt->clip_selected++;
  multitrack_view_clips(NULL,mt);
  clip_select(mt);
  return TRUE;
}

void set_timeline_end_secs (lives_mt *mt, gdouble secs) {
  mt->end_secs=secs;
  gtk_ruler_set_range (GTK_RULER (mt->timeline), 0., mt->end_secs, 0., mt->end_secs);
  gtk_widget_queue_draw (mt->timeline);
  gtk_widget_queue_draw (mt->timeline_table);
  gtk_spin_button_set_range (GTK_SPIN_BUTTON (mt->spinbutton_start),0.,mt->end_secs);
  gtk_spin_button_set_range (GTK_SPIN_BUTTON (mt->spinbutton_end),0.,mt->end_secs);

}

gboolean mt_tlfor (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;

  if (mainw->playing_file>-1) return TRUE;

  gdouble pos=GTK_RULER (mt->timeline)->position;

  pos+=(gint)mt->fps/mt->fps;

  if (pos>GTK_RULER (mt->timeline)->upper) pos=GTK_RULER (mt->timeline)->upper;
  GTK_RULER (mt->timeline)->position=pos;
  mt->region_start=mt->region_end=0.;
  gtk_widget_queue_draw (mt->timeline);
  mt->last_direction=DIRECTION_POSITIVE;

  return TRUE;
}


gboolean mt_tlback (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  gdouble pos=GTK_RULER (mt->timeline)->position;

  if (mainw->playing_file>-1) return TRUE;

  pos-=(gint)mt->fps/mt->fps;

  if (pos<0.) pos=0.;

  GTK_RULER (mt->timeline)->position=pos;
  mt->region_start=mt->region_end=0.;
  gtk_widget_queue_draw (mt->timeline);
  mt->last_direction=DIRECTION_NEGATIVE;

  return TRUE;
}



static void scroll_pre (gint track, lives_mt *mt) {
  gint scrolled;
  int num_vis=0;
  GList *liste=g_list_nth(mt->video_draws,track);
  GtkWidget *xeventbox;

  while (liste!=NULL) {
    xeventbox=(GTK_WIDGET(liste->data));
    scrolled=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(xeventbox),"hidden"));
    if (scrolled&TRACK_I_OUTSCROLLED_PRE) {
      scrolled^=TRACK_I_OUTSCROLLED_PRE;
      g_object_set_data(G_OBJECT(xeventbox),"hidden",GINT_TO_POINTER(scrolled));
    }
    if (num_vis>=mt->max_disp_vtracks) {
       if (scrolled&TRACK_I_OUTSCROLLED_POST) break;
       g_object_set_data(G_OBJECT(xeventbox),"hidden",GINT_TO_POINTER(scrolled|TRACK_I_OUTSCROLLED_POST));
    }
    else {
      if (scrolled&TRACK_I_OUTSCROLLED_POST) {
	scrolled^=TRACK_I_OUTSCROLLED_POST;
	g_object_set_data(G_OBJECT(xeventbox),"hidden",GINT_TO_POINTER(scrolled));
      }
      if (scrolled==0) num_vis++;
    }
    liste=liste->next;
  }
}


static void scroll_post (gint track, lives_mt *mt) {
  gint scrolled;
  int num_vis=0;
  GList *liste=g_list_nth(mt->video_draws,track);
  GtkWidget *xeventbox;

  while (liste!=NULL) {
    xeventbox=(GTK_WIDGET(liste->data));
    scrolled=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(xeventbox),"hidden"));
    if (scrolled&TRACK_I_OUTSCROLLED_POST) {
      scrolled^=TRACK_I_OUTSCROLLED_POST;
      g_object_set_data(G_OBJECT(xeventbox),"hidden",GINT_TO_POINTER(scrolled));
    }
    if (num_vis>=mt->max_disp_vtracks) {
      if (scrolled&TRACK_I_OUTSCROLLED_PRE) break;
      g_object_set_data(G_OBJECT(xeventbox),"hidden",GINT_TO_POINTER(scrolled|TRACK_I_OUTSCROLLED_PRE));
    }
    else {
      if (scrolled&TRACK_I_OUTSCROLLED_PRE) {
	scrolled^=TRACK_I_OUTSCROLLED_PRE;
	g_object_set_data(G_OBJECT(xeventbox),"hidden",GINT_TO_POINTER(scrolled));
      }
      if (scrolled==0) num_vis++;
    }
    liste=liste->prev;
  }
}



static void scroll_track_on_screen(gint track, lives_mt *mt) {
  GtkWidget *xeventbox=GTK_WIDGET(g_list_nth_data(mt->video_draws,track));
  gint scrolled=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(xeventbox),"hidden"));

  if (scrolled==0) {
    return;
  }
  else {
    if (scrolled==TRACK_I_OUTSCROLLED_PRE) scroll_pre(track,mt);
    if (scrolled==TRACK_I_OUTSCROLLED_POST) scroll_post(track,mt);
  }
  scroll_tracks(mt);
  return;
}


static void scroll_track_to_top(gint track, lives_mt *mt) {
  GList *liste=g_list_nth(mt->video_draws,track);
  GtkWidget *xeventbox=liste->data;
  gint scrolled=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(xeventbox),"hidden"));

  if (scrolled==0) {
    while (liste!=NULL) {
      xeventbox=GTK_WIDGET(liste->data);
      scrolled=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(xeventbox),"hidden"));
      if (scrolled&TRACK_I_OUTSCROLLED_PRE) break;
      scrolled|=TRACK_I_OUTSCROLLED_PRE;
      g_object_set_data(G_OBJECT(xeventbox),"hidden",GINT_TO_POINTER(scrolled|TRACK_I_OUTSCROLLED_PRE));
      liste=liste->prev;
    }
  }

  scroll_track_on_screen(track,mt);
  return;
}

void
scroll_track_by_scrollbar (GtkVScrollbar *sbar, gpointer user_data) {
  scroll_track_to_top(GTK_ADJUSTMENT(GTK_RANGE(sbar)->adjustment)->value,(lives_mt *)user_data);
}



gboolean mt_trdown (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;

  mt->current_track+=(mt->current_track<(mt->num_video_tracks-1));
  scroll_track_on_screen(mt->current_track,mt);
  track_select(mt);

  return TRUE;
}


gboolean mt_trup (GtkAccelGroup *group, GObject *obj, guint keyval, GdkModifierType mod, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;

  mt->current_track-=(mt->current_track>0);
  scroll_track_on_screen(mt->current_track,mt);
  track_select(mt);

  return TRUE;
}


void
paned_pos (GtkWidget *paned, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  gtk_widget_queue_draw (mt->timeline_table);
}






void mt_init_start_end_spins(lives_mt *mt) {
  GtkWidget *hbox;
  GtkWidget *arrow;
  GtkObject *spinbutton_start_adj;
  GtkObject *spinbutton_end_adj;
  
  hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (hbox);
  
  gtk_box_pack_start (GTK_BOX (mt->top_vbox), hbox, FALSE, FALSE, 0);

  spinbutton_start_adj = gtk_adjustment_new (0, 0, 0, 1, 100, 10);
  mt->spinbutton_start = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_start_adj), 1, 3);
  gtk_widget_set_sensitive(mt->spinbutton_start,FALSE);

  gtk_box_pack_start (GTK_BOX (hbox), mt->spinbutton_start, TRUE, FALSE, MAIN_SPIN_SPACER);
  GTK_WIDGET_SET_FLAGS (mt->spinbutton_start, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (mt->spinbutton_start), TRUE);
  gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (mt->spinbutton_start),GTK_UPDATE_ALWAYS);

  arrow = gtk_arrow_new (GTK_ARROW_LEFT, GTK_SHADOW_OUT);
  gtk_box_pack_start (GTK_BOX (hbox), arrow, FALSE, FALSE, 0);
  gtk_widget_modify_fg(arrow, GTK_STATE_NORMAL, &palette->normal_fore);

  gtk_entry_set_width_chars (GTK_ENTRY (mt->spinbutton_start),12);
  mt->sel_label = gtk_label_new(NULL);

  mt->preview_button=gtk_button_new_with_label(_("Show frame preview"));
  gtk_box_pack_start (GTK_BOX (hbox), mt->preview_button, FALSE, FALSE, 0);

  g_signal_connect (GTK_OBJECT (mt->preview_button), "clicked",
		    G_CALLBACK (on_frame_preview_clicked),
		    (gpointer)mt);

  set_sel_label(mt->sel_label);

  gtk_box_pack_start (GTK_BOX (hbox), mt->sel_label, FALSE, FALSE, 0);
  gtk_label_set_justify (GTK_LABEL (mt->sel_label), GTK_JUSTIFY_LEFT);
  gtk_widget_modify_fg(mt->sel_label, GTK_STATE_NORMAL, &palette->normal_fore);

  arrow = gtk_arrow_new (GTK_ARROW_RIGHT, GTK_SHADOW_OUT);
  gtk_box_pack_start (GTK_BOX (hbox), arrow, FALSE, FALSE, 3);
  gtk_widget_modify_fg(arrow, GTK_STATE_NORMAL, &palette->normal_fore);

  spinbutton_end_adj = gtk_adjustment_new (0, 0, 0, 1, 100, 10);
  mt->spinbutton_end = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_end_adj), 1, 3);
  gtk_widget_set_sensitive(mt->spinbutton_end,FALSE);
  gtk_box_pack_start (GTK_BOX (hbox), mt->spinbutton_end, TRUE, FALSE, MAIN_SPIN_SPACER);
  gtk_entry_set_width_chars (GTK_ENTRY (mt->spinbutton_end),12);

  GTK_WIDGET_SET_FLAGS (mt->spinbutton_end, GTK_CAN_DEFAULT|GTK_CAN_FOCUS);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (mt->spinbutton_end), TRUE);
  gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (mt->spinbutton_end),GTK_UPDATE_ALWAYS);

  if (palette->style&STYLE_1&&palette->style&STYLE_2) {
    gtk_widget_modify_base(mt->spinbutton_start, GTK_STATE_NORMAL, &palette->normal_back);
    gtk_widget_modify_base(mt->spinbutton_start, GTK_STATE_INSENSITIVE, &palette->normal_back);
    gtk_widget_modify_base(mt->spinbutton_end, GTK_STATE_NORMAL, &palette->normal_back);
    gtk_widget_modify_base(mt->spinbutton_end, GTK_STATE_INSENSITIVE, &palette->normal_back);
    gtk_widget_modify_text(mt->spinbutton_start, GTK_STATE_NORMAL, &palette->normal_fore);
    gtk_widget_modify_text(mt->spinbutton_start, GTK_STATE_INSENSITIVE, &palette->normal_fore);
    gtk_widget_modify_text(mt->spinbutton_end, GTK_STATE_NORMAL, &palette->normal_fore);
    gtk_widget_modify_text(mt->spinbutton_end, GTK_STATE_INSENSITIVE, &palette->normal_fore);
    gtk_widget_modify_fg(mt->sel_label, GTK_STATE_NORMAL, &palette->normal_fore);
  }
}



void mouse_mode_context(lives_mt *mt) {
  clear_context(mt);
  
  if (mt->mouse_mode==MOUSE_MODE_MOVE) {
    add_context_label (mt,(_("Single click on timeline to select a frame.\n")));
    add_context_label (mt,(_("Double click on timeline to select a block.\n")));
    add_context_label (mt,(_("Clips can be dragged onto the timeline.\n")));
    
    add_context_label (mt,(_("Mouse mode is: Move\n")));
    add_context_label (mt,(_("clips can be or moved around.\n")));
  }
  else if (mt->mouse_mode==MOUSE_MODE_SELECT) {
    clear_context(mt);
    
    add_context_label (mt,(_("Mouse mode is: Select.\n")));
    add_context_label (mt,(_("Drag with mouse on timeline\n")));
    add_context_label (mt,(_("to select tracks and time.\n")));
  }
}



static void on_snapt_mode_changed (GtkMenuItem *menuitem, gpointer user_data) {
 lives_mt *mt=(lives_mt *)user_data;

  if (menuitem==(GtkMenuItem *)mt->snapt_on) {
    set_menu_text(mt->snapt_menuitem,_("_Snap mode: On"),TRUE);
    mt->snapt=TRUE;
  }
  else if (menuitem==(GtkMenuItem *)mt->snapt_off) {
    set_menu_text(mt->snapt_menuitem,_("_Snap mode: Off"),TRUE);
    mt->snapt=FALSE;
  }

  if (!mt->snapt) {
    g_signal_handler_block(mt->snapt_on,mt->snapt_on_func);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->snapt_on),FALSE);
    g_signal_handler_unblock(mt->snapt_on,mt->snapt_on_func);
  }
  else {
    g_signal_handler_block(mt->snapt_off,mt->snapt_off_func);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->snapt_off),FALSE);
    g_signal_handler_unblock(mt->snapt_off,mt->snapt_off_func);
  }

}



static void on_mouse_mode_changed (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;

  if (menuitem==(GtkMenuItem *)mt->mm_move) {
    set_menu_text(mt->mm_menuitem,_("_Mouse mode: Move"),TRUE);
    mt->mouse_mode=MOUSE_MODE_MOVE;
  }
  else if (menuitem==(GtkMenuItem *)mt->mm_select) {
    set_menu_text(mt->mm_menuitem,_("_Mouse mode: Select"),TRUE);
    mt->mouse_mode=MOUSE_MODE_SELECT;
  }

  mouse_mode_context(mt);

  if (mt->mouse_mode!=MOUSE_MODE_MOVE) {
    g_signal_handler_block(mt->mm_move,mt->mm_move_func);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->mm_move),FALSE);
    g_signal_handler_unblock(mt->mm_move,mt->mm_move_func);
  }
  if (mt->mouse_mode!=MOUSE_MODE_SELECT) {
    g_signal_handler_block(mt->mm_select,mt->mm_select_func);
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->mm_select),FALSE);
    g_signal_handler_unblock(mt->mm_select,mt->mm_select_func);
  }

}


static size_t estimate_space(lives_mt *mt, gint undo_type) {
  size_t needed=sizeof(mt_undo);

  switch (undo_type) {
  case MT_UNDO_REMOVE_GAPS:
  case MT_UNDO_MOVE_BLOCK:
  case MT_UNDO_DELETE_BLOCK:
  case MT_UNDO_SPLIT_MULTI:
  case MT_UNDO_SPLIT:
  case MT_UNDO_APPLY_FILTER:
  case MT_UNDO_INSERT_BLOCK:
    needed+=event_list_get_byte_size(mt->event_list,NULL);
    break;
  }
  return needed;
}

static gchar *get_undo_text(gint action, void *extra) {
  gchar *filtname,*ret;
  int error;

  switch (action) {
  case MT_UNDO_REMOVE_GAPS:
    return g_strdup(_("Close gaps"));
  case MT_UNDO_MOVE_BLOCK:
    return g_strdup(_("Move block"));
  case MT_UNDO_DELETE_BLOCK:
    return g_strdup(_("Delete block"));
  case MT_UNDO_SPLIT_MULTI:
    return g_strdup(_("Split tracks"));
  case MT_UNDO_SPLIT:
    return g_strdup(_("Split block"));
  case MT_UNDO_APPLY_FILTER:
    filtname=weed_get_string_value((weed_plant_t *)extra,"name",&error);
    ret=g_strdup_printf(_("Apply %s"),filtname);
    g_free(filtname);
    return ret;
  case MT_UNDO_INSERT_BLOCK:
    return g_strdup(_("Insert block"));
  }
  return g_strdup("");
}





static void mt_set_undoable (lives_mt *mt, gint what, void *extra, gboolean sensitive) {
  mt->undoable=sensitive;
  if (what!=MT_UNDO_NONE) {
    gchar *what_safe;
    gchar *text=get_undo_text(what,extra);
    what_safe=g_strdelimit (g_strdup (text),"_",' ');
    g_snprintf(mt->undo_text,32,_ ("_Undo %s"),what_safe);
    g_free (what_safe);
    g_free(text);
  }
  else {
    mt->undoable=FALSE;
    g_snprintf(mt->undo_text,32,_ ("_Undo"));
  }
  set_menu_text(mt->undo,mt->undo_text,TRUE);

  gtk_widget_set_sensitive (mt->undo,sensitive);

}


static void mt_set_redoable (lives_mt *mt, gint what, void *extra, gboolean sensitive) {
  mt->redoable=sensitive;
  if (what!=MT_UNDO_NONE) {
    gchar *what_safe;
    gchar *text=get_undo_text(what,extra);
    what_safe=g_strdelimit (g_strdup (text),"_",' ');
    g_snprintf(mt->redo_text,32,_ ("_Redo %s"),what_safe);
    g_free (what_safe);
    g_free(text);
  }
  else {
    mt->redoable=FALSE;
    g_snprintf(mt->redo_text,32,_ ("_Redo"));
  }
  set_menu_text(mt->redo,mt->redo_text,TRUE);

  gtk_widget_set_sensitive (mt->redo,sensitive);

}



gboolean make_backup_space (lives_mt *mt, size_t space_needed) {
  // read thru mt->undos and eliminate that space until we have space_needed
  size_t space_avail=(size_t)(prefs->mt_undo_buf*1024*1024)-mt->undo_buffer_used;
  size_t space_freed=0;
  GList *xundo=mt->undos,*ulist;
  int count=0;
  mt_undo *undo;

  while (xundo!=NULL) {
    count++;
    undo=(mt_undo *)(xundo->data);
    space_freed+=undo->data_len;
    if ((space_avail+space_freed)>=space_needed) {
      memmove(mt->undo_mem,mt->undo_mem+space_freed,mt->undo_buffer_used-space_freed);
      ulist=g_list_copy(g_list_nth(mt->undos,count));
      ulist->prev=NULL;
      g_list_free(mt->undos);
      mt->undos=ulist;
      while (ulist!=NULL) {
	ulist->data=(unsigned char *)(ulist->data)-space_freed;
	ulist=ulist->next;
      }
      mt->undo_buffer_used-=space_freed;
      if (mt->undo_offset>g_list_length(mt->undos)) {
	mt->undo_offset=g_list_length(mt->undos);
	mt_set_undoable(mt,MT_UNDO_NONE,NULL,FALSE);
	mt_set_redoable(mt,((mt_undo *)(mt->undos->data))->action,NULL,TRUE);
      }
      return TRUE;
    }
    xundo=xundo->next;
  }
  mt->undo_buffer_used=0;
  g_list_free(mt->undos);
  mt->undos=NULL;
  mt->undo_offset=0;
  mt_set_undoable(mt,MT_UNDO_NONE,NULL,FALSE);
  mt_set_redoable(mt,MT_UNDO_NONE,NULL,FALSE);
  return FALSE;
}

static void save_event_list_inner(int fd, weed_plant_t *event_list, unsigned char **mem) {
  weed_plant_t *event=get_first_event(event_list);

  weed_plant_serialise(fd,event_list,mem);
  while (event!=NULL) {
    if (WEED_EVENT_IS_FILTER_INIT(event)) weed_set_voidptr_value(event,"event_id",(void *)event);
    weed_plant_serialise(fd,event,mem);
    event=get_next_event(event);
  }
}




void mt_backup(lives_mt *mt, gint undo_type, gpointer data) {
  size_t space_needed=0;
  mt_undo *undo;
  unsigned char *memblock;

  mt_undo *last_valid_undo;

  mt->did_backup=TRUE;

  mt->changed=TRUE;

  if (mt->undo_mem==NULL) return;

  if (mt->undos!=NULL&&mt->undo_offset!=0) {
    // invalidate redo's - we are backing up, so we can't redo any more
    // invalidate from g_list_length-undo_offset onwards
    if ((g_list_length(mt->undos))==mt->undo_offset) {
      mt->undos=NULL;
      mt->undo_buffer_used=0;
    }
    else {
      int i=0;
      GList *ulist=mt->undos;
      while (i<((int)g_list_length(mt->undos)-1-mt->undo_offset)) {
	ulist=ulist->next;
	i++;
      }
      if (ulist!=NULL) {
	memblock=ulist->data;
	last_valid_undo=(mt_undo *)memblock;
	memblock+=last_valid_undo->data_len;
	mt->undo_buffer_used=memblock-mt->undo_mem;
	if (ulist->next!=NULL) {
	  ulist=ulist->next;
	  ulist->prev->next=NULL;
	  g_list_free(ulist);
	}
      }
    }
    mt_set_redoable(mt,MT_UNDO_NONE,NULL,FALSE);
    mt->undo_offset=0;
  }

  undo=g_malloc(sizeof(mt_undo));
  undo->action=undo_type;
  undo->extra=NULL;

  switch (undo_type) {
  case MT_UNDO_APPLY_FILTER:
    undo->extra=get_weed_filter(mt->current_fx);
  case MT_UNDO_REMOVE_GAPS:
  case MT_UNDO_MOVE_BLOCK:
  case MT_UNDO_DELETE_BLOCK:
  case MT_UNDO_SPLIT_MULTI:
  case MT_UNDO_SPLIT:
  case MT_UNDO_INSERT_BLOCK:
    add_markers(mt,mt->event_list);
    if ((space_needed=estimate_space(mt,undo_type))>((size_t)(prefs->mt_undo_buf*1024*1024)-mt->undo_buffer_used)) {
      if (!make_backup_space(mt,space_needed)) {
	remove_markers(mt->event_list);
	return;
      }
      memblock=(unsigned char *)(mt->undo_mem+mt->undo_buffer_used+sizeof(mt_undo));
    }
    undo->data_len=space_needed;
    memblock=(unsigned char *)(mt->undo_mem+mt->undo_buffer_used+sizeof(mt_undo));
    save_event_list_inner(0,mt->event_list,&memblock);
    remove_markers(mt->event_list);
    break;
  }
  memcpy(mt->undo_mem+mt->undo_buffer_used,undo,sizeof(mt_undo));
  mt->undos=g_list_append(mt->undos,mt->undo_mem+mt->undo_buffer_used);
  mt->undo_buffer_used+=space_needed;
  mt_set_undoable(mt,undo->action,undo->extra,TRUE);
  g_free(undo);
}






lives_mt *multitrack (weed_plant_t *event_list, gint orig_file, gdouble fps) {
  GtkWidget *hseparator;
  GtkWidget *vseparator;
  GtkWidget *menubar;
  GtkWidget *menuitem;
  GtkWidget *menuitemsep;
  GtkWidget *menuitem_menu;
  GtkWidget *image;
  GtkWidget *separator;
  GtkWidget *playsel;
  GtkWidget *rewind;
  GtkWidget *full_screen;
  GtkWidget *sepwin;
  GtkWidget *fade;
  GtkWidget *loop_video;
  GtkWidget *loop_continue;
  GtkWidget *mute_audio;
  GtkWidget *sticky;
  GtkWidget *quit;
  GtkWidget *add_vid_behind;
  GtkWidget *add_vid_front;
  GtkWidget *view_mt_details;
  GtkWidget *show_messages;
  GtkWidget *tl_vbox;
  GtkWidget *scrollbar;
  GtkWidget *hbox;
  GtkWidget *vbox;
  GtkWidget *eventbox;
  GtkWidget *label;
  GtkWidget *submenu;
  GtkWidget *time_snap;
  GtkWidget *submenu_menu;
  GtkWidget *show_frame_events;
  GtkObject *vadjustment;
  GtkObject *spinbutton_adj;
  gint num_filters;
  int i;

  lives_mt *mt=(lives_mt *)g_malloc(sizeof(lives_mt));

  if (event_list==NULL) {
    mt->undo_mem=g_try_malloc(prefs->mt_undo_buf*1024*1024);
    if (mt->undo_mem==NULL) {
      do_mt_undo_mem_error();
    }
    mt->undo_buffer_used=0;
    mt->undos=NULL;
    mt->undo_offset=0;
    mt->changed=FALSE;
  }

  mt->tl_fixed_length=0.;
  mt->ptr_time=0.;
  mt->video_draws=NULL;
  mt->block_selected=NULL;
  mt->event_list=event_list;
  mt->accel_group = GTK_ACCEL_GROUP(gtk_accel_group_new ());
  mt->fps=fps;
  mt->hotspot_x=mt->hotspot_y=0;

  mt->region_start=mt->region_end=0.;
  mt->snap_over=FALSE;
  mt->region_updating=FALSE;
  mt->is_rendering=FALSE;
  mt->selected_tracks=NULL;
  mt->max_disp_vtracks=MAX_DISP_VTRACKS;
  mt->mt_frame_preview=FALSE;
  mt->current_rfx=NULL;
  mt->putative_block=NULL;

  mt->display=gdk_display_get_default(); // TODO

  mt->move_effects=TRUE;
  mt->moving_block=FALSE;

  mt->mouse_mode=MOUSE_MODE_MOVE;
  mt->snapt=FALSE;
  mt->insert_start=mt->insert_end=-1;

  memset(mt->layout_name,0,1);

  mt->did_backup=FALSE;

  mt->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  if (palette->style&STYLE_1) {
    if (palette->style&STYLE_3) gtk_widget_modify_bg(mt->window, GTK_STATE_NORMAL, &palette->normal_back);
    else gtk_widget_modify_bg(mt->window, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  gtk_window_set_title (GTK_WINDOW (mt->window), g_strdup_printf(_("LiVES-%s: Multitrack Window"),LiVES_VERSION));

  mt->top_vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (mt->window), mt->top_vbox);

  menubar = gtk_menu_bar_new ();
  gtk_box_pack_start (GTK_BOX (mt->top_vbox), menubar, FALSE, FALSE, 0);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menubar, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  // File

  menuitem = gtk_menu_item_new_with_mnemonic (_ ("_File"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);
  
  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  mt->save_event_list = gtk_image_menu_item_new_with_mnemonic (_("_Save layout"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->save_event_list);
  gtk_widget_set_sensitive (mt->save_event_list, FALSE);

  mt->load_event_list = gtk_image_menu_item_new_with_mnemonic (_("_Load layout"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->load_event_list);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  quit = gtk_image_menu_item_new_from_stock ("gtk-quit", mt->accel_group);
  gtk_container_add (GTK_CONTAINER (menuitem_menu), quit);

  // Edit

  menuitem = gtk_menu_item_new_with_mnemonic (_("_Edit"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);

  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  mt->undo = gtk_image_menu_item_new_with_mnemonic (_("_Undo"));
  gtk_widget_show (mt->undo);
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->undo);
  gtk_widget_set_sensitive (mt->undo, FALSE);

  gtk_widget_add_accelerator (mt->undo, "activate", mt->accel_group,
                              GDK_u, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  image = gtk_image_new_from_stock ("gtk-undo", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mt->undo), image);

  mt_set_undoable(mt,MT_UNDO_NONE,NULL,FALSE);

  g_signal_connect (GTK_OBJECT (mt->undo), "activate",
		    G_CALLBACK (multitrack_undo),
		    (gpointer)mt);

  mt->redo = gtk_image_menu_item_new_with_mnemonic (_("_Redo"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->redo);
  gtk_widget_set_sensitive (mt->redo, FALSE);

  gtk_widget_add_accelerator (mt->redo, "activate", mt->accel_group,
                              GDK_z, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  image = gtk_image_new_from_stock ("gtk-redo", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mt->redo), image);

  mt_set_redoable(mt,MT_UNDO_NONE,NULL,FALSE);

  g_signal_connect (GTK_OBJECT (mt->redo), "activate",
		    G_CALLBACK (multitrack_redo),
		    (gpointer)mt);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);


  mt->clipedit = gtk_image_menu_item_new_with_mnemonic (_("_CLIP EDITOR"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->clipedit);

  gtk_widget_add_accelerator (mt->clipedit, "activate", mt->accel_group,
                              GDK_e, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);


  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  mt->insert = gtk_image_menu_item_new_with_mnemonic (_("_Insert selected clip"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->insert);

  gtk_widget_add_accelerator (mt->insert, "activate", mt->accel_group,
                              GDK_i, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);
  gtk_widget_set_sensitive (mt->insert, FALSE);


  mt->delblock = gtk_image_menu_item_new_with_mnemonic (_("_Delete selected block"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->delblock);
  gtk_widget_set_sensitive (mt->delblock, FALSE);

  gtk_widget_add_accelerator (mt->delblock, "activate", mt->accel_group,
                              GDK_d, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);


  mt->jumpto = gtk_image_menu_item_new_with_mnemonic (_("_Jump to next block boundary"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->jumpto);

  gtk_widget_add_accelerator (mt->jumpto, "activate", mt->accel_group,
                              GDK_j, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  gtk_widget_set_sensitive (mt->jumpto, FALSE);

  // Play

  menuitem = gtk_menu_item_new_with_mnemonic (_ ("_Play"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);
  
  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  mt->playall = gtk_image_menu_item_new_with_mnemonic (_("_Play All"));
  gtk_widget_add_accelerator (mt->playall, "activate", mt->accel_group,
                              GDK_p, 0,
                              GTK_ACCEL_VISIBLE);
  gtk_widget_set_sensitive (mt->playall, FALSE);


  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->playall);

  image = gtk_image_new_from_stock ("gtk-refresh", GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mt->playall), image);

  playsel = gtk_image_menu_item_new_with_mnemonic (_("Pla_y Selection"));
  gtk_widget_add_accelerator (playsel, "activate", mt->accel_group,
                              GDK_y, 0,
                              GTK_ACCEL_VISIBLE);
  gtk_container_add (GTK_CONTAINER (menuitem_menu), playsel);
  gtk_widget_set_sensitive (playsel, FALSE);

  mt->stop = gtk_image_menu_item_new_with_mnemonic (_("_Stop"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->stop);
  gtk_widget_set_sensitive (mt->stop, FALSE);
  gtk_widget_add_accelerator (mt->stop, "activate", mt->accel_group,
                              GDK_q, 0,
                              GTK_ACCEL_VISIBLE);

  image = gtk_image_new_from_stock ("gtk-stop", GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mt->stop), image);

  rewind = gtk_image_menu_item_new_with_mnemonic (_("Re_wind"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), rewind);

  image = gtk_image_new_from_stock ("gtk-back", GTK_ICON_SIZE_MENU);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (rewind), image);
  gtk_widget_set_sensitive (rewind, FALSE);

  gtk_widget_add_accelerator (rewind, "activate", mt->accel_group,
                              GDK_w, 0,
                              GTK_ACCEL_VISIBLE);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  full_screen = gtk_check_menu_item_new_with_mnemonic (_("_Full Screen"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), full_screen);

  gtk_widget_add_accelerator (full_screen, "activate", mt->accel_group,
                              GDK_f, 0,
                              GTK_ACCEL_VISIBLE);

  sepwin = gtk_check_menu_item_new_with_mnemonic (_("Play in _Separate Window"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), sepwin);

  gtk_widget_add_accelerator (sepwin, "activate", mt->accel_group,
                              GDK_s, 0,
                              GTK_ACCEL_VISIBLE);

  fade = gtk_check_menu_item_new_with_mnemonic (_("_Blank Background"));
  gtk_widget_add_accelerator (fade, "activate", mt->accel_group,
                              GDK_b, 0,
                              GTK_ACCEL_VISIBLE);

  gtk_container_add (GTK_CONTAINER (menuitem_menu), fade);
  gtk_widget_set_sensitive (fade, FALSE);

  loop_video = gtk_check_menu_item_new_with_mnemonic (_("(Auto)_loop Video (to fit audio track)"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), loop_video);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(loop_video),FALSE);
  gtk_widget_add_accelerator (loop_video, "activate", mt->accel_group,
                              GDK_l, 0,
                              GTK_ACCEL_VISIBLE);
  gtk_widget_set_sensitive (loop_video, FALSE);

  loop_continue = gtk_check_menu_item_new_with_mnemonic (_("L_oop Continuously"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), loop_continue);

  gtk_widget_add_accelerator (loop_continue, "activate", mt->accel_group,
                              GDK_o, 0,
                              GTK_ACCEL_VISIBLE);

  mute_audio = gtk_check_menu_item_new_with_mnemonic (_("_Mute"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mute_audio);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  sticky = gtk_check_menu_item_new_with_mnemonic (_("Separate Window 'S_ticky' Mode"));

  gtk_widget_add_accelerator (sticky, "activate", mt->accel_group,
                              GDK_t, 0,
                              GTK_ACCEL_VISIBLE);

  gtk_container_add (GTK_CONTAINER (menuitem_menu), sticky);

  if (capable->smog_version_correct&&prefs->sepwin_type==1) {
    gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(sticky),TRUE);
  }



  // Effects

  menuitem = gtk_menu_item_new_with_mnemonic (_ ("Effect_s"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);
  
  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  mt->move_fx = gtk_check_menu_item_new_with_mnemonic (_("Move effects with blocks"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->move_fx);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->move_fx),mt->move_effects);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  mt->fx_block = gtk_menu_item_new_with_mnemonic (_("Apply effect to _block..."));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->fx_block);
  gtk_widget_set_sensitive(mt->fx_block,FALSE);

  submenu_menu=gtk_menu_new();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (mt->fx_block), submenu_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(submenu_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }


  // all filter effects to submenu
  num_filters=rte_get_numfilters();
  for (i=0;i<num_filters;i++) {
    weed_plant_t *filter=get_weed_filter(i);
    if (filter!=NULL&&enabled_in_channels(filter)==1&&enabled_out_channels(filter)==1) {
      gchar *fname=weed_filter_get_name(i);
      GtkWidget *menuitem = gtk_image_menu_item_new_with_label (fname);
      gtk_container_add (GTK_CONTAINER (submenu_menu), menuitem);
      g_free(fname);
      g_object_set_data(G_OBJECT(menuitem),"idx",GINT_TO_POINTER(i));
      g_signal_connect (GTK_OBJECT (menuitem), "activate",
			G_CALLBACK (mt_add_block_effect),
			(gpointer)mt);
    }
  }


  mt->fx_region = gtk_menu_item_new_with_mnemonic (_("Apply effect to _region..."));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->fx_region);
  gtk_widget_set_sensitive(mt->fx_region,FALSE);

  submenu_menu=gtk_menu_new();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (mt->fx_region), submenu_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(submenu_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }




  // Tracks

  menuitem = gtk_menu_item_new_with_mnemonic (_ ("_Tracks"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);
  
  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  add_vid_behind = gtk_image_menu_item_new_with_mnemonic (_("Add Video Track at _Rear"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), add_vid_behind);

  gtk_widget_add_accelerator (add_vid_behind, "activate", mt->accel_group,
                              GDK_t, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);


  add_vid_front = gtk_image_menu_item_new_with_mnemonic (_("Add Video Track at _Front"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), add_vid_front);

  gtk_widget_add_accelerator (add_vid_front, "activate", mt->accel_group,
                              GDK_t, GDK_CONTROL_MASK|GDK_SHIFT_MASK,
                              GTK_ACCEL_VISIBLE);


  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);


  menuitem = gtk_menu_item_new_with_mnemonic (_("_Split selected tracks"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), menuitem);

  g_signal_connect (GTK_OBJECT (menuitem), "activate",
		    G_CALLBACK (on_split_sel_activate),
		    (gpointer)mt);

  // Selection

  menuitem = gtk_menu_item_new_with_mnemonic (_ ("Se_lection"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);

  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  time_snap = gtk_check_menu_item_new_with_mnemonic (_("_Snap to overlap"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), time_snap);

  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(time_snap),mt->snap_over);

  g_signal_connect_after (GTK_OBJECT (time_snap), "toggled",
			  G_CALLBACK (on_snap_over_changed),
			  (gpointer)mt);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  mt->select_track = gtk_check_menu_item_new_with_mnemonic (_("_Select Current Track"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->select_track);

  gtk_widget_add_accelerator (mt->select_track, "activate", mt->accel_group,
                              GDK_space, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  menuitem = gtk_menu_item_new_with_mnemonic (_("Select _all video tracks"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), menuitem);

  g_signal_connect (GTK_OBJECT (menuitem), "activate",
		    G_CALLBACK (select_all_vid),
		    (gpointer)mt);

  menuitem = gtk_menu_item_new_with_mnemonic (_("Select _no video tracks"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), menuitem);

  g_signal_connect (GTK_OBJECT (menuitem), "activate",
		    G_CALLBACK (select_no_vid),
		    (gpointer)mt);

  menuitem = gtk_menu_item_new_with_mnemonic (_("Select all _time"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), menuitem);

  g_signal_connect (GTK_OBJECT (menuitem), "activate",
		    G_CALLBACK (select_all_time),
		    (gpointer)mt);

  gtk_widget_add_accelerator (menuitem, "activate", mt->accel_group,
                              GDK_a, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  menuitem = gtk_menu_item_new_with_mnemonic (_("Select from _zero time"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), menuitem);

  g_signal_connect (GTK_OBJECT (menuitem), "activate",
		    G_CALLBACK (select_from_zero_time),
		    (gpointer)mt);

  menuitem = gtk_menu_item_new_with_mnemonic (_("Select to _end time"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), menuitem);

  g_signal_connect (GTK_OBJECT (menuitem), "activate",
		    G_CALLBACK (select_to_end_time),
		    (gpointer)mt);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  mt->remove_gaps = gtk_menu_item_new_with_mnemonic (_("Close all _gaps in selection"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->remove_gaps);

  g_signal_connect (GTK_OBJECT (mt->remove_gaps), "activate",
		    G_CALLBACK (remove_gaps),
		    (gpointer)mt);

  gtk_widget_add_accelerator (mt->remove_gaps, "activate", mt->accel_group,
                              GDK_g, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  mt->remove_first_gaps = gtk_menu_item_new_with_mnemonic (_("Close _first gap(s) in selection"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->remove_first_gaps);

  g_signal_connect (GTK_OBJECT (mt->remove_first_gaps), "activate",
		    G_CALLBACK (remove_first_gaps),
		    (gpointer)mt);

  gtk_widget_add_accelerator (mt->remove_first_gaps, "activate", mt->accel_group,
                              GDK_f, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  // Tools

  menuitem = gtk_menu_item_new_with_mnemonic (_ ("_Tools"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);
  
  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  menuitem = gtk_image_menu_item_new_with_mnemonic (_("_Preferences..."));
  gtk_widget_show (menuitem);
  gtk_container_add (GTK_CONTAINER (menuitem_menu), menuitem);
  gtk_widget_add_accelerator (menuitem, "activate", mt->accel_group,
                              GDK_p, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  image = gtk_image_new_from_stock ("gtk-preferences", GTK_ICON_SIZE_MENU);
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (menuitem), image);

  g_signal_connect (GTK_OBJECT (menuitem), "activate",
                      G_CALLBACK (on_preferences_activate),
                      NULL);



  // Render

  menuitem = gtk_menu_item_new_with_mnemonic (_ ("_Render"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);
  
  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  mt->render = gtk_image_menu_item_new_with_mnemonic (_("_Render all to new clip..."));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->render);

  gtk_widget_add_accelerator (mt->render, "activate", mt->accel_group,
                              GDK_r, GDK_CONTROL_MASK,
                              GTK_ACCEL_VISIBLE);

  // View

  menuitem = gtk_menu_item_new_with_mnemonic (_ ("_View"));
  gtk_container_add (GTK_CONTAINER (menubar), menuitem);
  
  menuitem_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), menuitem_menu);

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menuitem_menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }


  mt->view_clips = gtk_menu_item_new_with_mnemonic (_("_Clips"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->view_clips);

  gtk_widget_add_accelerator (mt->view_clips, "activate", mt->accel_group,
                              GDK_c, 0,
                              GTK_ACCEL_VISIBLE);

  mt->view_in_out = gtk_menu_item_new_with_mnemonic (_("_In/out points"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->view_in_out);

  gtk_widget_add_accelerator (mt->view_in_out, "activate", mt->accel_group,
                              GDK_i, 0,
                              GTK_ACCEL_VISIBLE);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  view_mt_details = gtk_menu_item_new_with_mnemonic (_("Multitrack _details"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), view_mt_details);

  show_messages = gtk_image_menu_item_new_with_mnemonic (_("Show _Messages"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), show_messages);

  separator = gtk_menu_item_new ();
  gtk_container_add (GTK_CONTAINER (menuitem_menu), separator);
  gtk_widget_set_sensitive (separator, FALSE);

  mt->view_events = gtk_image_menu_item_new_with_mnemonic (_("_Event Window"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), mt->view_events);
  gtk_widget_set_sensitive (mt->view_events, FALSE);

  show_frame_events = gtk_check_menu_item_new_with_mnemonic (_("_Show FRAME events"));
  gtk_container_add (GTK_CONTAINER (menuitem_menu), show_frame_events);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(show_frame_events),prefs->event_window_show_frame_events);

  menuitemsep = gtk_menu_item_new_with_label ("  |  ");
  gtk_container_add (GTK_CONTAINER(menubar), menuitemsep);
  gtk_widget_set_sensitive (menuitemsep,FALSE);

  mt->mm_menuitem = gtk_menu_item_new_with_label ("");
  gtk_container_add (GTK_CONTAINER(menubar), mt->mm_menuitem);

  submenu = gtk_menu_new ();
  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(submenu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (mt->mm_menuitem), submenu);

  mt->mm_move = gtk_check_menu_item_new_with_mnemonic (_("Mouse mode: _Move"));
  gtk_container_add (GTK_CONTAINER(submenu), mt->mm_move);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->mm_move),mt->mouse_mode==MOUSE_MODE_MOVE);

  mt->mm_move_func=g_signal_connect (GTK_OBJECT (mt->mm_move), "toggled",
				     G_CALLBACK (on_mouse_mode_changed),
				     (gpointer)mt);

  mt->mm_select = gtk_check_menu_item_new_with_mnemonic (_("Mouse mode: _Select"));
  gtk_container_add (GTK_CONTAINER(submenu), mt->mm_select);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->mm_select),mt->mouse_mode==MOUSE_MODE_SELECT);

  mt->mm_select_func=g_signal_connect (GTK_OBJECT (mt->mm_select), "toggled",
				       G_CALLBACK (on_mouse_mode_changed),
				       (gpointer)mt);

  menuitemsep = gtk_menu_item_new_with_label ("  |  ");
  gtk_container_add (GTK_CONTAINER(menubar), menuitemsep);
  gtk_widget_set_sensitive (menuitemsep,FALSE);

  mt->snapt_menuitem = gtk_menu_item_new_with_label ("");
  gtk_container_add (GTK_CONTAINER(menubar), mt->snapt_menuitem);

  submenu = gtk_menu_new ();
  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(submenu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (mt->snapt_menuitem), submenu);

  mt->snapt_off = gtk_check_menu_item_new_with_mnemonic (_("Snap Mode: O_ff"));
  gtk_container_add (GTK_CONTAINER(submenu), mt->snapt_off);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->snapt_off),!mt->snapt);

  mt->snapt_off_func=g_signal_connect (GTK_OBJECT (mt->snapt_off), "toggled",
				       G_CALLBACK (on_snapt_mode_changed),
				       (gpointer)mt);

  mt->snapt_on = gtk_check_menu_item_new_with_mnemonic (_("Snap mode: O_n"));
  gtk_container_add (GTK_CONTAINER(submenu), mt->snapt_on);
  gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(mt->snapt_on),mt->snapt);

  mt->snapt_on_func=g_signal_connect (GTK_OBJECT (mt->snapt_on), "toggled",
				      G_CALLBACK (on_snapt_mode_changed),
				      (gpointer)mt);

  menuitemsep = gtk_menu_item_new_with_label ("  |  ");
  gtk_container_add (GTK_CONTAINER(menubar), menuitemsep);
  gtk_widget_set_sensitive (menuitemsep,FALSE);

  gtk_widget_ref(mainw->m_sepwinbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_sepwinbutton->parent),mainw->m_sepwinbutton);
  gtk_container_add (GTK_CONTAINER(menubar), mainw->m_sepwinbutton);
  gtk_widget_unref(mainw->m_sepwinbutton);

  gtk_widget_ref(mainw->m_rewindbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_rewindbutton->parent),mainw->m_rewindbutton);

  gtk_widget_ref(mainw->m_playbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_playbutton->parent),mainw->m_playbutton);
  gtk_container_add (GTK_CONTAINER(menubar), mainw->m_playbutton);
  gtk_widget_unref(mainw->m_playbutton);

  gtk_widget_ref(mainw->m_stopbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_stopbutton->parent),mainw->m_stopbutton);
  gtk_container_add (GTK_CONTAINER(menubar), mainw->m_stopbutton);
  gtk_widget_unref(mainw->m_stopbutton);

  gtk_widget_ref(mainw->m_playselbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_playselbutton->parent),mainw->m_playselbutton);

  gtk_widget_ref(mainw->m_loopbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_loopbutton->parent),mainw->m_loopbutton);
  gtk_container_add (GTK_CONTAINER(menubar), mainw->m_loopbutton);
  gtk_widget_unref(mainw->m_loopbutton);

  gtk_widget_ref(mainw->m_mutebutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_mutebutton->parent),mainw->m_mutebutton);


  g_signal_connect (GTK_OBJECT (quit), "activate",
		    G_CALLBACK (on_quit_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (mt->clipedit), "activate",
		    G_CALLBACK (multitrack_end_cb),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->playall), "activate",
		    G_CALLBACK (multitrack_playall),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->insert), "activate",
		    G_CALLBACK (multitrack_insert),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->view_events), "activate",
		    G_CALLBACK (multitrack_view_events),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (view_mt_details), "activate",
		    G_CALLBACK (multitrack_view_details),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->view_clips), "activate",
		    G_CALLBACK (multitrack_view_clips),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->view_in_out), "activate",
		    G_CALLBACK (multitrack_view_in_out),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (show_messages), "activate",
		    G_CALLBACK (on_show_messages_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (mt->stop), "activate",
		    G_CALLBACK (on_stop_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (sepwin), "activate",
		    G_CALLBACK (on_sepwin_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (full_screen), "activate",
		    G_CALLBACK (on_full_screen_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (fade), "activate",
		    G_CALLBACK (on_fade_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (loop_video), "activate",
		    G_CALLBACK (on_loop_video_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (loop_continue), "activate",
		    G_CALLBACK (on_loop_cont_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (mute_audio), "activate",
		    G_CALLBACK (on_mute_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (sticky), "activate",
		    G_CALLBACK (on_sticky_activate),
		    NULL);
  g_signal_connect (GTK_OBJECT (add_vid_behind), "activate",
		    G_CALLBACK (add_video_track_behind),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (add_vid_front), "activate",
		    G_CALLBACK (add_video_track_front),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->render), "activate",
		    G_CALLBACK (on_render_activate),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->jumpto), "activate",
		    G_CALLBACK (on_jumpto_activate),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->delblock), "activate",
		    G_CALLBACK (on_delblock_activate),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->save_event_list), "activate",
		    G_CALLBACK (on_save_event_list_activate),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (mt->load_event_list), "activate",
		    G_CALLBACK (on_load_event_list_activate),
		    (gpointer)mt);
  g_signal_connect (GTK_OBJECT (show_frame_events), "activate",
                      G_CALLBACK (show_frame_events_activate),
                      NULL);

  g_signal_connect (GTK_OBJECT (mt->select_track), "activate",
		    G_CALLBACK (on_seltrack_activate),
		    (gpointer)mt);

  hbox = gtk_hbox_new (FALSE, 10);
  gtk_box_pack_start (GTK_BOX (mt->top_vbox), hbox, FALSE, FALSE, 0);

  label=gtk_label_new (_("Timecode"));

  if (palette->style&STYLE_1) {
    gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
  }

  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 10);

  mt->timecode=gtk_entry_new();
  time_to_string (mt,0.,TIMECODE_LENGTH);
  gtk_editable_set_editable (GTK_EDITABLE(mt->timecode),FALSE);
  gtk_entry_set_max_length(GTK_ENTRY (mt->timecode),TIMECODE_LENGTH);
  gtk_entry_set_width_chars (GTK_ENTRY (mt->timecode),TIMECODE_LENGTH);
  GTK_WIDGET_UNSET_FLAGS (mt->timecode, GTK_CAN_FOCUS);
  gtk_box_pack_start (GTK_BOX (hbox), mt->timecode, FALSE, FALSE, 0);

  hseparator = gtk_hseparator_new ();
  gtk_box_pack_start (GTK_BOX (mt->top_vbox), hseparator, FALSE, FALSE, 0);

  mt->hbox = gtk_hbox_new (FALSE, 10);
  gtk_box_pack_start (GTK_BOX (mt->top_vbox), mt->hbox, FALSE, FALSE, 0);

  mt->play_blank = gtk_image_new_from_pixbuf (mainw->imframe);
  mt->play_box = gtk_vbox_new (FALSE, 10);
  gtk_widget_set_app_paintable(mt->play_box,TRUE);

  gtk_widget_set_size_request (mt->play_box, MT_PLAY_WIDTH, MT_PLAY_HEIGHT);
  gtk_box_pack_start (GTK_BOX (mt->hbox), mt->play_box, FALSE, FALSE, 0);
  gtk_container_add (GTK_CONTAINER (mt->play_box), mt->play_blank);


  vseparator = gtk_vseparator_new ();
  gtk_box_pack_start (GTK_BOX (mt->hbox), vseparator, FALSE, FALSE, 0);


  // prepare polymorph box
  mt->poly_box = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (mt->hbox), mt->poly_box, TRUE, TRUE, 0);

  // poly clip scroll
  mt->clip_scroll = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_ref (mt->clip_scroll);

  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (mt->clip_scroll), GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);

  mt->clip_inner_box = gtk_hbox_new (FALSE, 10);
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (mt->clip_scroll), mt->clip_inner_box);

  init_clips (mt,orig_file,FALSE);


  // poly in_out_box
  mt->in_out_box = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (mt->in_out_box);

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start(GTK_BOX(mt->in_out_box),vbox,FALSE,TRUE,0);
  spinbutton_adj = gtk_adjustment_new (0.,0.,0.,1./mt->fps,1.,0.);
  mt->spinbutton_in = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_adj), 1./mt->fps, 2);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (mt->spinbutton_in),TRUE);
  mt->in_image=gtk_image_new();
  gtk_box_pack_start(GTK_BOX(vbox),mt->in_image,FALSE,TRUE,0);

  mt->in_hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),mt->in_hbox,FALSE,FALSE,0);
  gtk_box_pack_start(GTK_BOX(mt->in_hbox),mt->spinbutton_in,FALSE,FALSE,0);

  mt->checkbutton_start_anchored = gtk_check_button_new_with_mnemonic (_ ("Anchor _start"));
  gtk_box_pack_end(GTK_BOX(mt->in_hbox),mt->checkbutton_start_anchored,FALSE,FALSE,0);

  mt->spin_in_func=g_signal_connect_after (GTK_OBJECT (mt->spinbutton_in), "value_changed",
					   G_CALLBACK (in_out_start_changed),
					   mt);

  mt->check_start_func=g_signal_connect_after (GTK_OBJECT (mt->checkbutton_start_anchored), "toggled",
					    G_CALLBACK (in_anchor_toggled),
					    mt);


  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_end(GTK_BOX(mt->in_out_box),vbox,FALSE,TRUE,0);
  spinbutton_adj = gtk_adjustment_new (0.,0.,0.,1./mt->fps,1.,0.);
  mt->spinbutton_out = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_adj), 1./mt->fps, 2);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (mt->spinbutton_out),TRUE);
  mt->out_image=gtk_image_new();
  gtk_box_pack_start(GTK_BOX(vbox),mt->out_image,FALSE,TRUE,0);
  mt->out_hbox = gtk_hbox_new (FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),mt->out_hbox,FALSE,FALSE,0);
  gtk_box_pack_start(GTK_BOX(mt->out_hbox),mt->spinbutton_out,FALSE,FALSE,0);
  mt->checkbutton_end_anchored = gtk_check_button_new_with_mnemonic (_ ("Anchor _end"));
  gtk_box_pack_end(GTK_BOX(mt->out_hbox),mt->checkbutton_end_anchored,FALSE,FALSE,0);

  mt->spin_out_func=g_signal_connect_after (GTK_OBJECT (mt->spinbutton_out), "value_changed",
					    G_CALLBACK (in_out_end_changed),
					    mt);

  mt->check_end_func=g_signal_connect_after (GTK_OBJECT (mt->checkbutton_end_anchored), "toggled",
					     G_CALLBACK (out_anchor_toggled),
					     mt);

  gtk_widget_show_all(mt->in_out_box);

  mt->poly_state=POLY_NONE;
  polymorph(mt,POLY_CLIPS);

  // context/info box
  vseparator = gtk_vseparator_new ();
  gtk_box_pack_start (GTK_BOX (mt->hbox), vseparator, FALSE, FALSE, 0);

  mt->context_box = gtk_vbox_new (FALSE, 10);
  gtk_box_pack_start (GTK_BOX (mt->hbox), mt->context_box, FALSE, FALSE, 0);

  if (mainw->imsep==NULL) {
    mt->sep_image=NULL;
    hseparator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (mt->top_vbox), hseparator, FALSE, FALSE, 2);
  }
  else {
    hseparator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (mt->top_vbox), hseparator, FALSE, FALSE, 0);
    mt->sep_image = gtk_image_new_from_pixbuf (mainw->imsep);
    gtk_box_pack_start (GTK_BOX (mt->top_vbox), mt->sep_image, FALSE, FALSE, 0);
    hseparator = gtk_hseparator_new ();
    gtk_box_pack_start (GTK_BOX (mt->top_vbox), hseparator, FALSE, FALSE, 0);
  }

  mt_init_start_end_spins(mt);

  mt->vpaned=gtk_vpaned_new();
  gtk_box_pack_start (GTK_BOX (mt->top_vbox), mt->vpaned, TRUE, TRUE, 0);

  g_signal_connect (GTK_OBJECT (mt->vpaned), "accept_position",
		    G_CALLBACK (paned_pos),
		    (gpointer)mt);

  tl_vbox = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (tl_vbox), 0);

  gtk_container_add (GTK_CONTAINER (mt->vpaned), tl_vbox);


  mt->timeline_table_header = gtk_table_new (1, 40, TRUE);

  eventbox=gtk_event_box_new();
  gtk_box_pack_start (GTK_BOX (tl_vbox), eventbox, TRUE, TRUE, 0);

  if (palette->style&STYLE_1) {
    if (palette->style&STYLE_3) gtk_widget_modify_bg(eventbox, GTK_STATE_NORMAL, &palette->normal_back);
    else gtk_widget_modify_bg(eventbox, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  g_signal_connect (GTK_OBJECT (eventbox), "button_press_event",
		    G_CALLBACK (on_track_header_click),
		    (gpointer)mt);

  g_signal_connect (GTK_OBJECT (eventbox), "button_release_event",
		    G_CALLBACK (on_track_header_release),
		    (gpointer)mt);

  gtk_widget_set_events (eventbox, GDK_BUTTON1_MOTION_MASK | GDK_BUTTON_RELEASE | GDK_BUTTON_PRESS);
  mt->mouse_mot1=g_signal_connect (GTK_OBJECT (eventbox), "motion_notify_event",
				   G_CALLBACK (on_track_header_move),
				   (gpointer)mt);
  g_signal_handler_block (eventbox,mt->mouse_mot1);

  hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (eventbox), hbox);

  vadjustment = gtk_adjustment_new (1.0,1.0,1.0,1.0,1.0,1.0);
  scrollbar=gtk_vscrollbar_new(GTK_ADJUSTMENT(vadjustment));
  gtk_widget_set_sensitive(scrollbar,FALSE);

  gtk_box_pack_start (GTK_BOX (hbox), mt->timeline_table_header, TRUE, TRUE, 0);
  gtk_box_pack_end (GTK_BOX (hbox), scrollbar, FALSE, FALSE, 10);

  mt->tl_hbox = gtk_hbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (mt->tl_hbox), 0);
    
  gtk_box_pack_start (GTK_BOX (tl_vbox), mt->tl_hbox, TRUE, TRUE, 0);

  mt->vadjustment = gtk_adjustment_new (0.0,0.0,1.0,1.0,mt->max_disp_vtracks,1.0);
  mt->scrollbar=gtk_vscrollbar_new(GTK_ADJUSTMENT(mt->vadjustment));

  g_signal_connect (GTK_OBJECT (mt->scrollbar), "value_changed",
		    G_CALLBACK (scroll_track_by_scrollbar),
		    (gpointer)mt);

  mt->tl_eventbox=gtk_event_box_new();
  gtk_box_pack_start (GTK_BOX (mt->tl_hbox), mt->tl_eventbox, TRUE, TRUE, 0);

  if (palette->style&STYLE_1) {
    if (palette->style&STYLE_3) gtk_widget_modify_bg(mt->tl_eventbox, GTK_STATE_NORMAL, &palette->normal_back);
    else gtk_widget_modify_bg(mt->tl_eventbox, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }

  g_signal_connect (GTK_OBJECT (mt->tl_eventbox), "button_press_event",
		    G_CALLBACK (on_track_between_click),
		    (gpointer)mt);

  g_signal_connect (GTK_OBJECT (mt->tl_eventbox), "button_release_event",
		    G_CALLBACK (on_track_between_release),
		    (gpointer)mt);

  gtk_widget_set_events (mt->tl_eventbox, GDK_BUTTON1_MOTION_MASK | GDK_BUTTON_RELEASE | GDK_BUTTON_PRESS);
  mt->mouse_mot2=g_signal_connect (GTK_OBJECT (mt->tl_eventbox), "motion_notify_event",
				  G_CALLBACK (on_track_move),
				  (gpointer)mt);

  g_signal_handler_block (mt->tl_eventbox,mt->mouse_mot2);

  gtk_box_pack_end (GTK_BOX (mt->tl_hbox), mt->scrollbar, FALSE, FALSE, 10);

  hseparator = gtk_hseparator_new ();
  gtk_box_pack_start (GTK_BOX (tl_vbox), hseparator, FALSE, FALSE, 20);

  mt->num_video_tracks=0;

  mt->timeline_table=NULL;
  init_tracks (mt);
  mainw->unordered_blocks=FALSE;

  gtk_widget_reparent (mainw->scrolledwindow,mt->vpaned);

  if (prefs->open_maximised) {
    gtk_window_maximize (GTK_WINDOW(mt->window));
  }

  // add info bar

  gtk_window_add_accel_group (GTK_WINDOW (mt->window), mt->accel_group);

  g_signal_connect (GTK_OBJECT (mt->window), "delete_event",
		    G_CALLBACK (on_mt_delete_event),
		    (gpointer)mt);

  mainw->multitrack=mt;

  gtk_accel_group_connect (GTK_ACCEL_GROUP (mt->accel_group), GDK_Page_Up, GDK_CONTROL_MASK, 0, g_cclosure_new (G_CALLBACK (mt_nextclip),mt,NULL));
  gtk_accel_group_connect (GTK_ACCEL_GROUP (mt->accel_group), GDK_Page_Down, GDK_CONTROL_MASK, 0, g_cclosure_new (G_CALLBACK (mt_prevclip),mt,NULL));

  gtk_accel_group_connect (GTK_ACCEL_GROUP (mt->accel_group), GDK_Left, GDK_CONTROL_MASK, 0, g_cclosure_new (G_CALLBACK (mt_tlback),mt,NULL));
  gtk_accel_group_connect (GTK_ACCEL_GROUP (mt->accel_group), GDK_Right, GDK_CONTROL_MASK, 0, g_cclosure_new (G_CALLBACK (mt_tlfor),mt,NULL));

  gtk_accel_group_connect (GTK_ACCEL_GROUP (mt->accel_group), GDK_Up, GDK_CONTROL_MASK, 0, g_cclosure_new (G_CALLBACK (mt_trup),mt,NULL));
  gtk_accel_group_connect (GTK_ACCEL_GROUP (mt->accel_group), GDK_Down, GDK_CONTROL_MASK, 0, g_cclosure_new (G_CALLBACK (mt_trdown),mt,NULL));

  mt->insert_mode=INSERT_MODE_NORMAL;
  mt->last_direction=DIRECTION_POSITIVE;

  on_mouse_mode_changed(GTK_MENU_ITEM(mt->mm_move),(gpointer)mt);
  on_snapt_mode_changed(GTK_MENU_ITEM(mt->snapt_off),(gpointer)mt);

  mt_sensitise(mt);

  return mt;
}



void multitrack_delete (lives_mt *mt) {
  // free lives_mt struct
  int i;
  gint new_file=-1;

  mainw->cancelled=CANCEL_NONE;

  if (mt->changed&&!(prefs->warning_mask&WARN_MASK_DISCARD_LAYOUT)) {
    _entryw *cdsw=create_cds_dialog(0,WARN_MASK_DISCARD_LAYOUT);
    gint resp=gtk_dialog_run(GTK_DIALOG(cdsw->dialog));
    gtk_widget_destroy(cdsw->dialog);
    g_free(cdsw);
    if (resp==0) return;
    if (resp==2) {
      on_save_event_list_activate(NULL,mt);
      if (mainw->cancelled!=CANCEL_NONE) {
	mainw->cancelled=CANCEL_NONE;
	return;
      }
    }
  }

  if (mt->poly_state==POLY_EFFECT) polymorph(mt,POLY_CLIPS);

  if (mt->undo_mem!=NULL) g_free(mt->undo_mem);

  g_list_free(mt->selected_tracks);

  if (mt->event_list!=NULL) event_list_free(mt->event_list);

  if (cfile->frames==0) close_current_file();

  if (mt->clip_selected>=0&&mainw->files[mt->clip_selected+1]!=NULL) new_file=mt->clip_selected+1;

  gtk_widget_reparent (mainw->scrolledwindow,mainw->message_box);
  gtk_widget_show (mainw->LiVES);


  // put buttons back in mainw->menubar
  gtk_widget_ref(mainw->m_sepwinbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_sepwinbutton->parent),mainw->m_sepwinbutton);
  gtk_container_add (GTK_CONTAINER(mainw->menubar), mainw->m_sepwinbutton);
  gtk_widget_unref(mainw->m_sepwinbutton);

  gtk_container_add (GTK_CONTAINER(mainw->menubar), mainw->m_rewindbutton);
  gtk_widget_unref(mainw->m_rewindbutton);

  gtk_widget_ref(mainw->m_playbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_playbutton->parent),mainw->m_playbutton);
  gtk_container_add (GTK_CONTAINER(mainw->menubar), mainw->m_playbutton);
  gtk_widget_unref(mainw->m_playbutton);

  gtk_widget_ref(mainw->m_stopbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_stopbutton->parent),mainw->m_stopbutton);
  gtk_container_add (GTK_CONTAINER(mainw->menubar), mainw->m_stopbutton);
  gtk_widget_unref(mainw->m_stopbutton);

  gtk_container_add (GTK_CONTAINER(mainw->menubar), mainw->m_playselbutton);
  gtk_widget_unref(mainw->m_playselbutton);

  gtk_widget_ref(mainw->m_loopbutton);
  gtk_container_remove(GTK_CONTAINER(mainw->m_loopbutton->parent),mainw->m_loopbutton);
  gtk_container_add (GTK_CONTAINER(mainw->menubar), mainw->m_loopbutton);
  gtk_widget_unref(mainw->m_loopbutton);

  gtk_container_add (GTK_CONTAINER(mainw->menubar), mainw->m_mutebutton);
  gtk_widget_unref(mainw->m_mutebutton);


  // free our track_rects
  if (mt->video_draws!=NULL) {
    for (i=0;i<mt->num_video_tracks;i++) {
      delete_video_track(mt,i,FALSE);
    }
    g_list_free (mt->video_draws);
  }
  gtk_widget_unref(mt->clip_scroll);
  gtk_widget_unref(mt->in_out_box);

  gtk_widget_destroy (mt->window);

  if (mainw->play_window!=NULL) {
    gtk_window_remove_accel_group (GTK_WINDOW (mainw->play_window), mt->accel_group);
    gtk_window_add_accel_group (GTK_WINDOW (mainw->play_window), mainw->accel_group);
  }

  g_free (mt);
  mainw->multitrack=NULL;

  if (new_file!=-1) switch_to_file ((mainw->current_file=0),mt->file_selected);

  reset_clip_menu();

  for (i=1;i<MAX_FILES;i++) {
    if (mainw->files[i]!=NULL) {
      if (mainw->files[i]->event_list!=NULL) event_list_free(mainw->files[i]->event_list);
      mainw->files[i]->event_list=NULL;
    }
  }
  if (mainw->current_file>0) sensitize();
  gtk_widget_hide(mainw->playframe);
  mainw->is_rendering=FALSE;
  d_print (_ ("\n==============================\nSwitched to Clip Edit mode\n"));
}



gboolean
on_mt_delete_event (GtkWidget *widget, GdkEvent *event, gpointer user_data) {
  multitrack_delete((lives_mt *)user_data);
  return FALSE;
}


inline track_rect *add_block_start_point (GtkWidget *eventbox, gint64 tc, gint filenum, gint64 offset_start, weed_plant_t *event, gboolean ordered) {
  // each mt->video_draw (eventbox) has a gulong data which points to a linked list of track_rect
  // here we create a new linked list item and set the start timecode in the timeline, offset in the source file, and start event
  // then append it to our list

  // block_last points to the last block added - not the last block in the track !!

  track_rect *block=(track_rect *)g_object_get_data (G_OBJECT(eventbox), "blocks");
  track_rect *new_block=(track_rect *)g_malloc (sizeof(track_rect));

  new_block->next=new_block->prev=NULL;
  new_block->state=BLOCK_UNSELECTED;
  new_block->start_anchored=new_block->end_anchored=FALSE;
  new_block->start_event=event;
  new_block->ordered=ordered;
  new_block->eventbox=eventbox;
  new_block->offset_start=offset_start;
  new_block->min_tc=new_block->max_tc=-1;

  g_object_set_data (G_OBJECT(eventbox),"block_last",(gpointer)new_block);

  while (block!=NULL) {
    if (get_event_timecode(block->start_event)>tc) {
      if (block->prev!=NULL) {
	block->prev->next=new_block;
	new_block->prev=block->prev;
      }
      else g_object_set_data (G_OBJECT(eventbox),"blocks",(gpointer)new_block);
      new_block->next=block;
      block->prev=new_block;
      break;
    }
    if (block->next==NULL) {
      block->next=new_block;
      new_block->prev=block;
      break;
    }
    block=block->next;
  }
  if (block==NULL) g_object_set_data (G_OBJECT(eventbox),"blocks",(gpointer)new_block);
  return new_block;
}


inline void add_block_end_point (GtkWidget *eventbox, weed_plant_t *event) {
  // here we add the end point to our last track_rect
  track_rect *block=(track_rect *)g_object_get_data (G_OBJECT(eventbox),"block_last");
  block->end_event=event;
}


void init_tracks (lives_mt *mt) {
  GtkWidget *label;

  if (mt->video_draws!=NULL) g_list_free (mt->video_draws);
  mt->video_draws=NULL;

  label=gtk_label_new (_("Timeline (seconds)"));

  if (palette->style&STYLE_1) {
    gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
  }

  gtk_table_attach (GTK_TABLE (mt->timeline_table_header), label, 0, 6, 0, 1, GTK_FILL, 0, 0, 0);

  mt->current_track=0;

  // start with 2 video tracks
  add_video_track_behind (NULL,mt);
  add_video_track_behind (NULL,mt);

  mt->current_track=0;

  mt->timeline = gtk_hruler_new();
  mt->timeline_eb=gtk_event_box_new();
  gtk_widget_show(mt->timeline_eb);
  gtk_widget_set_events (mt->timeline_eb, GDK_POINTER_MOTION_MASK | GDK_BUTTON1_MOTION_MASK | GDK_BUTTON_RELEASE | GDK_BUTTON_PRESS);

  g_signal_connect (GTK_OBJECT (mt->timeline), "motion_notify_event",
		    G_CALLBACK (return_true),
		    NULL);

  g_signal_connect (GTK_OBJECT (mt->timeline_eb), "motion_notify_event",
		    G_CALLBACK (on_timeline_update),
		    (gpointer)mt);

  g_signal_connect (GTK_OBJECT (mt->timeline_eb), "button_release_event",
		    G_CALLBACK (on_timeline_release),
		    (gpointer)mt);

  g_signal_connect (GTK_OBJECT (mt->timeline_eb), "button_press_event",
		    G_CALLBACK (on_timeline_press),
		    (gpointer)mt);

  g_signal_connect_after (GTK_OBJECT (mt->timeline), "expose_event",
			  G_CALLBACK (expose_timeline_event),
			  (gpointer)mt);


  gtk_container_add (GTK_CONTAINER (mt->timeline_eb), mt->timeline);

  gtk_table_attach (GTK_TABLE (mt->timeline_table_header), mt->timeline_eb, 6, 40, 0, 1,
		    (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
		    (GtkAttachOptions) (GTK_EXPAND), 0, 0);

  if (mt->event_list!=NULL) {
    gint last_tracks=1; // number of video tracks on timeline
    int *frame_index;
    int *clip_index,*new_clip_index,*new_frame_index;
    weed_plant_t *event,*last_event=NULL,*next_frame_event;
    gint num_tracks;
    int tracks[65536]; // TODO - use linked list
    int j,error;
    gint64 tc;
    gint last_valid_frame;
    weed_timecode_t offset_start;
    weed_timecode_t block_marker_tc=-1;
    int block_marker_num_tracks=0;
    int *block_marker_tracks=NULL;
    gboolean forced_end;
    weed_timecode_t block_marker_uo_tc=-1;
    int block_marker_uo_num_tracks=0;
    int *block_marker_uo_tracks=NULL;
    gboolean ordered;

    for (j=0;j<65536;j++) tracks[j]=0;

    // draw coloured blocks to represent the FRAME events
    event=get_first_event(mt->event_list);
    while (event!=NULL) {
      if (WEED_EVENT_IS_MARKER(event)) {
	if (weed_get_int_value(event,"lives_type",&error)==EVENT_MARKER_BLOCK_START) {
	  block_marker_tc=get_event_timecode(event);
	  block_marker_num_tracks=weed_leaf_num_elements(event,"tracks");
	  if (block_marker_tracks!=NULL) weed_free(block_marker_tracks);
	  block_marker_tracks=weed_get_int_array(event,"tracks",&error);
	}
	else if (weed_get_int_value(event,"lives_type",&error)==EVENT_MARKER_BLOCK_UNORDERED) {
	  block_marker_uo_tc=get_event_timecode(event);
	  block_marker_uo_num_tracks=weed_leaf_num_elements(event,"tracks");
	  if (block_marker_uo_tracks!=NULL) weed_free(block_marker_uo_tracks);
	  block_marker_uo_tracks=weed_get_int_array(event,"tracks",&error);
	}
      }
      else if (WEED_EVENT_IS_FRAME(event)) {
	tc=get_event_timecode (event);
	num_tracks=weed_leaf_num_elements (event,"clips");

	clip_index=weed_get_int_array(event,"clips",&error);
	frame_index=weed_get_int_array(event,"frames",&error);

	if (num_tracks<last_tracks) {
	  for (j=num_tracks;j<last_tracks;j++) {
	    // TODO - tracks should be linked list
	    if (tracks[j]>0) {
	      add_block_end_point (GTK_WIDGET(g_list_nth_data(mt->video_draws,j)),last_event); // end of previous rectangle
	      tracks[j]=0;
	    }
	  }
	}

	if (num_tracks>mt->num_video_tracks) {
	  for (j=mt->num_video_tracks;j<num_tracks;j++) {
	    add_video_track_behind(NULL,mt);
	  }
	}

	last_tracks=num_tracks;
	new_clip_index=g_malloc(num_tracks*sizint);
	new_frame_index=g_malloc(num_tracks*sizint);
	last_valid_frame=0;

	for (j=0;j<num_tracks;j++) {
	  // TODO - tracks should be linked list
	  if (renumbered_clips[clip_index[j]]>0&&frame_index[j]<=mainw->files[renumbered_clips[clip_index[j]]]->frames) {
	    forced_end=FALSE;
	    if (tc==block_marker_tc&&int_array_contains_value(block_marker_tracks,block_marker_num_tracks,j)) forced_end=TRUE;
	    if ((tracks[j]!=renumbered_clips[clip_index[j]])||forced_end) {
	      if (tracks[j]>0) {
		add_block_end_point (GTK_WIDGET(g_list_nth_data(mt->video_draws,j)),last_event); // end of previous rectangle
	      }
	      if (clip_index[j]>0) {
		ordered=!mainw->unordered_blocks;
		if (tc==block_marker_uo_tc&&int_array_contains_value(block_marker_uo_tracks,block_marker_uo_num_tracks,j)) ordered=FALSE;
		// start a new rectangle
		offset_start=calc_variable_time(clip_index[j],frame_index[j])*U_SEC;
		add_block_start_point (GTK_WIDGET(g_list_nth_data(mt->video_draws,j)),tc,renumbered_clips[clip_index[j]],offset_start,event,ordered);
	      }
	      tracks[j]=renumbered_clips[clip_index[j]];
	    }
	    new_clip_index[j]=renumbered_clips[clip_index[j]];
	    new_frame_index[j]=frame_index[j];
	    last_valid_frame=j+1;
	  }
	  else {
	    // clip has probably been closed, so we remove its frames
	    new_clip_index[j]=-1;
	    new_frame_index[j]=0;
	    if (tracks[j]>0) {
	      add_block_end_point (GTK_WIDGET(g_list_nth_data(mt->video_draws,j)),last_event); // end of previous rectangle
	      tracks[j]=0;
	    }
	  }
	}

	if (last_valid_frame==0) {
	  g_free(new_clip_index);
	  g_free(new_frame_index);
	  new_clip_index=g_malloc(sizint);
	  new_frame_index=g_malloc(sizint);
	  *new_clip_index=-1;
	  *new_frame_index=0;
	  num_tracks=1;
	}
	else {
	  if (last_valid_frame<num_tracks) {
	    g_free(new_clip_index);
	    g_free(new_frame_index);
	    new_clip_index=g_malloc(last_valid_frame*sizint);
	    new_frame_index=g_malloc(last_valid_frame*sizint);
	    for (j=0;j<last_valid_frame;j++) {
	      new_clip_index[j]=clip_index[j];
	      new_frame_index[j]=frame_index[j];
	    }
	  }
	}

	weed_set_int_array(event,"clips",num_tracks,new_clip_index);
	weed_set_int_array(event,"frames",num_tracks,new_frame_index);

	weed_free(clip_index);
	g_free(new_clip_index);
	weed_free(frame_index);
	g_free(new_frame_index);

	next_frame_event=get_next_frame_event(event);
      
	if (next_frame_event==NULL) {
	  // this is the last FRAME event, so close all our rectangles
	  for (j=0;j<mt->num_video_tracks;j++) {
	    if (tracks[j]>0) {
	      add_block_end_point (GTK_WIDGET(g_list_nth_data(mt->video_draws,j)),event);
	    }
	  }
	}
	last_event=event;
      }
      event=get_next_event(event);
    }
    remove_end_blank_frames(mt->event_list);
    if (block_marker_tracks!=NULL) weed_free(block_marker_tracks);
    if (block_marker_uo_tracks!=NULL) weed_free(block_marker_uo_tracks);
  }
  if (mt->event_list!=NULL) set_timeline_end_secs(mt,event_list_get_end_secs(mt->event_list)*8.);
  else set_timeline_end_secs(mt,DEF_TIME);
}



void delete_video_track(lives_mt *mt, gint layer, gboolean full) {
  // WARNING - does not yet delete events from event_list
  // only deletes visually

  GtkWidget *eventbox=g_list_nth_data(mt->video_draws,layer);
  track_rect *block=(track_rect *)g_object_get_data (G_OBJECT(eventbox), "blocks"),*blocknext;

  GtkWidget *checkbutton;
  GtkWidget *label,*labelbox;

  while (block!=NULL) {
    blocknext=block->next;
    if (mt->block_selected==block) mt->block_selected=NULL;
    g_free(block);
    block=blocknext;
  }

  checkbutton=g_object_get_data (G_OBJECT(eventbox), "checkbutton");
  label=g_object_get_data (G_OBJECT(eventbox), "label");
  labelbox=g_object_get_data (G_OBJECT(eventbox), "labelbox");
  gtk_widget_destroy(checkbutton);
  gtk_widget_destroy(label);
  gtk_widget_destroy(eventbox);
  gtk_widget_destroy(labelbox);
}



void add_video_track (lives_mt *mt, gboolean behind) {
  // add another video track to our timeline_table
  GtkWidget *label;
  GtkWidget *checkbutton;
  GtkWidget *eventbox;  // each track has an eventbox, which we store in GList *video_draws
  int i;
  GList *liste;

  mt_desensitise(mt);

  mt->num_video_tracks++;

  checkbutton = gtk_check_button_new ();
  gtk_widget_ref(checkbutton);

  eventbox=gtk_event_box_new();
  gtk_widget_ref(eventbox);

  g_object_set_data (G_OBJECT(eventbox),"track_name",g_strdup_printf(_("Video %d"),mt->num_video_tracks));
  g_object_set_data (G_OBJECT(eventbox),"checkbutton",(gpointer)checkbutton);

  g_object_set_data (G_OBJECT(eventbox), "blocks", (gpointer)NULL);
  g_object_set_data (G_OBJECT(eventbox), "hidden", GINT_TO_POINTER(0));

  GTK_ADJUSTMENT(mt->vadjustment)->upper=(gdouble)mt->num_video_tracks;
  GTK_ADJUSTMENT(mt->vadjustment)->page_size=mt->max_disp_vtracks>mt->num_video_tracks?(gdouble)mt->num_video_tracks:(gdouble)mt->max_disp_vtracks;

  if (!behind) {
    // track in front of (above) stack
    // shift all rows down
    // increment "layer_number"s, change labels
    g_object_set_data (G_OBJECT(eventbox),"layer_number",GINT_TO_POINTER(0));
    for (i=0;i<mt->num_video_tracks-1;i++) {
      gchar *newtext;
      GtkWidget *xeventbox=g_list_nth_data(mt->video_draws,i);
      GtkWidget *xcheckbutton=g_object_get_data(G_OBJECT(xeventbox),"checkbutton");
      g_object_set_data (G_OBJECT(xeventbox),"layer_number",GINT_TO_POINTER(i+1));
      g_object_set_data (G_OBJECT(xcheckbutton),"layer_number",GINT_TO_POINTER(i+1));
      label=GTK_WIDGET(g_object_get_data(G_OBJECT(xeventbox),"label"));
      newtext=g_strdup_printf(_("%s (layer %d)"),g_object_get_data(G_OBJECT(xeventbox),"track_name"),i+1);
      g_free(GTK_LABEL(label)->text);
      GTK_LABEL(label)->text=newtext;
    }
    // add a -1,0 in all frame events
    // renumber "in_tacks", "out_tracks" in effect_init events
    event_list_add_track(mt->event_list,0);

    mt->video_draws=g_list_prepend (mt->video_draws, (gpointer)eventbox);
    mt->current_track=0;

    //renumber all tracks in mt->selected_tracks
    liste=mt->selected_tracks;
    while (liste!=NULL) {
      liste->data=GINT_TO_POINTER(GPOINTER_TO_INT(liste->data)+1);
      liste=liste->next;
    }
    g_object_set_data(G_OBJECT(eventbox),"hidden",GINT_TO_POINTER(TRACK_I_OUTSCROLLED_PRE));
  }
  else {
    // add track behind (below) stack
    mt->video_draws=g_list_append (mt->video_draws, (gpointer)eventbox);
    g_object_set_data (G_OBJECT(eventbox),"layer_number",GINT_TO_POINTER(mt->num_video_tracks-1));
    g_object_set_data (G_OBJECT(checkbutton),"layer_number",GINT_TO_POINTER(mt->num_video_tracks-1));
    g_object_set_data(G_OBJECT(eventbox),"hidden",GINT_TO_POINTER(TRACK_I_OUTSCROLLED_POST));
    mt->current_track=mt->num_video_tracks-1;
  }

  label=gtk_label_new (g_strdup_printf(_("%s (layer %d)"),g_object_get_data (G_OBJECT(eventbox),"track_name"),GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"layer_number"))));
  gtk_widget_ref(label);

  if (palette->style&STYLE_1) {
    if (!palette->style&STYLE_3) {
      gtk_widget_modify_fg (label, GTK_STATE_NORMAL, &palette->normal_back);
      gtk_widget_modify_fg (label, GTK_STATE_SELECTED, &palette->normal_back);
    }
  }

  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_LEFT);

  g_object_set_data (G_OBJECT(eventbox),"label",label);

  if (!behind) scroll_track_on_screen(0,mt);
  else scroll_track_on_screen(mt->num_video_tracks-1,mt);


  mt_sensitise(mt);

  gtk_widget_queue_draw (mt->vpaned);

}

void add_video_track_behind (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  add_video_track (mt,TRUE);
}

void add_video_track_front (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  add_video_track (mt,FALSE);
}


static void set_cursor_style(lives_mt *mt, gint cstyle, gint width, gint height, gint clip, gint hsx, gint hsy) {
  GdkCursor *cursor;
  GdkPixbuf *pixbuf=NULL;
  guchar *cpixels,*tpixels;
  int i,j,k;
  GdkPixbuf *thumbnail;
  gint twidth=0,twidth3,trow;
  file *sfile=mainw->files[clip];

  switch(cstyle) {
  case LIVES_CURSOR_NORMAL:
    gdk_window_set_cursor (mainw->multitrack->window->window, NULL);
    return;
  case LIVES_CURSOR_BLOCK:
    pixbuf=gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, width, height);
    for (i=0;i<width;i+=BLOCK_THUMB_WIDTH) {
      // create a small thumb
      twidth=BLOCK_THUMB_WIDTH;
      if ((i+twidth)>width) twidth=width-i;
      thumbnail=make_thumb(clip,twidth,height,sfile->start+(gint)((gdouble)i/(gdouble)width*(gdouble)(sfile->end-sfile->start)));
      // render it in the eventbox
      if (thumbnail!=NULL) {
	trow=gdk_pixbuf_get_rowstride(thumbnail);
	twidth=gdk_pixbuf_get_width(thumbnail);
	twidth3=twidth*3;
	cpixels=gdk_pixbuf_get_pixels(pixbuf)+(i*4);
	tpixels=gdk_pixbuf_get_pixels(thumbnail);
	for (j=0;j<height;j++) {
	  for (k=0;k<twidth3;k+=3) {
	    memcpy(cpixels,&tpixels[k],3);
	    memset(cpixels+3,0xFF,1);
	    cpixels+=4;
	  }
	  tpixels+=trow;
	  cpixels+=(width-twidth)<<2;
	}
      }
    }

    break;
  }

  cursor = gdk_cursor_new_from_pixbuf (gdk_display_get_default(), pixbuf, hsx, hsy);

  gdk_window_set_cursor (mainw->multitrack->window->window, cursor);
  gdk_pixbuf_unref (pixbuf);

}



static gboolean
clip_ebox_pressed (GtkWidget *eventbox, GdkEventButton *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  if (event->type!=GDK_BUTTON_PRESS) {
    set_cursor_style(mt,LIVES_CURSOR_NORMAL,0,0,0,0,0);
    // double click, open up in clip editor
    multitrack_delete(mt);
    return TRUE;
  }
  mt->clip_selected=GPOINTER_TO_INT(g_object_get_data (G_OBJECT(eventbox),"clip"));
  clip_select(mt);
  return FALSE;
}





static gboolean
on_drag_clip_start           (GtkWidget       *widget,
			      GdkEventButton  *event,
			      gpointer         user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  int width,height;
  gint ebwidth;
  file *sfile;
  gint file;

  // change cursor to block
  if (mt->video_draws==NULL) {
    return FALSE;
  }
  else {
    ebwidth=GTK_WIDGET(mt->timeline)->allocation.width;
    file=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(widget),"file"));
    sfile=mainw->files[file];
    width=(sfile->end-sfile->start+1.)/sfile->fps;
    width=(gint)((gdouble)width/mt->end_secs*(gdouble)ebwidth+1.);
    if (width>ebwidth) width=ebwidth;
    if (width<2) width=2;
    height=GTK_WIDGET(g_list_nth_data(mt->video_draws,0))->allocation.height;
    set_cursor_style(mt,LIVES_CURSOR_BLOCK,width,height,file,0,height/2);
    mt->hotspot_x=mt->hotspot_y=0;
  }
  
  return FALSE;

}



static gboolean
on_drag_clip_end           (GtkWidget       *widget,
			      GdkEventButton  *event,
			      gpointer         user_data) {
  GdkWindow *window;
  GtkWidget *eventbox;
  lives_mt *mt=(lives_mt *)user_data;
  int win_x,win_y;
  int i;
  gdouble timesecs;

  set_cursor_style(mt,LIVES_CURSOR_NORMAL,0,0,0,0,0);
  window=gdk_display_get_window_at_pointer (mt->display,&win_x,&win_y);

  for (i=0;i<g_list_length(mt->video_draws);i++) {
    if (GTK_WIDGET((eventbox=g_list_nth_data(mt->video_draws,i)))->window==window) {
      gdk_window_get_pointer(GDK_WINDOW (mt->timeline->window), &mt->sel_x, &mt->sel_y, NULL);
      timesecs=(gdouble)mt->sel_x/(gdouble)mt->timeline->allocation.width*mt->end_secs;
      if (timesecs>mt->end_secs) timesecs=mt->end_secs;
      
      timesecs=(gint)(timesecs*mt->fps+.5)/mt->fps; // quantise to frame timing
      
      if (timesecs<0.) timesecs=0.;

      mt->current_track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"layer_number"));
      track_select(mt);

      if (mainw->playing_file==-1) {
	GTK_RULER (mt->timeline)->position=timesecs;
	gtk_widget_queue_draw (mainw->multitrack->timeline);
      }
      insert_here_cb(NULL,mt);
      break;
    }
  }
  return FALSE;
}




void init_clips (lives_mt *mt, gint orig_file, gboolean add) {
  GtkWidget *thumb_image=NULL;
  GtkWidget *vbox, *label;
  GtkWidget *eventbox;
  gchar clip_name[CLIP_LABEL_LENGTH];
  GdkPixbuf *thumbnail;
  int i=1,j=0;
  gint width=CLIP_THUMB_WIDTH,height=CLIP_THUMB_HEIGHT;
  gchar filename[256];

  mt->clip_selected=-1;

  if (add) {
    i=orig_file;
    j=i-1;
  }

  for (;(!add&&(i<MAX_FILES))||(add&&(i==orig_file));i++) {
    if (mainw->files[i]!=NULL&&i!=mainw->scrap_file&&mainw->files[i]->clip_type==CLIP_TYPE_DISK) {
      if (i==orig_file||(mt->clip_selected==-1&&i==mainw->pre_src_file)) {
	mt->clip_selected=j;
      }
      // make a small thumbnail, add it to the clips box
      thumbnail=make_thumb(i,width,height,1);
      if (thumbnail!=NULL) {
	eventbox=gtk_event_box_new();
	if (palette->style&STYLE_1) {
	  gtk_widget_modify_bg (eventbox, GTK_STATE_SELECTED, &palette->menu_and_bars);
	}
	gtk_widget_set_events (eventbox, GDK_BUTTON_RELEASE | GDK_BUTTON_PRESS);
	g_object_set_data(G_OBJECT(eventbox),"clip",GINT_TO_POINTER(i-1)); // clip number to hightlight
	vbox = gtk_vbox_new (FALSE, 10);
	g_object_set_data (G_OBJECT(eventbox),"file",GINT_TO_POINTER(i)); // LiVES mainw->files[] number 
	thumb_image=gtk_image_new();
	gtk_image_set_from_pixbuf(GTK_IMAGE(thumb_image),thumbnail);
	gdk_pixbuf_unref(thumbnail);
	gtk_container_add (GTK_CONTAINER (eventbox), vbox);
	gtk_box_pack_start (GTK_BOX (mt->clip_inner_box), eventbox, FALSE, FALSE, 0);
	g_snprintf (filename,256,"%s",mainw->files[i]->name);
	get_basename(filename);
	g_snprintf (clip_name,CLIP_LABEL_LENGTH,"  %s  ",filename);
	label=gtk_label_new (clip_name);
	if (palette->style&STYLE_3) gtk_widget_modify_fg (label, GTK_STATE_SELECTED, &palette->info_text);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), thumb_image, FALSE, FALSE, 0);
	label=gtk_label_new (g_strdup_printf (_("%d frames"),mainw->files[i]->frames));
	if (palette->style&STYLE_3) gtk_widget_modify_fg (label, GTK_STATE_SELECTED, &palette->info_text);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
	label=gtk_label_new (g_strdup_printf (_("  %d to %d selected  "),mainw->files[i]->start,mainw->files[i]->end));
	if (palette->style&STYLE_3) gtk_widget_modify_fg (label, GTK_STATE_SELECTED, &palette->info_text);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
	label=gtk_label_new (g_strdup_printf (_("%.2f sec."),(mainw->files[i]->end-mainw->files[i]->start+1.)/mainw->files[i]->fps));
	if (palette->style&STYLE_3) gtk_widget_modify_fg (label, GTK_STATE_SELECTED, &palette->info_text);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
	j++;

	g_signal_connect (GTK_OBJECT (eventbox), "button_press_event",
			  G_CALLBACK (clip_ebox_pressed),
			  (gpointer)mt);
	g_signal_connect (GTK_OBJECT (eventbox), "button_release_event",
			  G_CALLBACK (on_drag_clip_end),
			  (gpointer)mt);
	g_signal_connect (GTK_OBJECT (eventbox), "button_press_event",
			  G_CALLBACK (on_drag_clip_start),
			  (gpointer)mt);

	if (add) gtk_widget_show_all(eventbox);
      }
    }
  }
}


gboolean on_multitrack_activate (GtkMenuItem *menuitem, weed_plant_t *event_list) {
  //returns TRUE if we go into mt mode
  gint orig_file;
  int i;
  gboolean event_warn=FALSE;
  gboolean response;

  lives_mt *multi;

  for (i=1;i<MAX_FILES;i++) {
    if (mainw->files[i]!=NULL&&mainw->files[i]->event_list!=NULL) {
      if (!event_warn&&!do_warning_dialog(_("Some clips have frame changes recorded.\nIf you go to multitrack mode, you will lose these events\n\nClick OK to continue, Cancel to return to the clip editor.\n"))) return FALSE;
      event_warn=TRUE;
      event_list_free(mainw->files[i]->event_list);
      mainw->files[i]->event_list=NULL;
    }
  }

  if (prefs->mt_enter_prompt) {
    rdet=create_render_details(FALSE);  // WARNING !! - rdet is global in events.h
    rdet->enc_changed=FALSE;
    if ((response=gtk_dialog_run(GTK_DIALOG(rdet->dialog)))==GTK_RESPONSE_OK) if (rdet->enc_changed) check_encoder_restrictions(FALSE);
    gtk_widget_destroy (rdet->dialog); 
    
    if (response==GTK_RESPONSE_CANCEL) {
      g_free(rdet->encoder_name);
      g_free(rdet);
      rdet=NULL;
      return FALSE;
    }
  }

  // create new file for rendering to
  renumber_clips();
  orig_file=mainw->current_file;
  mainw->current_file=mainw->first_free_file;

  if (!get_new_handle(mainw->current_file,NULL)) {
    mainw->current_file=orig_file;

    if (prefs->mt_enter_prompt) {
      g_free(rdet->encoder_name);
      g_free(rdet);
      rdet=NULL;
    }
    return FALSE; // show dialog again
  }

  if (prefs->mt_enter_prompt) {
    cfile->hsize=rdet->width;
    cfile->vsize=rdet->height;
    cfile->pb_fps=cfile->fps=rdet->fps;
    cfile->ratio_fps=rdet->ratio_fps;
    g_free(rdet->encoder_name);
    g_free(rdet);
    rdet=NULL;
  }
  else {
    cfile->hsize=prefs->mt_def_width;
    cfile->vsize=prefs->mt_def_height;
    cfile->pb_fps=cfile->fps=prefs->mt_def_fps;
    cfile->ratio_fps=FALSE;
  }

  cfile->bpp=24;
  cfile->changed=TRUE;
  cfile->is_loaded=TRUE;
  
  cfile->old_frames=cfile->frames;

  // if we have an existing event list, we will quantise it to the selected fps
  if (event_list!=NULL) {
    weed_plant_t *qevent_list=quantise_events(event_list,cfile->fps,FALSE);
    event_list_replace_events(event_list,qevent_list);
    weed_set_double_value(event_list,"fps",cfile->fps);
    event_list_rectify(event_list,FALSE);
  }

  multi=multitrack(event_list,orig_file,cfile->fps);
  gtk_widget_show_all (multi->window);
  gtk_widget_hide (mainw->LiVES);
  gtk_widget_hide(multi->preview_button);

  multi->render_file=mainw->current_file;
  multi->achans=multi->arate=multi->asampsize=0;

  track_select(multi);
  clip_select(multi);

  gtk_window_set_title (GTK_WINDOW (multi->window), g_strdup_printf(_("LiVES-%s: Multitrack %dx%d : %d bpp %.3f fps"),LiVES_VERSION,cfile->hsize,cfile->vsize,cfile->bpp,cfile->fps));

  if (mainw->preview_box!=NULL&&mainw->preview_box->parent!=NULL) {
    g_object_unref(mainw->preview_box);
    gtk_container_remove (GTK_CONTAINER (mainw->play_window), mainw->preview_box);
    mainw->preview_box=NULL;
  }
  
  if (mainw->play_window!=NULL) {
    g_signal_handler_unblock(mainw->play_window,mainw->pw_exp_func);
    gtk_window_remove_accel_group (GTK_WINDOW (mainw->play_window), mainw->accel_group);
    gtk_window_add_accel_group (GTK_WINDOW (mainw->play_window), multi->accel_group);
    resize_play_window();
  }
  d_print (_ ("\n==============================\nSwitched to Multitrack mode\n"));
  return TRUE;
}


gboolean block_overlap (GtkWidget *eventbox, gdouble time_start, gdouble time_end) {
  gint64 tc_start=time_start*U_SECL;
  gint64 tc_end=time_end*U_SECL;
  track_rect *block=(track_rect *)g_object_get_data (G_OBJECT(eventbox),"blocks");

  while (block!=NULL) {
    if (get_event_timecode(block->start_event)>tc_end) return FALSE;
    if (get_event_timecode(block->end_event)>=tc_start) return TRUE;
    block=block->next;
  }
  return FALSE;
}  



track_rect *get_block_from_time (GtkWidget *eventbox, gdouble time, lives_mt *mt) {
  gint64 tc=time*U_SECL;
  track_rect *block=(track_rect *)g_object_get_data (G_OBJECT(eventbox),"blocks");
  tc=q_gint64(tc,mt->fps);

  while (block!=NULL) {
    if (get_event_timecode(block->start_event)>tc) return NULL;
    if (get_event_timecode(block->end_event)>=tc) break;
    block=block->next;
  }
  return block;
}


track_rect *get_block_before (GtkWidget *eventbox, gdouble time) {
  gint64 tc=time*U_SECL;
  track_rect *block=(track_rect *)g_object_get_data (G_OBJECT(eventbox),"blocks"),*last_block=NULL;

  while (block!=NULL) {
    if (get_event_timecode(block->start_event)>tc) break;
    last_block=block;
    block=block->next;
  }
  return last_block;
}

track_rect *get_block_after (GtkWidget *eventbox, gdouble time) {
  gint64 tc=time*U_SECL;
  track_rect *block=(track_rect *)g_object_get_data (G_OBJECT(eventbox),"blocks");

  while (block!=NULL) {
    if (get_event_timecode(block->start_event)>tc) break;
    block=block->next;
  }
  return block;
}


static track_rect *move_block (lives_mt *mt, track_rect *block, gdouble timesecs, gint old_track, gint new_track) {
  weed_plant_t *first_event;
  weed_timecode_t new_start_tc,end_tc;
  weed_timecode_t start_tc=get_event_timecode(block->start_event);
  GtkWidget *eventbox;
  gint clip;
  gboolean did_backup=mt->did_backup;

  if (!did_backup) mt_backup(mt,MT_UNDO_MOVE_BLOCK,NULL);
  
  mt->block_selected=block;
  end_tc=get_event_timecode(block->end_event);
  
  first_event=get_prev_event(block->start_event);
  while (first_event!=NULL&&get_event_timecode(first_event)==start_tc) {
    first_event=get_prev_event(first_event);
  }
  
  mt->insert_start=block->offset_start;
  mt->insert_end=block->offset_start+end_tc-start_tc;

  clip=get_frame_event_clip(block->start_event,old_track);

  mt->moving_block=TRUE;
  delete_block_cb(NULL,(gpointer)mt);
  mt->block_selected=NULL;
  mt->current_track=new_track;
  track_select(mt);
  mt->clip_selected=clip-1;
  clip_select(mt);
  GTK_RULER (mt->timeline)->position=timesecs;

  insert_here_cb(NULL,(gpointer)mt);
  
  mt->insert_start=mt->insert_end=-1;

  new_start_tc=q_gint64(timesecs*U_SEC,mt->fps);
  if (mt->move_effects) update_filter_events(mt,first_event,start_tc,end_tc,old_track,new_start_tc,mt->current_track);
  mt->moving_block=FALSE;
  remove_end_blank_frames(mt->event_list);

  if (mt->snapt&&!did_backup) {
    gdouble oldr_start=mt->region_start;
    gdouble oldr_end=mt->region_end;
    GList *tracks_sel=g_list_copy(mt->selected_tracks);
    if (mt->selected_tracks!=NULL) g_list_free(mt->selected_tracks);
    mt->selected_tracks=NULL;
    mt->selected_tracks=g_list_append(mt->selected_tracks,GINT_TO_POINTER(old_track));
    mt->region_start=start_tc/U_SEC;
    mt->region_end=mt->end_secs;
    remove_first_gaps(NULL,mt);
    g_list_free(mt->selected_tracks);
    mt->selected_tracks=g_list_copy(tracks_sel);
    if (tracks_sel!=NULL) g_list_free(tracks_sel);
    mt->region_start=oldr_start;
    mt->region_end=oldr_end;
  }

  eventbox=g_list_nth_data(mt->video_draws,mt->current_track);
  mt->did_backup=did_backup;

  return (track_rect *)g_object_get_data(G_OBJECT(eventbox),"block_last");
}


void unselect_all (lives_mt *mt) {
  // unselect all blocks 
  int i;
  GtkWidget *eventbox;
  track_rect *trec;

  if (mt->block_selected!=NULL) gtk_widget_queue_draw(mt->block_selected->eventbox);

  for (i=0;i<mt->num_video_tracks;i++) {
    eventbox=g_list_nth_data(mt->video_draws,i);
    if (eventbox!=NULL) {
      trec=(track_rect *)g_object_get_data (G_OBJECT(eventbox),"blocks");
      while (trec!=NULL) {
	trec->state=BLOCK_UNSELECTED;
	trec=trec->next;
      }
    }
  }
  mt->block_selected=NULL;
  gtk_widget_set_sensitive(mt->view_in_out,FALSE);
  gtk_widget_set_sensitive (mt->delblock, FALSE);
  gtk_widget_set_sensitive (mt->fx_block, FALSE);
  polymorph (mt,POLY_CLIPS);
}


void clear_context (lives_mt *mt) {
  gtk_widget_destroy(mt->context_box);
  mt->context_box = gtk_vbox_new (FALSE, 10);
  gtk_box_pack_start (GTK_BOX (mt->hbox), mt->context_box, FALSE, FALSE, 0);
  gtk_widget_show (mt->context_box);
}

void add_context_label (lives_mt *mt, gchar *text) {
  GtkWidget *label=gtk_label_new (text);
  if (palette->style&STYLE_1) {
    if (palette->style&STYLE_3) gtk_widget_modify_fg (label, GTK_STATE_SELECTED, &palette->info_text);
    else gtk_widget_modify_fg(label, GTK_STATE_NORMAL, &palette->normal_fore);
  }
  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
  gtk_label_set_justify (GTK_LABEL (label), GTK_JUSTIFY_CENTER);
  gtk_widget_show (label);
  gtk_box_pack_start (GTK_BOX (mt->context_box), label, FALSE, FALSE, 0);
}



gboolean resize_timeline (lives_mt *mt) {
  gdouble end_secs;

  if (mt->event_list==NULL||get_first_event(mt->event_list)==NULL||mt->tl_fixed_length>0.) return FALSE;

  end_secs=event_list_get_end_secs(mt->event_list);

  if (end_secs>mt->end_secs) {
    set_timeline_end_secs(mt,end_secs);
    return TRUE;
  }
  return FALSE;
}



static void set_in_out_spin_ranges(lives_mt *mt, gint64 start_tc, gint64 end_tc) {
  track_rect *block=mt->block_selected;
  gint64 min_tc=block->min_tc,max_tc=block->max_tc,offset_start=get_event_timecode(block->start_event);
  gint track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(block->eventbox),"layer_number"));
  gint filenum=get_frame_event_clip(block->start_event,track);

  gdouble in_val=start_tc/U_SEC,out_val=end_tc/U_SEC,in_start_range=0.,out_start_range=in_val+1./mt->fps,out_end_range=count_resampled_frames(mainw->files[filenum]->frames,mainw->files[filenum]->fps,mt->fps)/mt->fps,in_end_range=out_val-1./mt->fps,real_out_end_range=out_end_range,real_in_start_range=in_start_range;

   if (mt->insert_mode!=INSERT_MODE_OVERWRITE) {
     if (!block->end_anchored&&max_tc>-1&&(((max_tc-offset_start)/U_SEC+in_val)<out_end_range)) real_out_end_range=(max_tc-offset_start)/U_SEC+in_val;
     if (!block->start_anchored&&min_tc>-1&&(((min_tc-offset_start)/U_SEC+in_val)>in_start_range)) real_in_start_range=(min_tc-offset_start)/U_SEC+in_val;
     if (!block->start_anchored) out_end_range=real_out_end_range; 
     if (!block->end_anchored) in_start_range=real_in_start_range; 
   }

   if (block->end_anchored&&(out_val-in_val>out_start_range)) out_start_range=in_start_range+out_val-in_val;
   if (block->start_anchored&&(out_end_range-out_val+in_val)<in_end_range) in_end_range=out_end_range-out_val+in_val;

   gtk_spin_button_set_range(GTK_SPIN_BUTTON(mt->spinbutton_out), out_start_range, real_out_end_range);
   gtk_spin_button_set_range(GTK_SPIN_BUTTON(mt->spinbutton_in), real_in_start_range, in_end_range);

   // make sure next and prev blocks can't overwrite
   if (block->prev!=NULL) block->prev->max_tc=offset_start;
   if (block->next!=NULL) block->next->min_tc=get_event_timecode(block->end_event)+U_SEC/mt->fps;

}

static void update_in_image(lives_mt *mt) {
  GdkPixbuf *pixbuf,*thumb;
  track_rect *block=mt->block_selected;
  gint track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(block->eventbox),"layer_number"));
  gint filenum=get_frame_event_clip(block->start_event,track);
  gint frame_start=calc_variable_frame(filenum,block->offset_start/U_SEC);
  gint width=cfile->hsize;
  gint height=cfile->vsize;

  if (width>mt->poly_box->allocation.width/2) width=mt->poly_box->allocation.width/2;
  if (height>mt->poly_box->allocation.height-(block->ordered?mt->spinbutton_in->allocation.height:0)) height=mt->poly_box->allocation.height-(block->ordered?mt->spinbutton_in->allocation.height:0);
  thumb=make_thumb(filenum,width,height,frame_start);
  if ((pixbuf=gtk_image_get_pixbuf(GTK_IMAGE(mt->in_image)))!=NULL) gdk_pixbuf_unref(pixbuf);
  gtk_image_set_from_pixbuf (GTK_IMAGE(mt->in_image),thumb);
}

static void update_out_image(lives_mt *mt, gint64 end_tc) {
  GdkPixbuf *pixbuf,*thumb;
  track_rect *block=mt->block_selected;
  gint track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(block->eventbox),"layer_number"));
  gint filenum=get_frame_event_clip(block->start_event,track);
  gint frame_end=calc_variable_frame(filenum,end_tc/U_SEC-1./mt->fps);
  gint width=cfile->hsize;
  gint height=cfile->vsize;
  
  if (width>mt->poly_box->allocation.width/2) width=mt->poly_box->allocation.width/2;
  if (height>mt->poly_box->allocation.height-(block->ordered?mt->spinbutton_out->allocation.height:0)) height=mt->poly_box->allocation.height-(block->ordered?mt->spinbutton_out->allocation.height:0);
  thumb=make_thumb(filenum,width,height,frame_end);
  if ((pixbuf=gtk_image_get_pixbuf(GTK_IMAGE(mt->out_image)))!=NULL) gdk_pixbuf_unref(pixbuf);
  gtk_image_set_from_pixbuf (GTK_IMAGE(mt->out_image),thumb);
}


static void show_preview (lives_mt *mt, weed_timecode_t tc) {
  // *** TODO - for STATEFUL effects we must store keyframes
  weed_timecode_t start_tc=get_event_timecode(mt->block_selected->start_event);
  weed_plant_t *layer=weed_plant_new(WEED_PLANT_CHANNEL);
  weed_plant_t *layers[]={layer,NULL};
  weed_plant_t *event=mt->block_selected->start_event;
  gint track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(mt->block_selected->eventbox),"layer_number"));
  int filenum=get_frame_event_clip(event,track);
  int frame=get_frame_event_frame(event,track);
  GdkPixbuf *pbuf,*pixbuf;
  int error;
  
  if (!mt->mt_frame_preview) return;

  weed_set_int_value(layer,"clip",filenum);
  weed_set_int_value(layer,"frame",frame);
  weed_set_int_value(layer,"current_palette",WEED_PALETTE_RGB24);

  pull_frame(layer,prefs->image_ext,tc);
  weed_apply_instance(mt->current_rfx->source,NULL,layers,weed_get_int_value(layer,"width",&error),weed_get_int_value(layer,"height",&error),tc-start_tc);
  pixbuf=layer_to_pixbuf(layer);
  pbuf=gdk_pixbuf_scale_simple(pixbuf,MT_PLAY_WIDTH,MT_PLAY_HEIGHT,GDK_INTERP_HYPER);

  gtk_image_set_from_pixbuf(GTK_IMAGE(mainw->image274),pbuf);

  gdk_pixbuf_unref(pixbuf);
  gdk_pixbuf_unref(pbuf);
  weed_plant_free(layer);
}


void
in_out_start_changed (GtkWidget *widget, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  track_rect *block=mt->block_selected;
  gdouble new_start=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
  weed_plant_t *event=block->start_event;
  weed_timecode_t new_start_tc,orig_start_tc=block->offset_start,offset_end,tl_start;
  gint track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(block->eventbox),"layer_number"));
  gint filenum;
  weed_plant_t *start_event=NULL,*event_next;
  gboolean was_moved;
  weed_timecode_t new_tl_tc;

  new_start_tc=q_dbl(new_start,mt->fps);

  if (new_start_tc==orig_start_tc||!block->ordered) {
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_in),new_start_tc/U_SEC);
    return;
  }

  tl_start=get_event_timecode(event);
  filenum=get_frame_event_clip(block->start_event,track);

  if (!block->start_anchored) {
    if (new_start_tc>block->offset_start) {
      start_event=get_prev_frame_event(block->start_event);

      // start increased, not anchored
      while (event!=NULL) {
	if ((get_event_timecode(event)-tl_start)>=new_start_tc-block->offset_start) {
	  // done
	  block->start_event=event;
	  block->offset_start=new_start_tc;
	  break;
	}
	remove_frame_from_event(mt->event_list,event,track);
	event=get_next_frame_event(event);
      }

      // move filter_inits right, and deinits left
      new_tl_tc=get_event_timecode(block->start_event);
      if (start_event==NULL) event=get_first_event(mt->event_list);
      else event=get_next_event(start_event);
      while (event!=NULL&&get_event_timecode(event)<tl_start) event=get_next_event(event);

      while (event!=NULL&&get_event_timecode(event)<new_tl_tc) {
	event_next=get_next_event(event);
	was_moved=FALSE;
	if (WEED_EVENT_IS_FILTER_INIT(event)) {
	  if (!move_event_right(mt->event_list,event,TRUE,mt->fps)) {
	    was_moved=TRUE;
	    if (event==start_event) start_event=NULL;
	  }
	}
	else {
	  if (WEED_EVENT_IS_FILTER_DEINIT(event)) {
	    if (!move_event_left(mt->event_list,event,TRUE,mt->fps)) {
	      was_moved=TRUE;
	    }
	  }
	}
	if (was_moved) {
	  if (start_event==NULL) event=get_first_event(mt->event_list);
	  else event=get_next_event(start_event);
	  while (event!=NULL&&get_event_timecode(event)<tl_start) event=get_next_event(event);
	}
	else {
	  event=event_next;
	  if (WEED_EVENT_IS_FRAME(event)) start_event=event;
	}
      }
    }
    else {
      // do an insert from offset_start down
      gint filenum=get_frame_event_clip(block->start_event,track);
      insert_frames (filenum,block->offset_start,new_start_tc,tl_start,DIRECTION_NEGATIVE,block->eventbox,mt,block);

      // any filter_inits with this track as owner get moved as far left as possible
      new_tl_tc=get_event_timecode(block->start_event);
      event=block->start_event;

      while (event!=NULL&&get_event_timecode(event)<tl_start) {
	start_event=event;
	event=get_next_event(event);
      }
      while (event!=NULL&&get_event_timecode(event)==tl_start) {
	if (WEED_EVENT_IS_FILTER_INIT(event)) {
	  if (filter_init_has_owner(event,track)) {
	    // candidate for moving
	    move_filter_init_event(mt->event_list,new_tl_tc,event,mt->fps);
	    // account for overlaps
	    move_event_right(mt->event_list,event,TRUE,mt->fps);
	    event=start_event;
	    while (event!=NULL&&get_event_timecode(event)<tl_start) event=get_next_event(event);
	  }
	}
	else event=get_next_event(event);
      }
    }
  }
  else {
    // start is anchored, do a re-insert from start to end
    gshort insert_mode=mt->insert_mode;
    offset_end=q_gint64((block->offset_start=new_start_tc)+(gint64)(U_SEC/mt->fps)+(get_event_timecode(block->end_event)-get_event_timecode(block->start_event)),mt->fps);
    mt->insert_mode=INSERT_MODE_OVERWRITE;
    insert_frames (filenum,new_start_tc,offset_end,tl_start,DIRECTION_POSITIVE,block->eventbox,mt,block);
    mt->insert_mode=insert_mode;
  }

  offset_end=q_gint64((new_start_tc=block->offset_start)+(gint64)(U_SEC/mt->fps)+(get_event_timecode(block->end_event)-get_event_timecode(block->start_event)),mt->fps);

  if (mt->poly_state==POLY_IN_OUT) {
    set_in_out_spin_ranges(mt,new_start_tc,offset_end);
  
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_out),offset_end/U_SEC);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_in),new_start_tc/U_SEC);

    // update images
    update_in_image(mt);
    if (block->start_anchored) update_out_image(mt,offset_end);
  }

  if (!resize_timeline(mt)) gtk_widget_queue_draw (block->eventbox); // redraw the track
}





void
in_out_end_changed (GtkWidget *widget, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  track_rect *block=mt->block_selected;
  gdouble new_end=gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));
  weed_timecode_t offset_end=block->offset_start+(gint64)(U_SEC/mt->fps)+(get_event_timecode(block->end_event)-get_event_timecode(block->start_event)),orig_end_tc;
  weed_plant_t *event=block->end_event,*prevevent;
  gint track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(block->eventbox),"layer_number"));
;
  gint filenum=get_frame_event_clip(block->end_event,track);
  weed_timecode_t new_end_tc,tl_end;
  weed_plant_t *start_event,*event_next,*init_event;
  gboolean was_moved;
  weed_timecode_t new_tl_tc;
  int error;

  if (new_end>mainw->files[filenum]->frames/mainw->files[filenum]->fps) new_end=mainw->files[filenum]->frames/mainw->files[filenum]->fps;

  new_end_tc=q_dbl(new_end,mt->fps);
  orig_end_tc=offset_end=q_gint64(offset_end,mt->fps);

  if (new_end_tc==orig_end_tc||!block->ordered) {
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_out),new_end_tc/U_SEC);
    return;
  }

  tl_end=get_event_timecode(event);
  start_event=get_prev_frame_event(event);

  if (!block->end_anchored) {
    if (new_end_tc<offset_end) {
      // end decreased, not anchored

      while (event!=NULL) {
	if ((tl_end-get_event_timecode(event))>=(offset_end-new_end_tc)) {
	  // done
	  block->end_event=event;
	  break;
	}
	prevevent=get_prev_frame_event(event);
	remove_frame_from_event(mt->event_list,event,track);
	event=prevevent;
      }

      // move filter_inits right, and deinits left
      new_tl_tc=get_event_timecode(block->end_event);
      if (start_event==NULL) event=get_first_event(mt->event_list);
      else event=get_next_event(start_event);
      while (event!=NULL&&get_event_timecode(event)<=new_tl_tc) event=get_next_event(event);

      while (event!=NULL&&get_event_timecode(event)<=tl_end) {
	event_next=get_next_event(event);
	was_moved=FALSE;
	if (WEED_EVENT_IS_FILTER_INIT(event)) {
	  if (!move_event_right(mt->event_list,event,TRUE,mt->fps)) {
	    was_moved=TRUE;
	    if (event==start_event) start_event=NULL;
	  }
	}
	else {
	  if (WEED_EVENT_IS_FILTER_DEINIT(event)) {
	    if (!move_event_left(mt->event_list,event,TRUE,mt->fps)) {
	      was_moved=TRUE;
	    }
	  }
	}
	if (was_moved) {
	  if (start_event==NULL) event=get_first_event(mt->event_list);
	  else event=get_next_event(start_event);
	  while (event!=NULL&&get_event_timecode(event)<=new_tl_tc) event=get_next_event(event);
	}
	else {
	  event=event_next;
	  if (WEED_EVENT_IS_FRAME(event)) start_event=event;
	}
      }
      remove_end_blank_frames(mt->event_list);
    }
    else {
      // end increased, not anchored
      // do an insert from end_tc up, starting with end_frame and finishing at new_end
      insert_frames (filenum,offset_end,new_end_tc,tl_end+(gint64)(U_SEC/mt->fps),DIRECTION_POSITIVE,block->eventbox,mt,block);

      new_tl_tc=get_event_timecode(block->end_event);

      start_event=event;

      while (event!=NULL&&get_event_timecode(event)==tl_end) {
	if (WEED_EVENT_IS_FILTER_DEINIT(event)) {
	  init_event=weed_get_voidptr_value(event,"init_event",&error);
	  if (filter_init_has_owner(init_event,track)) {
	    // candidate for moving
	    move_filter_deinit_event(mt->event_list,new_tl_tc,event,mt->fps);
	    // account for overlaps
	    //move_event_left(mt->event_list,event,TRUE,mt->fps);
	    event=start_event;
	  }
	}
	else event=get_next_event(event);
      }
    }
  }
  else {
    // end is anchored, do a re-insert from end to start
    gint64 offset_start;
    gshort insert_mode=mt->insert_mode;

    offset_end=q_gint64((offset_start=block->offset_start+new_end_tc-orig_end_tc)+(gint64)(U_SEC/mt->fps)+(get_event_timecode(block->end_event)-get_event_timecode(block->start_event)),mt->fps);
    mt->insert_mode=INSERT_MODE_OVERWRITE;
    insert_frames (filenum,offset_end,offset_start,tl_end+(gint64)(U_SEC/mt->fps),DIRECTION_NEGATIVE,block->eventbox,mt,block);
    mt->insert_mode=insert_mode;
  }

  // get new offset_end
  new_end_tc=q_gint64(block->offset_start+(gint64)(U_SEC/mt->fps)+(get_event_timecode(block->end_event)-get_event_timecode(block->start_event)),mt->fps);

  if (mt->poly_state==POLY_IN_OUT) {
    set_in_out_spin_ranges(mt,block->offset_start,new_end_tc);

    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_in), block->offset_start/U_SEC);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_out), new_end_tc/U_SEC);

    // update image
    update_out_image(mt,new_end_tc);
    if (block->end_anchored) update_in_image(mt);
  }

  resize_timeline(mt);

  if (!resize_timeline(mt)) gtk_widget_queue_draw (block->eventbox); // redraw the track
}

void
in_anchor_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  track_rect *block=mt->block_selected;
  gint64 offset_end=block->offset_start+(gint64)(U_SEC/mt->fps)+(get_event_timecode(block->end_event)-get_event_timecode(block->start_event));

  block->start_anchored=!block->start_anchored;
  set_in_out_spin_ranges(mt,block->offset_start,offset_end);
}

void
out_anchor_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  track_rect *block=mt->block_selected;
  gint64 offset_end=block->offset_start+(gint64)(U_SEC/mt->fps)+(get_event_timecode(block->end_event)-get_event_timecode(block->start_event));

  block->end_anchored=!block->end_anchored;
  set_in_out_spin_ranges(mt,block->offset_start,offset_end);
}




void polymorph (lives_mt *mt, gshort poly) {
  GdkPixbuf *thumb,*pixbuf;
  gint width,height,track;
  gint frame_start,frame_end;
  int filenum;
  track_rect *block=mt->block_selected;
  gint64 offset_end;
  weed_plant_t *filter;
  GtkWidget *hbox;
  GtkWidget *button;
  GtkWidget *label;
  GtkWidget *scale;
  GtkObject *spinbutton_adj;
  gdouble start_time,end_time;
  gboolean sepwin=mainw->sep_win;
  
  if (poly==mt->poly_state&&poly!=POLY_EFFECT) return;

  if (poly==POLY_IN_OUT&&block==NULL) return;

  switch (mt->poly_state) {
  case (POLY_CLIPS) :
    gtk_container_remove (GTK_CONTAINER(mt->poly_box),mt->clip_scroll);
    break;
  case (POLY_IN_OUT) :
    gtk_container_remove (GTK_CONTAINER(mt->poly_box),mt->in_out_box);
    break;
  case (POLY_EFFECT) :

    rfx_free(mt->current_rfx);
    mt->current_rfx=NULL;
    gtk_widget_destroy(mt->fx_box);

    if (mt->mt_frame_preview) {
      // put blank back in preview window
      gtk_widget_hide(mt->preview_button);
      gtk_widget_show(mt->sel_label);
      gtk_widget_ref (mainw->playarea);
      if (mainw->plug1!=NULL) {
	gtk_container_remove (GTK_CONTAINER(mainw->plug1),mainw->image274);
	gtk_widget_destroy (mainw->plug1);
	mainw->plug1=NULL;
      }
      
      if (mainw->playarea->parent!=NULL) gtk_container_remove (GTK_CONTAINER(mainw->playarea->parent),mainw->playarea);
      gtk_container_add (GTK_CONTAINER(mt->play_box),mt->play_blank);
      gtk_container_add (GTK_CONTAINER (mainw->playframe), mainw->playarea);
      gtk_widget_unref (mainw->playarea);
      gtk_widget_unref(mt->play_blank);
      
      mainw->xwin=0;
      mt->mt_frame_preview=FALSE;
    }
    if (pchain!=NULL) g_free(pchain);
    mouse_mode_context(mt); // reset context box text
    break;
  }

  mt->poly_state=poly;

  switch (poly) {
  case (POLY_IN_OUT) :
    if (block->ordered) {
      gtk_widget_show(mt->in_hbox);
      gtk_widget_show(mt->out_hbox);
    }
    else {
      gtk_widget_hide(mt->in_hbox);
      gtk_widget_hide(mt->out_hbox);
    }

    track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(block->eventbox),"layer_number"));
    filenum=get_frame_event_clip(block->start_event,track);
    offset_end=block->offset_start+(gint64)(U_SEC/mt->fps)+(get_event_timecode(block->end_event)-get_event_timecode(block->start_event));
    width=cfile->hsize;
    height=cfile->vsize;
    if (width>mt->poly_box->allocation.width/2) width=mt->poly_box->allocation.width/2;
    if (height>mt->poly_box->allocation.height-(block->ordered?mainw->spinbutton_start->allocation.height:0)) height=mt->poly_box->allocation.height-(block->ordered?mainw->spinbutton_start->allocation.height:0);

    if (mainw->playing_file==filenum) {
      // we must do this, because during playback with an event_list, cfile->event_list holds a pointer to our mt->event_list
      // our resampled frames will have been moved to cfile->event_list_back
      // this is definately NOT good coding style
      mainw->files[filenum]->event_list=mainw->files[filenum]->event_list_back;
    }
    frame_start=calc_variable_frame(filenum,block->offset_start/U_SEC);
    frame_end=calc_variable_frame(filenum,offset_end/U_SEC-1./mt->fps);
    if (mainw->playing_file==filenum) {
      mainw->files[filenum]->event_list=mt->event_list;
    }

    // start image
    thumb=make_thumb(filenum,width,height,frame_start);
    if ((pixbuf=gtk_image_get_pixbuf(GTK_IMAGE(mt->in_image)))!=NULL) gdk_pixbuf_unref(pixbuf);
    gtk_image_set_from_pixbuf (GTK_IMAGE(mt->in_image),thumb);

    g_signal_handler_block (mt->spinbutton_in,mt->spin_in_func);
    gtk_spin_button_set_range(GTK_SPIN_BUTTON(mt->spinbutton_in),0., offset_end/U_SEC-1./mt->fps);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_in),block->offset_start/U_SEC);
    g_signal_handler_unblock (mt->spinbutton_in,mt->spin_in_func);
    g_signal_handler_block (mt->checkbutton_start_anchored,mt->check_start_func);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mt->checkbutton_start_anchored),block->start_anchored);
    g_signal_handler_unblock (mt->checkbutton_start_anchored,mt->check_start_func);

    // end image
    thumb=make_thumb(filenum,width,height,frame_end);
    if ((pixbuf=gtk_image_get_pixbuf(GTK_IMAGE(mt->out_image)))!=NULL) gdk_pixbuf_unref(pixbuf);
    gtk_image_set_from_pixbuf (GTK_IMAGE(mt->out_image),thumb);

    g_signal_handler_block (mt->spinbutton_out,mt->spin_out_func);
    gtk_spin_button_set_range(GTK_SPIN_BUTTON(mt->spinbutton_out), block->offset_start/U_SEC+1./mt->fps, count_resampled_frames(mainw->files[filenum]->frames,mainw->files[filenum]->fps,mt->fps)/mt->fps);
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_out),offset_end/U_SEC);
    g_signal_handler_unblock (mt->spinbutton_out,mt->spin_out_func);

    g_signal_handler_block (mt->checkbutton_end_anchored,mt->check_end_func);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(mt->checkbutton_end_anchored),block->end_anchored);
    g_signal_handler_unblock (mt->checkbutton_end_anchored,mt->check_end_func);
    gtk_box_pack_start(GTK_BOX(mt->poly_box),mt->in_out_box,TRUE,TRUE,0);

    if (mainw->playing_file>-1) mt_desensitise(mt);
    else mt_sensitise (mt);

    break;
  case (POLY_CLIPS) :
    gtk_box_pack_start(GTK_BOX(mt->poly_box),mt->clip_scroll,TRUE,TRUE,0);
    break;
  case (POLY_EFFECT):

    start_time=get_event_timecode(mt->block_selected->start_event)/U_SEC;
    end_time=get_event_timecode(mt->block_selected->end_event)/U_SEC;

    filter=get_weed_filter(mt->current_fx);
    
    if (fx_dialog[1]!=NULL) {
      lives_rendered_effect_t *rfx=g_object_get_data (G_OBJECT (fx_dialog[1]),"rfx");
      gtk_widget_destroy(fx_dialog[1]);
      on_paramwindow_cancel_clicked2(NULL,rfx);
    }
    mt->fx_box=gtk_vbox_new(FALSE,0);
    gtk_box_pack_start(GTK_BOX(mt->poly_box),mt->fx_box,TRUE,TRUE,0);

    label=gtk_label_new(weed_filter_get_name(mt->current_fx));
    gtk_box_pack_start (GTK_BOX (mt->fx_box), label, FALSE, FALSE, 0);
    mt->current_rfx=weed_to_rfx(filter,FALSE);
    make_param_box(GTK_VBOX (mt->fx_box), mt->current_rfx); // *** TODO - hide params which need reinit
    
    hbox=gtk_hbox_new(FALSE,0);
    gtk_box_pack_end (GTK_BOX (mt->fx_box), hbox, FALSE, FALSE, 0);
    button = gtk_button_new_with_mnemonic (_("_Set value(s) at time"));
    gtk_box_pack_start (GTK_BOX (hbox), button, FALSE, FALSE, 0);
    
    g_signal_connect (GTK_OBJECT (button), "clicked",
		      G_CALLBACK (on_set_pvals_clicked),
		      (gpointer)mt);

    spinbutton_adj = gtk_adjustment_new (start_time, start_time, end_time, 1./cfile->fps, 10./cfile->fps, 0.);
    mt->node_spinbutton = gtk_spin_button_new (GTK_ADJUSTMENT (spinbutton_adj), 0, 3);
    gtk_widget_show (mt->node_spinbutton);
    gtk_box_pack_start (GTK_BOX (hbox), mt->node_spinbutton, FALSE, TRUE, 0);
    
    g_signal_connect_after (GTK_OBJECT (mt->node_spinbutton), "value_changed",
			    G_CALLBACK (on_node_spin_value_changed),
			    (gpointer)mt);

    scale=gtk_hscale_new(GTK_ADJUSTMENT(spinbutton_adj));
    gtk_box_pack_start (GTK_BOX (hbox), scale, TRUE, TRUE, 10);
    gtk_scale_set_draw_value(GTK_SCALE(scale),FALSE);
    
    mt->prev_node_button = gtk_button_new_with_mnemonic (_("_Prev node"));
    gtk_box_pack_start (GTK_BOX (hbox), mt->prev_node_button, FALSE, FALSE, 0);
    gtk_widget_set_sensitive(mt->prev_node_button,FALSE);
    
    g_signal_connect (GTK_OBJECT (mt->prev_node_button), "clicked",
		      G_CALLBACK (on_prev_node_clicked),
		      (gpointer)mt);

    mt->next_node_button = gtk_button_new_with_mnemonic (_("_Next node"));
    gtk_box_pack_start (GTK_BOX (hbox), mt->next_node_button, FALSE, FALSE, 0);
    gtk_widget_set_sensitive(mt->next_node_button,FALSE);
    
    g_signal_connect (GTK_OBJECT (mt->next_node_button), "clicked",
		      G_CALLBACK (on_next_node_clicked),
		      (gpointer)mt);

    mt->del_node_button = gtk_button_new_with_mnemonic (_("_Delete node"));
    gtk_box_pack_start (GTK_BOX (hbox), mt->del_node_button, FALSE, FALSE, 0);
    gtk_widget_set_sensitive(mt->del_node_button,FALSE);

    g_signal_connect (GTK_OBJECT (mt->del_node_button), "clicked",
		      G_CALLBACK (on_del_node_clicked),
		      (gpointer)mt);

    gtk_widget_show_all(mt->fx_box);

    if (mainw->playing_file<0) {
      // prepare internal play window for previews
      gtk_widget_hide(mt->sel_label);
      gtk_widget_show(mt->preview_button);
      gtk_widget_ref(mt->play_blank);
      gtk_container_remove (GTK_CONTAINER(mt->play_box),mt->play_blank);
      gtk_widget_ref (mainw->playarea);
      if (mainw->plug1!=NULL) {
	gtk_container_remove (GTK_CONTAINER(mainw->plug1),mainw->image274);
	gtk_widget_destroy (mainw->plug1);
	mainw->plug1=NULL;
      }
      mainw->xwin=0;
      if (mainw->playarea->parent!=NULL) gtk_container_remove (GTK_CONTAINER(mainw->playarea->parent),mainw->playarea);
      gtk_box_pack_start (GTK_BOX (mt->play_box), mainw->playarea, TRUE, TRUE, 0);
      gtk_widget_unref (mainw->playarea);
      mainw->sep_win=FALSE; // force showing in internal window
      add_to_playframe();
      mainw->sep_win=sepwin;
      mt->mt_frame_preview=TRUE;
      show_preview(mt,q_dbl(start_time,mt->fps));
    }
    clear_context(mt);
    add_context_label(mt,_("Drag the time slider to where you want\n"));
    add_context_label(mt,_("to set effect parameters\n"));
    add_context_label(mt,_("Set parameters, and then click\n"));
    add_context_label(mt,_("\"Set value(s) at time\"\n"));
    add_context_label(mt,_("You can preview by clicking on\n"));
    add_context_label(mt,_("\"Show Frame Preview\" below.\n"));
    break;
  }
  gtk_widget_queue_draw(mt->poly_box);
}


static void mouse_select_start(GtkWidget *widget, lives_mt *mt) {
  gdouble timesecs;
  gint min_x;
  gtk_widget_set_sensitive(mt->mm_menuitem,FALSE);
  
  gdk_window_get_pointer(GDK_WINDOW (mt->timeline->window), &mt->sel_x, &mt->sel_y, NULL);
  timesecs=(gdouble)mt->sel_x/(gdouble)mt->timeline->allocation.width*mt->end_secs;
  if (timesecs>mt->end_secs) timesecs=mt->end_secs;

  timesecs=(gint)(timesecs*mt->fps+.5)/mt->fps; // quantise to frame timing

  if (timesecs<0.) timesecs=0.;
  mt->region_start=mt->region_end=timesecs;

  mt->region_updating=TRUE;
  on_timeline_update(mt->timeline_eb,NULL,mt);
  mt->region_updating=FALSE;

  gdk_window_get_pointer(GDK_WINDOW (mt->tl_eventbox->window), &mt->sel_x, &mt->sel_y, NULL);
  gdk_window_get_position(GDK_WINDOW (mt->timeline_eb->window), &min_x, NULL);

  if (mt->sel_x<min_x) mt->sel_x=min_x;
  if (mt->sel_y<0.) mt->sel_y=0.;

  gtk_widget_queue_draw(mt->tl_hbox);
  gtk_widget_queue_draw(mt->timeline);

  mt->tl_selecting=TRUE;

}

static void mouse_select_end(GtkWidget *widget, lives_mt *mt) {
  mt->tl_selecting=FALSE;
  gdk_window_get_pointer(GDK_WINDOW (mt->timeline->window), &mt->sel_x, &mt->sel_y, NULL);
  gtk_widget_set_sensitive(mt->mm_menuitem,TRUE);
  gtk_widget_queue_draw(mt->tl_eventbox);
  on_timeline_release(mt->timeline_eb,NULL,mt);
}


static void mouse_select_move(GtkWidget *widget, lives_mt *mt) {
  gint x,y;
  gint start_x,start_y,width,height;
  gint current_track=mt->current_track;
  int i;
  GtkWidget *xeventbox;
  gint rel_x,rel_y,min_x;
  gint offs_y_start,offs_y_end,xheight;
  GtkWidget *checkbutton;

  if (mt->block_selected!=NULL) unselect_all(mt);

  gdk_window_get_pointer(GDK_WINDOW (mt->tl_eventbox->window), &x, &y, NULL);
  gdk_window_get_position(GDK_WINDOW (mt->timeline_eb->window), &min_x, NULL);

  if (x<min_x) x=min_x;
  if (y<0.) y=0.;

  gtk_widget_queue_draw(mt->tl_hbox);
  gdk_window_process_updates(mt->tl_eventbox->window,FALSE);

  if (x>=mt->sel_x) {
    start_x=mt->sel_x;
    width=x-mt->sel_x;
  }
  else {
    start_x=x;
    width=mt->sel_x-x;
  }
  if (y>=mt->sel_y) {
    start_y=mt->sel_y;
    height=y-mt->sel_y;
  }
  else {
    start_y=y;
    height=mt->sel_y-y;
  }

  if (start_x<0) start_x=0;
  if (start_y<0) start_y=0;

  gdk_draw_rectangle (GDK_DRAWABLE(mt->tl_eventbox->window), mainw->LiVES->style->black_gc, TRUE, start_x, start_y, width, height);

  for (i=0;i<mt->num_video_tracks;i++) {
    xeventbox=g_list_nth_data(mt->video_draws,i);
    if (!GPOINTER_TO_INT(g_object_get_data(G_OBJECT(xeventbox),"hidden"))) {
      xheight=xeventbox->allocation.height;
      checkbutton=g_object_get_data(G_OBJECT(xeventbox),"checkbutton");
      gdk_window_get_position(xeventbox->window,&rel_x,&rel_y);
      if (start_y>(rel_y+xheight/2)||(start_y+height)<(rel_y+xheight/2)) {
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton))) {
	  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),FALSE);
	  mt->current_track=current_track;
	  track_select(mt);
	}
	continue;
      }
      offs_y_start=0;
      offs_y_end=xheight;
      if (start_y<rel_y+xheight) {
	offs_y_start=start_y-rel_y;
	gdk_draw_line (xeventbox->window, mainw->LiVES->style->black_gc, start_x-rel_x, offs_y_start, start_x+width-rel_x-1, offs_y_start);
      }
      if (start_y+height<rel_y+xheight) {
	offs_y_end=start_y-rel_y+height;
	gdk_draw_line (xeventbox->window, mainw->LiVES->style->black_gc, start_x-rel_x, offs_y_end, start_x+width-rel_x-1, offs_y_end);
      }
      gdk_draw_line (xeventbox->window, mainw->LiVES->style->black_gc, start_x-rel_x, offs_y_start, start_x-rel_x, offs_y_end);
      gdk_draw_line (xeventbox->window, mainw->LiVES->style->black_gc, start_x+width-rel_x-1, offs_y_start, start_x+width-rel_x-1, offs_y_end);

      if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton))) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),TRUE);
	mt->current_track=current_track;
	track_select(mt);
      }
    }
  }

  if (widget!=mt->timeline_eb) {
    mt->region_updating=TRUE;
    on_timeline_update(mt->timeline_eb,NULL,mt);
    mt->region_updating=FALSE;
  }
  
}

void do_block_context (lives_mt *mt, GdkEventButton *event, track_rect *block) {
  // pop up a context menu when a selected block is right clicked on

  // unfinished...

  GtkWidget *delete_block;
  GtkWidget *split_here;
  GtkWidget *menu=gtk_menu_new();

  mouse_select_end(NULL,mt);

  gtk_menu_set_title (GTK_MENU(menu),_("LiVES: Selected block"));

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }
  
  delete_block = gtk_menu_item_new_with_mnemonic (_("_Delete this block"));
  gtk_container_add (GTK_CONTAINER (menu), delete_block);

  g_signal_connect (GTK_OBJECT (delete_block), "activate",
		    G_CALLBACK (delete_block_cb),
		    (gpointer)mt);
  
  if (block->ordered) { // TODO
    split_here = gtk_menu_item_new_with_mnemonic (_("_Split block here"));
    gtk_container_add (GTK_CONTAINER (menu), split_here);
    
    g_signal_connect (GTK_OBJECT (split_here), "activate",
		      G_CALLBACK (on_split_activate),
		      (gpointer)mt);
  }
  gtk_widget_show_all (menu);
  
  gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);


}

void do_track_context (lives_mt *mt, GdkEventButton *event, gdouble timesecs, gint track) {
  // pop up a context menu when track is right clicked on

  // unfinished...

  GtkWidget *insert_here;
  GtkWidget *menu=gtk_menu_new();

  mouse_select_end(NULL,mt);

  gtk_menu_set_title (GTK_MENU(menu),_("LiVES: Selected frame"));

  if (palette->style&STYLE_1) {
    gtk_widget_modify_bg(menu, GTK_STATE_NORMAL, &palette->menu_and_bars);
  }
  
  insert_here = gtk_menu_item_new_with_mnemonic (_("_Insert here"));
  gtk_container_add (GTK_CONTAINER (menu), insert_here);
  
  gtk_widget_show_all (menu);
  
  gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);

  g_signal_connect (GTK_OBJECT (insert_here), "activate",
		    G_CALLBACK (insert_here_cb),
		    (gpointer)mt);
  
}


gboolean on_track_release (GtkWidget *eventbox, GdkEventButton *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  gint x,y;
  gdouble timesecs;
  gint track=-1;
  GtkWidget *xeventbox;
  GdkWindow *window;
  gint win_x,win_y;
  //int filenum=-1;
  int i;
  gint old_track=mt->current_track;

  gdk_window_get_pointer(GDK_WINDOW (eventbox->window), &x, &y, NULL);
  if (x<0) x=0;

  window=gdk_display_get_window_at_pointer (mt->display,&win_x,&win_y);

  for (i=0;i<g_list_length(mt->video_draws);i++) {
    if (GTK_WIDGET((xeventbox=g_list_nth_data(mt->video_draws,i)))->window==window) {
      track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(xeventbox),"layer_number"));
    }
  }

  if (mt->mouse_mode==MOUSE_MODE_SELECT) {
    mouse_select_end(eventbox,mt);
    g_signal_handler_block (mt->tl_eventbox,mt->mouse_mot2);
    mt->sel_y-=y+2;
  }
  else {
    if (track!=-1&&mt->putative_block!=NULL&&mainw->playing_file==-1&&event->button==1) { // TODO **
      weed_timecode_t start_tc;
      mt_desensitise(mt);
      mt->move_effects=gtk_check_menu_item_get_active (GTK_CHECK_MENU_ITEM (mt->move_fx));
      timesecs=(gdouble)x/(gdouble)eventbox->allocation.width*mt->end_secs;
      if (timesecs>mt->end_secs) timesecs=mt->end_secs;
      timesecs=(gint)(timesecs*mt->fps)/mt->fps; // quantise to frame timing
      if (timesecs<0.) timesecs=0.;

      GTK_RULER (mt->timeline)->position=timesecs;
      gtk_widget_queue_draw (mainw->multitrack->timeline);
      start_tc=get_event_timecode(mt->putative_block->start_event);
      
      if (track!=mt->current_track||q_gint64(timesecs*U_SEC,mt->fps)!=q_gint64(start_tc,mt->fps)) {
	move_block(mt,mt->putative_block,timesecs,old_track,track);
      }
    }
    if (mt->hotspot_x!=0||mt->hotspot_y!=0) {
      GdkScreen *screen;
      gint abs_x,abs_y;
      gint height=GTK_WIDGET(g_list_nth_data(mt->video_draws,0))->allocation.height;
      mt_desensitise(mt);
      gdk_display_get_pointer(mt->display,&screen,&abs_x,&abs_y,NULL);
      gdk_display_warp_pointer(mt->display,screen,abs_x+mt->hotspot_x,abs_y+mt->hotspot_y-height/2);
      gdk_window_get_pointer(GDK_WINDOW (eventbox->window), &x, NULL, NULL);
      timesecs=(gdouble)x/(gdouble)eventbox->allocation.width*mt->end_secs;
      timesecs=(gint)(timesecs*mt->fps+.5)/mt->fps; // quantise to frame timing
      GTK_RULER (mt->timeline)->position=timesecs;
      gtk_widget_queue_draw (mainw->multitrack->timeline);
    }
  }
  
  mt_sensitise(mt);
  mt->hotspot_x=mt->hotspot_y=0;
  mt->putative_block=NULL;

  set_cursor_style(mt,LIVES_CURSOR_NORMAL,0,0,0,0,0);
  gtk_widget_set_sensitive(mt->mm_menuitem,TRUE);
  return TRUE;
}




gboolean on_track_header_click (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  if (mt->mouse_mode==MOUSE_MODE_SELECT) {
    mouse_select_start(widget,mt);
    g_signal_handler_unblock (widget,mt->mouse_mot1);
  }
  return TRUE;
}

gboolean on_track_header_release (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  if (mt->mouse_mode==MOUSE_MODE_SELECT) {
    mouse_select_end(widget,mt);
    g_signal_handler_block (widget,mt->mouse_mot1);
  }
  return TRUE;
}

gboolean on_track_between_click (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  if (mt->mouse_mode==MOUSE_MODE_SELECT) {
    mouse_select_start(widget,mt);
    g_signal_handler_unblock (mt->tl_eventbox,mt->mouse_mot2);
  }
  return TRUE;
}

gboolean on_track_between_release (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  if (mt->mouse_mode==MOUSE_MODE_SELECT) {
    mouse_select_end(widget,mt);
    g_signal_handler_block (mt->tl_eventbox,mt->mouse_mot2);
  }
  return TRUE;
}


gboolean on_track_click (GtkWidget *eventbox, GdkEventButton *event, gpointer user_data) {
  gint x,y;
  track_rect *block;
  lives_mt *mt=(lives_mt *)user_data;
  gdouble timesecs;
  gint track;
  int filenum=-1;

  gtk_widget_set_sensitive(mt->mm_menuitem,FALSE);

  gdk_window_get_pointer(GDK_WINDOW (eventbox->window), &x, &y, NULL);
  timesecs=(gdouble)x/(gdouble)eventbox->allocation.width*mt->end_secs;
  timesecs=(gint)(timesecs*mt->fps+.5)/mt->fps; // quantise to frame timing

  track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"layer_number"));
  block=mt->putative_block=get_block_from_time(eventbox,timesecs,mt);

  unselect_all(mt);

  if (mt->mouse_mode==MOUSE_MODE_SELECT&&event->type==GDK_BUTTON_PRESS) {
    mouse_select_start(eventbox,mt);
    g_signal_handler_unblock (mt->tl_eventbox,mt->mouse_mot2);
  }
  else {
    if (event->type!=GDK_BUTTON_PRESS) {
      if (block!=NULL) {
	// double click, select the rectangle
	filenum=get_frame_event_clip(block->start_event,track);
	block->state=BLOCK_SELECTED;
	mt->block_selected=block;
	clear_context(mt);
	add_context_label (mt,g_strdup_printf (_("Current track: %s (layer %d)\n"),g_object_get_data(G_OBJECT(eventbox),"track_name"),track));
	add_context_label (mt,g_strdup_printf (_("%.2f sec. to %.2f sec.\n"),get_event_timecode(block->start_event)/U_SEC,get_event_timecode(block->end_event)/U_SEC+1./mt->fps));
	add_context_label (mt,g_strdup_printf (_("Source: %s"),mainw->files[filenum]->name));
	add_context_label (mt,(_("Right click for context menu.\n")));
	add_context_label (mt,(_("Single click on timeline\nto select a frame.\n")));
	gtk_widget_set_sensitive(mt->view_in_out,TRUE);
	gtk_widget_set_sensitive (mt->fx_block, TRUE);
	draw_block(mt,block,mt->end_secs,0,0);
	multitrack_view_in_out(NULL,mt);
      }
    }
    else {
      // single click, TODO - locate the frame for the track in event_list
      clear_context(mt);
      add_context_label (mt,g_strdup_printf (_("Current track: %s (layer %d)\n"),g_object_get_data(G_OBJECT(eventbox),"track_name"),track));
      add_context_label (mt,g_strdup_printf (_("%.2f sec.\n"),timesecs));
      if (block!=NULL) {
	filenum=get_frame_event_clip(block->start_event,track);
	add_context_label (mt,g_strdup_printf (_("Source: %s"),mainw->files[filenum]->name));
	add_context_label (mt,(_("Right click for context menu.\n")));
      }
      add_context_label (mt,(_("Double click on a block to select it.\n")));
      
      if (mainw->playing_file==-1) {
	GTK_RULER (mt->timeline)->position=timesecs;
	gtk_widget_queue_draw (mainw->multitrack->timeline);
      }
      // TODO - list effects applied to frame
      
      if (block!=NULL&&filenum!=mainw->scrap_file) {
	mt->clip_selected=filenum-1;
	clip_select(mt);
      }
      else polymorph (mt,POLY_CLIPS);
      mt->region_start=mt->region_end=timesecs;
      
      if (block!=NULL&&event->button!=3) {
	gdouble start_secs,end_secs;
	
	GdkScreen *screen;
	gint abs_x,abs_y;
	
	gint ebwidth=GTK_WIDGET(mt->timeline)->allocation.width;
	
	gdouble width=((end_secs=(get_event_timecode(block->end_event)/U_SEC))-(start_secs=(get_event_timecode(block->start_event)/U_SEC))+1./mt->fps);
	gint height=GTK_WIDGET(g_list_nth_data(mt->video_draws,0))->allocation.height;
	
	width=(width/mt->end_secs*(gdouble)ebwidth);
	if (width>ebwidth) width=ebwidth;
	if (width<2) width=2;
	mt->hotspot_x=x-ebwidth*start_secs/mt->end_secs;
	mt->hotspot_y=y;
	gdk_display_get_pointer(mt->display,&screen,&abs_x,&abs_y,NULL);
	gdk_display_warp_pointer(mt->display,screen,abs_x-mt->hotspot_x,abs_y-y+height/2);
	set_cursor_style(mt,LIVES_CURSOR_BLOCK,width,height,filenum,0,height/2);
      }
    }
  }

  mt->current_track=track;
  track_select(mt);
  
  if (event->button==3&&mainw->playing_file==-1) {
    if (block!=NULL) {
      // context menu for a selected block
      mt->putative_block=block;
      do_block_context(mt,event,block);
      return TRUE;
    }
    else {
      do_track_context(mt,event,timesecs,track);
      return TRUE;
    }
  }

  return TRUE;
}

gboolean on_track_move (GtkWidget *widget, GdkEventMotion *event, gpointer user_data) {
  // used for mouse mode SELECT
  lives_mt *mt=(lives_mt *)user_data;
  if (mt->mouse_mode==MOUSE_MODE_SELECT) mouse_select_move(widget,mt);
  return TRUE;
}

gboolean on_track_header_move (GtkWidget *widget, GdkEventMotion *event, gpointer user_data) {
  // used for mouse mode SELECT
  lives_mt *mt=(lives_mt *)user_data;
  if (mt->mouse_mode==MOUSE_MODE_SELECT) mouse_select_move(widget,mt);
  return TRUE;
}


void animate_multitrack (lives_mt *mt) {
  // update timeline pointer(s)
  /*  gint offset=-1;
  int i;
  gint len=g_list_length (mainw->multitrack->video_draws); */
  gdouble currtime=(mainw->currticks-mainw->origticks)/U_SEC;
  time_to_string(mt,currtime,TIMECODE_LENGTH);

  GTK_RULER (mt->timeline)->position=currtime;
  gtk_widget_queue_draw (mt->timeline);

  // TODO - opt
  /*  for (i=0;i<len;i++) {
    GtkWidget *eventbox=g_list_nth_data(mainw->multitrack->video_draws,i);
    if (GTK_WIDGET_VISIBLE(eventbox)) {
      if (offset==-1) offset=currtime/mainw->multitrack->end_secs*eventbox->allocation.width;
      gdk_draw_line (GDK_DRAWABLE(eventbox->window), mainw->LiVES->style->black_gc, offset, 1, offset, eventbox->allocation.height-1);
    }
    } */
} 






////////////////////////////////////////////////////
// menuitem callbacks


gboolean multitrack_end (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  multitrack_delete (mt);
  return TRUE;
}



// callbacks for future adding to osc.c
void multitrack_end_cb (GtkMenuItem *menuitem, gpointer user_data) {
  multitrack_end(menuitem,user_data);
}


void insert_here_cb (GtkMenuItem *menuitem, gpointer user_data) {
  multitrack_insert(NULL,user_data);
}


void delete_block_cb (GtkMenuItem *menuitem, gpointer user_data) {
  on_delblock_activate(NULL,user_data);
}


///////////////////////////////////////////////////////////

void on_snap_over_changed (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  mt->snap_over=!mt->snap_over;
}



void select_all_time (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  mt->region_start=0.;
  mt->region_end=mt->end_secs;
  on_timeline_release(NULL,NULL,mt);
}


void select_from_zero_time (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  if (mt->region_start==0.&&mt->region_end==0.) mt->region_end=GTK_RULER (mt->timeline)->position;
  mt->region_start=0.;
  on_timeline_release(NULL,NULL,mt);
}

void select_to_end_time (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  if (mt->region_start==0.&&mt->region_end==0.) mt->region_start=GTK_RULER (mt->timeline)->position;
  mt->region_end=mt->end_secs;
  on_timeline_release(NULL,NULL,mt);
}



void select_all_vid (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  GtkWidget *eventbox,*checkbutton;
  gint current_track=mt->current_track;
  GList *vdr=mt->video_draws;

  while (vdr!=NULL) {
    eventbox=vdr->data;
    checkbutton=g_object_get_data(G_OBJECT(eventbox),"checkbutton");
    if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton))) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),TRUE);
    vdr=vdr->next;
  }
  mt->current_track=current_track;
  track_select(mt);

}

void select_no_vid (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  GtkWidget *eventbox,*checkbutton;
  gint current_track=mt->current_track;
  GList *vdr=mt->video_draws;

  while (vdr!=NULL) {
    eventbox=vdr->data;
    checkbutton=g_object_get_data(G_OBJECT(eventbox),"checkbutton");
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton))) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),FALSE);
    vdr=vdr->next;
  }
  mt->current_track=current_track;
  track_select(mt);

}


static void remove_gaps_inner (GtkMenuItem *menuitem, gpointer user_data, gboolean only_first) {
  lives_mt *mt=(lives_mt *)user_data;
  GList *vsel=mt->selected_tracks;
  track_rect *block;
  gint track;
  GtkWidget *eventbox;
  weed_timecode_t tc,blank_tc=0.,start_tc,end_tc,tc_last;
  gboolean got_first_blank;
  weed_plant_t *event,*shortcut;
  gint filenum;
  gboolean did_backup=mt->did_backup;
  weed_timecode_t offset=0;
  GList *track_sel;

  //go through selected tracks, move each block as far left as possible

  tc_last=q_gint64(mt->region_end*U_SEC,mt->fps);

  while (vsel!=NULL) {
    track=GPOINTER_TO_INT(vsel->data);
    eventbox=g_list_nth_data(mt->video_draws,track);
    tc=mt->region_start*U_SEC;
    tc=q_gint64(tc,mt->fps);
    got_first_blank=FALSE;
    shortcut=NULL;
    while (tc<=tc_last) {
      if ((event=get_frame_event_at(mt->event_list,tc,shortcut,TRUE))==NULL) break;
      shortcut=event;
      if (frame_event_has_frame_for_track(event,track)) {
	// frame at tc
	if (!got_first_blank) {
	  tc+=U_SEC/mt->fps;
	  tc=q_gint64(tc,mt->fps);
	  continue; // no blanks yet
	}
	// move this block to blank_tc
	block=get_block_from_time(eventbox,tc/U_SEC,mt);
	if (only_first&&block->next!=NULL) offset=get_event_timecode(block->next->start_event)-get_event_timecode(block->end_event);
	filenum=get_frame_event_clip(block->start_event,track);
	mt->clip_selected=filenum-1;
	clip_select(mt);
	if (!mt->did_backup) mt_backup(mt,MT_UNDO_REMOVE_GAPS,NULL);

	// save current selected_tracks, move_block may change this
	track_sel=mt->selected_tracks;
	mt->selected_tracks=NULL;
	block=move_block(mt,block,blank_tc/U_SEC,track,track);
	if (mt->selected_tracks!=NULL) g_list_free(mt->selected_tracks);
	mt->selected_tracks=track_sel;

	start_tc=get_event_timecode(block->start_event);
	end_tc=get_event_timecode(block->end_event);

	// set tc to new end_of_block
	tc=end_tc-start_tc+blank_tc;
	shortcut=block->end_event;
	if (!only_first) got_first_blank=FALSE;
	else blank_tc=q_gint64(end_tc+offset,mt->fps);
      }
      else {
	// blank at tc
	if (!got_first_blank) {
	  got_first_blank=TRUE;
	  blank_tc=tc;
	}
      }
      tc=q_gint64(tc+U_SEC/mt->fps,mt->fps);
    }
    vsel=vsel->next;
  }
  mt->did_backup=did_backup;
}


void remove_first_gaps (GtkMenuItem *menuitem, gpointer user_data) {
  remove_gaps_inner(menuitem,user_data,TRUE);
}

void remove_gaps (GtkMenuItem *menuitem, gpointer user_data) {
  remove_gaps_inner(menuitem,user_data,FALSE);
}



static weed_plant_t *load_event_list_inner (int fd, gboolean show_errors, gint *num_events, unsigned char **mem, unsigned char *mem_end) {
  weed_plant_t *event,*eventprev=NULL;
  int error;
  weed_plant_t *event_list=weed_plant_deserialise(fd,mem);

  if (event_list==NULL||!WEED_PLANT_IS_EVENT_LIST(event_list)) {
    if (show_errors) d_print(_("invalid event list. Failed.\n"));
    return NULL;
  }

  if (show_errors&&(!weed_plant_has_leaf(event_list,"fps")||weed_get_double_value(event_list,"fps",&error)<1.||weed_get_double_value(event_list,"fps",&error)>FPS_MAX)) {
    d_print(_("event list has invalid fps. Failed.\n"));
    return NULL;
  }

  if (weed_plant_has_leaf(event_list,"needs_set")) {
    if (show_errors) {
      gchar *set_needed=weed_get_string_value(event_list,"needs_set",&error);
      gchar *err;
      if (!mainw->was_set||strcmp(set_needed,mainw->set_name)) {
	err=g_strdup_printf(_("\nThis layout requires the set \"%s\"\nIn order to load it you must restart LiVES, \nthen load in that set from the VJ menu.\n"),set_needed);
	d_print(err);
	do_error_dialog(err);
	g_free(err);
	g_free(set_needed);
	return NULL;
      }
    }
  }
  else if (!show_errors&&mem==NULL) return NULL; // no change needed

  weed_set_plantptr_value(event_list,"first",NULL);
  weed_set_plantptr_value(event_list,"last",NULL);

  do {
    if (mem!=NULL&&(*mem)>=mem_end) break;
    event=weed_plant_deserialise(fd,mem);
    if (event!=NULL) {
      if (eventprev!=NULL) weed_set_plantptr_value(eventprev,"next",event);
      weed_set_plantptr_value(event,"previous",eventprev);
      weed_set_plantptr_value(event,"next",NULL);

      if (get_first_event(event_list)==NULL) {
	weed_set_plantptr_value(event_list,"first",event);
      }
      weed_set_plantptr_value(event_list,"last",event);
      eventprev=event;
      if (num_events!=NULL) (*num_events)++;
    }
  } while (event!=NULL);

  return event_list;
}



void
multitrack_undo            (GtkMenuItem     *menuitem,
			    gpointer         user_data) {

  lives_mt *mt=(lives_mt *)user_data;
  size_t space_avail=(size_t)(prefs->mt_undo_buf*1024*1024)-mt->undo_buffer_used;
  size_t space_needed;
  mt_undo *last_undo=g_list_nth_data(mt->undos,g_list_length(mt->undos)-1-mt->undo_offset);
  unsigned char *memblock,*mem_end;
  int i;
  mt_undo *new_redo=NULL;
  gint current_track;
  gdouble end_secs;
  GList *slist;
  gint num_tracks;
  GtkWidget *checkbutton,*eventbox,*label;
  GList *label_list=NULL;
  GList *vlist,*llist;
  gchar *txt;
  GList *seltracks=NULL;
  gint clip_sel;

  if (mt->undo_mem==NULL) return;

  switch (last_undo->action) {
  case MT_UNDO_REMOVE_GAPS:
  case MT_UNDO_MOVE_BLOCK:
  case MT_UNDO_DELETE_BLOCK:
  case MT_UNDO_SPLIT_MULTI:
  case MT_UNDO_SPLIT:
  case MT_UNDO_APPLY_FILTER:
  case MT_UNDO_INSERT_BLOCK:

    if (mt->undo_offset==0) {
      add_markers(mt,mt->event_list);
      if ((space_needed=estimate_space(mt,last_undo->action))>space_avail) {
	if (!make_backup_space(mt,space_needed)||mt->undos==NULL) {
	  remove_markers(mt->event_list);
	  do_mt_undo_buf_error();
	  return;
	}
      }

      new_redo=(mt_undo *)(mt->undo_mem+mt->undo_buffer_used);
      new_redo->action=last_undo->action;
      
      memblock=(unsigned char *)(new_redo)+sizeof(mt_undo);
      new_redo->data_len=space_needed;
      save_event_list_inner(0,mt->event_list,&memblock);
      mt->undo_buffer_used+=space_needed;
      mt->undos=g_list_append(mt->undos,new_redo);
      mt->undo_offset++;
    }

    current_track=mt->current_track;
    end_secs=mt->end_secs;
    num_tracks=mt->num_video_tracks;
    clip_sel=mt->clip_selected;

    seltracks=g_list_copy(mt->selected_tracks);

    vlist=mt->video_draws;
    while (vlist!=NULL) {
      eventbox=GTK_WIDGET(vlist->data);
      label=GTK_WIDGET(g_object_get_data(G_OBJECT(eventbox),"label"));
      txt=g_strdup(gtk_label_get_text(GTK_LABEL(label)));
      label_list=g_list_append(label_list,txt);
      vlist=vlist->next;
    }

    event_list_free(mt->event_list);
    last_undo=g_list_nth_data(mt->undos,g_list_length(mt->undos)-1-mt->undo_offset);
    memblock=(unsigned char *)(last_undo)+sizeof(mt_undo);
    mem_end=memblock+last_undo->data_len-sizeof(mt_undo);
    mt->event_list=load_event_list_inner(-1,FALSE,NULL,&memblock,mem_end);

    if (!event_list_rectify(mt->event_list,FALSE)) {
      event_list_free(mt->event_list);
      mt->event_list=NULL;
    }
    
    if (get_first_event(mt->event_list)==NULL) {
      event_list_free(mt->event_list);
      mt->event_list=NULL;
    }
    
    for (i=0;i<mt->num_video_tracks;i++) {
      delete_video_track(mt,i,FALSE);
    }
    g_list_free(mt->video_draws);
    mt->video_draws=NULL;
    mt->num_video_tracks=0;

    gtk_widget_destroy(mt->timeline_eb);
    init_tracks(mt);
    for (i=mt->num_video_tracks;i<num_tracks;i++) {
      add_video_track_behind (NULL,mt);
    }

    mt->clip_selected=clip_sel;
    clip_select(mt);

    vlist=mt->video_draws;
    llist=label_list;
    while (vlist!=NULL) {
      eventbox=GTK_WIDGET(vlist->data);
      label=GTK_WIDGET(g_object_get_data(G_OBJECT(eventbox),"label"));
      g_free(GTK_LABEL(label)->text);
      GTK_LABEL(label)->text=llist->data;
      vlist=vlist->next;
      llist=llist->next;
    }
    g_list_free(label_list);

    if (mt->event_list!=NULL) remove_markers(mt->event_list);
    gtk_widget_show_all(mt->timeline_eb);

    mt->selected_tracks=g_list_copy(seltracks);
    slist=mt->selected_tracks;
    while (slist!=NULL) {
      eventbox=g_list_nth_data(mt->video_draws,GPOINTER_TO_INT(slist->data));
      checkbutton=g_object_get_data(G_OBJECT(eventbox),"checkbutton");
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),TRUE);
      slist=slist->next;
    }
    if (seltracks!=NULL) g_list_free(seltracks);

    mt->current_track=current_track;
    track_select(mt);
    if (mt->end_secs!=end_secs&&event_list_get_end_secs(mt->event_list)<=end_secs) set_timeline_end_secs (mt, end_secs);
    break;
  }

  mt->undo_offset++;

  if (mt->undo_offset==g_list_length(mt->undos)) mt_set_undoable(mt,MT_UNDO_NONE,NULL,FALSE);
  else {
    mt_undo *undo=(mt_undo *)(g_list_nth_data(mt->undos,g_list_length(mt->undos)-mt->undo_offset-1));
    mt_set_undoable(mt,undo->action,undo->extra,TRUE);
  }
  mt_set_redoable(mt,last_undo->action,last_undo->extra,TRUE);
}


void
multitrack_redo            (GtkMenuItem     *menuitem,
			    gpointer         user_data) {

  lives_mt *mt=(lives_mt *)user_data;
  mt_undo *last_redo=g_list_nth_data(mt->undos,g_list_length(mt->undos)+1-mt->undo_offset);
  unsigned char *memblock,*mem_end;
  int i;
  gint current_track;
  gdouble end_secs;
  GList *slist;
  gint num_tracks;
  GtkWidget *checkbutton,*eventbox,*label;
  GList *label_list=NULL;
  GList *vlist,*llist;
  gchar *txt;
  GList *seltracks=NULL;
  gint clip_sel;

  if (mt->undo_mem==NULL) return;

  switch (last_redo->action) {
  case MT_UNDO_APPLY_FILTER:
  case MT_UNDO_REMOVE_GAPS:
  case MT_UNDO_MOVE_BLOCK:
  case MT_UNDO_DELETE_BLOCK:
  case MT_UNDO_SPLIT_MULTI:
  case MT_UNDO_SPLIT:
  case MT_UNDO_INSERT_BLOCK:

    current_track=mt->current_track;
    end_secs=mt->end_secs;
    num_tracks=mt->num_video_tracks;
    clip_sel=mt->clip_selected;

    seltracks=g_list_copy(mt->selected_tracks);

    vlist=mt->video_draws;
    while (vlist!=NULL) {
      eventbox=GTK_WIDGET(vlist->data);
      label=GTK_WIDGET(g_object_get_data(G_OBJECT(eventbox),"label"));
      txt=g_strdup(gtk_label_get_text(GTK_LABEL(label)));
      label_list=g_list_append(label_list,txt);
      vlist=vlist->next;
    }

    event_list_free(mt->event_list);

    memblock=(unsigned char *)(last_redo)+sizeof(mt_undo);
    mem_end=memblock+last_redo->data_len-sizeof(mt_undo);
    mt->event_list=load_event_list_inner(-1,FALSE,NULL,&memblock,mem_end);

    if (!event_list_rectify(mt->event_list,FALSE)) {
      event_list_free(mt->event_list);
      mt->event_list=NULL;
    }
    
    if (get_first_event(mt->event_list)==NULL) {
      event_list_free(mt->event_list);
      mt->event_list=NULL;
    }
    
    for (i=0;i<mt->num_video_tracks;i++) {
      delete_video_track(mt,i,FALSE);
    }
    g_list_free(mt->video_draws);
    mt->video_draws=NULL;
    mt->num_video_tracks=0;

    gtk_widget_destroy(mt->timeline_eb);
    init_tracks(mt);
    for (i=mt->num_video_tracks;i<num_tracks;i++) {
      add_video_track_behind (NULL,mt);
    }

    mt->clip_selected=clip_sel;
    clip_select(mt);

    vlist=mt->video_draws;
    llist=label_list;
    while (vlist!=NULL) {
      eventbox=GTK_WIDGET(vlist->data);
      label=GTK_WIDGET(g_object_get_data(G_OBJECT(eventbox),"label"));
      g_free(GTK_LABEL(label)->text);
      GTK_LABEL(label)->text=llist->data;
      vlist=vlist->next;
      llist=llist->next;
    }
    g_list_free(label_list);

    if (mt->event_list!=NULL) remove_markers(mt->event_list);
    gtk_widget_show_all(mt->timeline_eb);

    mt->selected_tracks=g_list_copy(seltracks);
    slist=mt->selected_tracks;
    while (slist!=NULL) {
      eventbox=g_list_nth_data(mt->video_draws,GPOINTER_TO_INT(slist->data));
      checkbutton=g_object_get_data(G_OBJECT(eventbox),"checkbutton");
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),TRUE);
      slist=slist->next;
    }
    if (seltracks!=NULL) g_list_free(seltracks);

    mt->current_track=current_track;
    track_select(mt);
    if (mt->end_secs!=end_secs&&event_list_get_end_secs(mt->event_list)<=end_secs) set_timeline_end_secs (mt, end_secs);
    break;
  }

  mt->undo_offset--;

  if (mt->undo_offset<=1) mt_set_redoable(mt,MT_UNDO_NONE,NULL,FALSE);
  else {
    mt_undo *redo=(mt_undo *)(g_list_nth_data(mt->undos,g_list_length(mt->undos)-mt->undo_offset));
    mt_set_redoable(mt,redo->action,redo->extra,TRUE);
  }
  last_redo=g_list_nth_data(mt->undos,g_list_length(mt->undos)-1-mt->undo_offset);
  mt_set_undoable(mt,last_redo->action,last_redo->extra,TRUE);

}



void
multitrack_view_details            (GtkMenuItem     *menuitem,
				    gpointer         user_data) {
  char buff[512];
  fileinfo *filew;
  lives_mt *mt=(lives_mt *)user_data;
  file *rfile=mainw->files[mt->render_file];
  guint bsize=0;
  gdouble time=0.;
  gint num_events=0;

  filew = create_info_window (mt->achans,TRUE);

  g_snprintf(buff,512,"LiVES - %s",_("Multitrack details"));
  gtk_window_set_title (GTK_WINDOW (filew->info_window), buff);
  
  // type
  g_snprintf(buff,512,"\n  Event List");
  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview24)),buff, -1);

  // fps
  if (mt->fps>0) {
      g_snprintf(buff,512,"\n  %.3f%s",mt->fps,rfile->ratio_fps?"...":"");
  }
  else {
    g_snprintf(buff,512,_ ("\n (variable)"));
  }

  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview25)),buff, -1);

  // image size
  g_snprintf(buff,512,"\n  %dx%d",rfile->hsize,rfile->vsize);
  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview26)),buff, -1);

  // elist time
  if (mt->event_list!=NULL) {
    bsize=event_list_get_byte_size(mt->event_list,&num_events);
    time=event_list_get_end_secs(mt->event_list);
  }

  // events
  g_snprintf(buff,512,"\n  %d",num_events);
  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview27)),buff, -1);

  g_snprintf(buff,512,"\n  %.3f sec",time);
  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview28)),buff, -1);

  // byte size
  g_snprintf(buff,512,"\n  %d bytes",bsize);
  gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview29)),buff, -1);

  if (mt->achans>0) {
    //g_snprintf(buff,512,"\n  %.2f sec",mt->laudio_time);
    //gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview_ltime)),buff, -1);
    g_snprintf(buff,512,"\n  %d Hz %d bit",mt->arate,mt->asampsize);
    gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview_lrate)),buff, -1);
  }
  
  if (mt->achans>1) {
    g_snprintf(buff,512,"\n  %d Hz %d bit",mt->arate,mt->asampsize);
    gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview_rrate)),buff, -1);
    //g_snprintf(buff,512,"\n  %.2f sec",mt->raudio_time);
    //gtk_text_buffer_set_text (gtk_text_view_get_buffer (GTK_TEXT_VIEW (filew->textview_rtime)),buff, -1);
  }
  gtk_widget_show (filew->info_window);
}



void mt_add_block_effect (GtkMenuItem *menuitem, gpointer user_data) {
  weed_plant_t *filter;
  lives_mt *mt=(lives_mt *)user_data;
  weed_plant_t *start_event=mt->block_selected->start_event;
  weed_plant_t *end_event=mt->block_selected->end_event;
  weed_timecode_t start_tc=get_event_timecode(start_event);
  weed_timecode_t end_tc=get_event_timecode(end_event);
  weed_plant_t *event;
  void **init_events;
  gchar *filter_name;
  gchar *text;
  int selected_track;

  gboolean did_backup=mt->did_backup;


  mt->current_fx=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(menuitem),"idx"));
  filter=get_weed_filter(mt->current_fx);
  if (!did_backup) mt_backup(mt,MT_UNDO_APPLY_FILTER,NULL);

  // add effect_init event
  append_filter_init_event(mt->event_list,start_tc,mt->current_fx);
  mt->init_event=get_last_event(mt->event_list);
  unlink_event(mt->event_list,mt->init_event);
  selected_track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(mt->block_selected->eventbox),"layer_number"));
  weed_set_int_value(mt->init_event,"in_tracks",selected_track);
  weed_set_int_value(mt->init_event,"out_tracks",selected_track);
  insert_filter_init_event_at(mt->event_list,start_event,mt->init_event);
  if (weed_plant_has_leaf(filter,"in_parameter_templates")) pchain=filter_init_add_pchanges(mt->event_list,filter,mt->init_event);

  // add effect map event
  init_events=get_init_events_before(mt->event_list,start_event,mt->init_event,TRUE);
  append_filter_map_event(mt->event_list,start_tc,init_events);
  g_free(init_events);
  event=get_last_event(mt->event_list);
  unlink_event(mt->event_list,event);
  insert_filter_map_event_at(mt->event_list,start_event,event,TRUE);

  // update all effect maps in block, appending init_event
  update_filter_maps(mt->event_list,start_event,end_event,mt->init_event);

  // add effect deinit event
  append_filter_deinit_event(mt->event_list,end_tc,(void *)mt->init_event,pchain);
  event=get_last_event(mt->event_list);
  unlink_event(mt->event_list,event);
  insert_filter_deinit_event_at(mt->event_list,end_event,event);

  // zip forward a bit, in case there is a FILTER_MAP at end_tc after our FILTER_DEINIT (e.g. if adding multiple filters)
  while (event!=NULL&&get_event_timecode(event)==end_tc) event=get_next_event(event);
  if (event==NULL) event=get_last_event(mt->event_list);
  else event=get_prev_event(event);

  // add effect map event 
  init_events=get_init_events_before(mt->event_list,event,mt->init_event,FALSE); // also deletes the effect
  append_filter_map_event(mt->event_list,end_tc,init_events);
  event=get_last_event(mt->event_list);
  unlink_event(mt->event_list,event);
  insert_filter_map_event_at(mt->event_list,end_event,event,FALSE);

  if (num_in_params(filter,FALSE,FALSE)>0) polymorph(mt,POLY_EFFECT);
  else polymorph(mt,POLY_IN_OUT);

  filter_name=weed_filter_get_name(mt->current_fx);
  text=g_strdup_printf(_("Added effect %s to track %s from %.4f to %.4f\n"),filter_name,get_track_name(mt,selected_track),start_tc/U_SEC,end_tc/U_SEC);
  d_print(text);
  g_free(text);
  g_free(filter_name);

  mt->did_backup=did_backup;
}
  

void on_jumpto_activate (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  GtkWidget *eventbox=g_list_nth_data(mt->video_draws,mt->current_track);
  gdouble secs=GTK_RULER (mt->timeline)->position/GTK_RULER (mt->timeline)->upper*mt->end_secs,new_secs;
  track_rect *block=get_block_from_time(eventbox,secs,mt);
  gint64 tc,new_tc;

  if (block==NULL) {
    if (mt->last_direction==DIRECTION_NEGATIVE) {
      block=get_block_before(eventbox,secs);
      if (block==NULL) return;
      new_secs=get_event_timecode(block->end_event)/U_SEC+1./mt->fps;
      new_tc=new_secs*U_SECL+.5;
      tc=secs*U_SECL+.5;
      secs=(new_tc!=tc)?new_secs:get_event_timecode(block->start_event)/U_SEC-1./mt->fps;
    }
    else {
      block=get_block_after(eventbox,secs);
      if (block==NULL) return;
      new_secs=get_event_timecode(block->start_event)/U_SEC-1./mt->fps;
      new_tc=new_secs*U_SECL+.5;
      tc=secs*U_SECL+.5;
      secs=(new_tc!=tc)?new_secs:get_event_timecode(block->end_event)/U_SEC+1./mt->fps;
    }
  }
  else {
    if (mt->last_direction==DIRECTION_NEGATIVE) secs=get_event_timecode(block->start_event)/U_SEC-1./mt->fps;
    else secs=get_event_timecode(block->end_event)/U_SEC+1./mt->fps;
  }

  if (secs<0.) secs=0.;
  if (secs>mt->end_secs) set_timeline_end_secs (mt, secs);

  GTK_RULER (mt->timeline)->position=secs;
  gtk_widget_queue_draw (mt->timeline);
}

void on_render_activate (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  gint current_file;

  mainw->event_list=mt->event_list;

  mt->is_rendering=TRUE;

  if (mt->poly_state==POLY_EFFECT) polymorph(mt,POLY_CLIPS);

  if (render_to_clip(FALSE)) {
    cfile->start=1;
    cfile->end=cfile->frames;
    set_undoable (NULL,FALSE);
    add_to_winmenu();
    current_file=mainw->current_file;
    d_print (g_strdup_printf (_ ("rendered %d frames to new clip.\n"),cfile->frames));
    mt->file_selected=current_file;
    mt->is_rendering=FALSE;
    save_clip_values(current_file);
    if (prefs->crash_recovery) add_to_recovery_file(cfile->handle);

    if (prefs->mt_exit_render) {
      if (!multitrack_end(menuitem,user_data)) return;
      switch_to_file ((mainw->current_file=0),current_file);
    }
    else {
      gint orig_file=mainw->current_file;
      mt->is_rendering=FALSE;
      mainw->current_file=mainw->first_free_file;

      if (!get_new_handle(mainw->current_file,NULL)) {
	mainw->current_file=orig_file;
	if (multitrack_end(menuitem,user_data)) switch_to_file ((mainw->current_file=0),current_file);
	return;
      }

      cfile->hsize=mainw->files[orig_file]->hsize;
      cfile->vsize=mainw->files[orig_file]->vsize;

      cfile->pb_fps=cfile->fps=mainw->files[orig_file]->fps;
      cfile->ratio_fps=mainw->files[orig_file]->ratio_fps;

      cfile->bpp=24;
      cfile->changed=TRUE;
      cfile->is_loaded=TRUE;
  
      cfile->old_frames=cfile->frames;

      mt->render_file=mainw->current_file;

      init_clips(mt,orig_file,TRUE);
      while (g_main_context_iteration(NULL,FALSE));
      clip_select(mt);
    }
  }
  else {
    cfile->frames=cfile->start=cfile->end=0;
    mt->is_rendering=FALSE;
    mainw->event_list=NULL;
    d_print(_("cancelled.\n"));
  }
}



void update_filter_events(lives_mt *mt, weed_plant_t *first_event, weed_timecode_t start_tc, weed_timecode_t end_tc, int track, weed_timecode_t new_start_tc, int new_track) {
  // move/remove filter_inits param_change and filter_deinits after deleting/moving a block
  weed_plant_t *event,*event_next;
  gboolean was_moved;
  int error;
  weed_plant_t *init_event,*deinit_event;
  GList *moved_events=NULL;
  gboolean leave_event;
  int i;

  if (first_event==NULL) event=get_first_event(mt->event_list);
  else event=get_next_event(first_event);
  while (event!=NULL&&get_event_timecode(event)<start_tc) event=get_next_event(event);

  while (event!=NULL&&get_event_timecode(event)<=end_tc) {
    event_next=get_next_event(event);
    was_moved=FALSE;
    if (WEED_EVENT_IS_FILTER_INIT(event)) {
      if (mt->move_effects&&mt->moving_block) {
	if (weed_plant_has_leaf(event,"deinit_event")&&weed_plant_has_leaf(event,"in_tracks")&&weed_leaf_num_elements(event,"in_tracks")==1&&weed_get_int_value(event,"in_tracks",&error)==track) {
	  deinit_event=weed_get_voidptr_value(event,"deinit_event",&error);
	  if (get_event_timecode(deinit_event)<=end_tc) {
	    if (g_list_index(moved_events,event)==-1) {
	      // update owners,in_tracks and out_tracks
	      weed_set_int_value(event,"in_tracks",new_track);

	      if (weed_plant_has_leaf(event,"out_tracks")) {
		int *out_tracks=weed_get_int_array(event,"out_tracks",&error);
		int num_tracks=weed_leaf_num_elements(event,"out_tracks");
		for (i=0;i<num_tracks;i++) {
		  if (out_tracks[i]==track) out_tracks[i]=new_track;
		}
		weed_set_int_array(event,"out_tracks",num_tracks,out_tracks);
		weed_free(out_tracks);
	      }

	      // move to new position
	      if (new_start_tc<start_tc) {
		move_filter_init_event(mt->event_list,get_event_timecode(event)+new_start_tc-start_tc,event,mt->fps);
		move_filter_deinit_event(mt->event_list,get_event_timecode(deinit_event)+new_start_tc-start_tc,deinit_event,mt->fps);
		if (event==first_event) first_event=NULL;
		was_moved=TRUE;
	      }
	      else if (new_start_tc>start_tc) {
		move_filter_deinit_event(mt->event_list,get_event_timecode(deinit_event)+new_start_tc-start_tc,deinit_event,mt->fps);
		move_filter_init_event(mt->event_list,get_event_timecode(event)+new_start_tc-start_tc,event,mt->fps);
		if (event==first_event) first_event=NULL;
		was_moved=TRUE;
	      }
	      moved_events=g_list_prepend(moved_events,event);
	    }
	  }
	}
      }
      if (g_list_index(moved_events,event)==-1&&!was_moved&&!move_event_right(mt->event_list,event,TRUE,mt->fps)) {
	was_moved=TRUE;
	if (event==first_event) first_event=NULL;
      }
    }
    else {
      leave_event=TRUE;
      if (WEED_EVENT_IS_FILTER_DEINIT(event)) {
	if (mt->move_effects&&mt->moving_block) {
	  if (weed_plant_has_leaf(event,"init_event")) {
	    init_event=weed_get_voidptr_value(event,"init_event",&error);
	    if (g_list_index(moved_events,init_event)==-1&&!(weed_plant_has_leaf(event,"in_tracks")&&weed_leaf_num_elements(event,"in_tracks")==1&&weed_get_int_value(event,"in_tracks",&error)==track)) {
	      leave_event=FALSE;
	    }
	  }
	}
	if (!leave_event&&!move_event_left(mt->event_list,event,TRUE,mt->fps)) {
	  was_moved=TRUE;
	}
      }
    }
    if (was_moved) {
      if (first_event==NULL) event=get_first_event(mt->event_list);
      else event=get_next_event(first_event);
      while (event!=NULL&&get_event_timecode(event)<start_tc) event=get_next_event(event);
    }
    else {
      event=event_next;
      if (WEED_EVENT_IS_FRAME(event)) first_event=event;
    }
  }
  if (moved_events!=NULL) g_list_free(moved_events);
}


static void split_block(lives_mt *mt, track_rect *block, weed_timecode_t tc, gint track) {
  weed_plant_t *event=block->start_event;
  tc=q_gint64(tc,mt->fps);
  weed_plant_t *old_end_event=block->end_event;
  gint frame,clip;
  GtkWidget *eventbox;
  track_rect *new_block;
  weed_timecode_t offset_start;

  while (get_event_timecode(event)<tc) event=get_next_event(event);
  block->end_event=get_prev_event(event);
  if (!WEED_EVENT_IS_FRAME(block->end_event)) block->end_event=get_prev_frame_event(event);

  if (!WEED_EVENT_IS_FRAME(event)) event=get_next_frame_event(event);

  frame=get_frame_event_frame(event,track);
  clip=get_frame_event_clip(event,track);

  eventbox=g_list_nth_data(mt->video_draws,track);
  if (block->ordered) offset_start=block->offset_start-get_event_timecode(block->start_event)+get_event_timecode(event);
  else offset_start=calc_variable_time(clip,frame)*U_SEC;

  new_block=add_block_start_point (GTK_WIDGET(eventbox),tc,clip,offset_start,event,FALSE);
  new_block->end_event=old_end_event;

  gtk_widget_queue_draw(eventbox);

}



void on_split_activate (GtkMenuItem *menuitem, gpointer user_data) {
  // split current block at current time
  lives_mt *mt=(lives_mt *)user_data;
  gdouble timesecs=GTK_RULER (mt->timeline)->position;
  gboolean did_backup=mt->did_backup;

  if (!did_backup) mt_backup(mt,MT_UNDO_SPLIT,NULL);

  timesecs=(gint)(timesecs*mt->fps+.5)/mt->fps; // quantise to frame timing

  split_block(mt,mt->putative_block,timesecs*U_SEC,mt->current_track);
  mt->did_backup=did_backup;
}



void on_split_sel_activate (GtkMenuItem *menuitem, gpointer user_data) {
  // split selected tracks at current time
  lives_mt *mt=(lives_mt *)user_data;
  GList *selt=mt->selected_tracks;
  GtkWidget *eventbox;
  gint track;
  track_rect *block;
  gdouble timesecs=GTK_RULER (mt->timeline)->position;
  gboolean did_backup=mt->did_backup;

  if (mt->selected_tracks==NULL) return;

  if (!did_backup) mt_backup(mt,MT_UNDO_SPLIT_MULTI,NULL);

  timesecs=(gint)(timesecs*mt->fps+.5)/mt->fps; // quantise to frame timing

  while (selt!=NULL) {
    track=GPOINTER_TO_INT(selt->data);
    eventbox=g_list_nth_data(mt->video_draws,track);
    block=get_block_from_time(eventbox,timesecs,mt);
    split_block(mt,block,timesecs*U_SEC,track);
    selt=selt->next;
  }
  mt->did_backup=did_backup;

}



void on_delblock_activate (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  track_rect *block;
  weed_plant_t *event,*prevevent;
  GtkWidget *eventbox;
  gint track;
  gboolean done=FALSE;
  weed_timecode_t start_tc,end_tc;
  weed_plant_t *first_event;
  track_rect *firstblock;
  gboolean did_backup=mt->did_backup;

  if (!did_backup) mt_backup(mt,MT_UNDO_DELETE_BLOCK,NULL);

  if (mt->block_selected==NULL) mt->block_selected=mt->putative_block;
  block=mt->block_selected;
  eventbox=block->eventbox;
  track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"layer_number"));
  mt_desensitise(mt);

  start_tc=get_event_timecode(block->start_event);
  end_tc=get_event_timecode(block->end_event);

  first_event=get_prev_event(block->start_event);
  while (first_event!=NULL&&get_event_timecode(first_event)==start_tc) {
    first_event=get_prev_event(first_event);
  }

  event=block->end_event;
  // delete frames
  while (event!=NULL&&!done) {
    prevevent=get_prev_frame_event(event);
    if (event==block->start_event) done=TRUE;
    remove_frame_from_event(mt->event_list,event,track);
    if (!done) event=prevevent;
  }
  firstblock=(track_rect *)g_object_get_data (G_OBJECT(eventbox), "blocks");
  if (firstblock==block) g_object_set_data (G_OBJECT(eventbox),"blocks",(gpointer)block->next);
  if (block->prev!=NULL) block->prev->next=block->next;
  if (block->next!=NULL) block->next->prev=block->prev;
  g_free(block);

  gtk_widget_queue_draw (eventbox);

  if (!mt->move_effects||!mt->moving_block) {
    update_filter_events(mt,first_event,start_tc,end_tc,track,start_tc,track);
    remove_end_blank_frames(mt->event_list);
    if (mt->block_selected==block) {
      mt->block_selected=NULL;
      unselect_all(mt);
    }
    mt_sensitise(mt);
  }

  if (mt->snapt&&!mt->moving_block&&!did_backup) {
    gdouble oldr_start=mt->region_start;
    gdouble oldr_end=mt->region_end;
    GList *tracks_sel=g_list_copy(mt->selected_tracks);
    if (mt->selected_tracks!=NULL) g_list_free(mt->selected_tracks);
    mt->selected_tracks=NULL;
    mt->selected_tracks=g_list_append(mt->selected_tracks,GINT_TO_POINTER(mt->current_track));
    mt->region_start=start_tc/U_SEC;
    mt->region_end=mt->end_secs;
    remove_first_gaps(NULL,mt);
    g_list_free(mt->selected_tracks);
    mt->selected_tracks=g_list_copy(tracks_sel);
    if (tracks_sel!=NULL) g_list_free(tracks_sel);
    mt->region_start=oldr_start;
    mt->region_end=oldr_end;
  }

  mt->did_backup=did_backup;

}



void on_seltrack_activate (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  GtkWidget *eventbox=g_list_nth_data(mt->video_draws,mt->current_track);
  GtkWidget *checkbutton=g_object_get_data(G_OBJECT(eventbox),"checkbutton");
  gulong seltrack_func=(gulong)g_object_get_data(G_OBJECT(checkbutton),"tfunc");
  gboolean mi_state=gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(menuitem));

  if (mi_state) {
    if (g_list_index(mt->selected_tracks,GINT_TO_POINTER(mt->current_track))==-1) mt->selected_tracks=g_list_append(mt->selected_tracks,GINT_TO_POINTER(mt->current_track));
  }
  else {
    if (g_list_index(mt->selected_tracks,GINT_TO_POINTER(mt->current_track))!=-1) mt->selected_tracks=g_list_remove(mt->selected_tracks,GINT_TO_POINTER(mt->current_track));
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton))!=mi_state) {
    g_signal_handler_block(checkbutton,seltrack_func);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton),mi_state);
    g_signal_handler_unblock(checkbutton,seltrack_func);
  }
  do_sel_context(mt);
}


void on_seltrack_toggled (GtkToggleButton *checkbutton, gpointer user_data) {
  gint track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(checkbutton),"layer_number"));
  lives_mt *mt=(lives_mt *)user_data;

  mt->current_track=track;

  // track_select will call on_seltrack_activate, which will set our new state
  track_select(mt);
}



///////////////////////////////////////////////////////////


void mt_desensitise (lives_mt *mt) {
  gdouble val;
  gtk_widget_set_sensitive (mt->clipedit,FALSE);
  gtk_widget_set_sensitive (mt->insert,FALSE);
  gtk_widget_set_sensitive (mt->playall,FALSE);
  gtk_widget_set_sensitive (mt->view_events,FALSE);
  gtk_widget_set_sensitive (mt->jumpto, FALSE);
  gtk_widget_set_sensitive (mt->render, FALSE);
  gtk_widget_set_sensitive (mt->delblock, FALSE);
  gtk_widget_set_sensitive (mt->save_event_list, FALSE);
  gtk_widget_set_sensitive (mt->load_event_list, FALSE);
  gtk_widget_set_sensitive (mt->move_fx, FALSE);
  gtk_widget_set_sensitive (mt->remove_gaps, FALSE);
  gtk_widget_set_sensitive (mt->remove_first_gaps, FALSE);
  gtk_widget_set_sensitive (mt->undo, FALSE);
  gtk_widget_set_sensitive (mt->redo, FALSE);

  if (mt->poly_state==POLY_IN_OUT) {
    val=gtk_spin_button_get_value(GTK_SPIN_BUTTON(mt->spinbutton_in));
    gtk_spin_button_set_range(GTK_SPIN_BUTTON(mt->spinbutton_in),val,val);
    
    val=gtk_spin_button_get_value(GTK_SPIN_BUTTON(mt->spinbutton_out));
    gtk_spin_button_set_range(GTK_SPIN_BUTTON(mt->spinbutton_out),val,val);
  }
}


void mt_sensitise (lives_mt *mt) {
  GtkWidget *eventbox;

  if (mt->event_list!=NULL&&get_first_event(mt->event_list)!=NULL) {
    gtk_widget_set_sensitive (mt->playall,TRUE);
    gtk_widget_set_sensitive (mt->view_events,TRUE);
    gtk_widget_set_sensitive (mt->render, TRUE);
    gtk_widget_set_sensitive (mt->save_event_list, TRUE);
  }
  else {
    gtk_widget_set_sensitive (mt->playall,FALSE);
    gtk_widget_set_sensitive (mt->view_events,FALSE);
    gtk_widget_set_sensitive (mt->render,FALSE);
    gtk_widget_set_sensitive (mt->save_event_list,FALSE);
  }
  gtk_widget_set_sensitive (mt->move_fx, TRUE);
  gtk_widget_set_sensitive (mt->remove_gaps, TRUE);
  gtk_widget_set_sensitive (mt->remove_first_gaps, TRUE);
  if (mt->undoable) gtk_widget_set_sensitive (mt->undo, TRUE);
  if (mt->redoable) gtk_widget_set_sensitive (mt->redo, TRUE);

  gtk_widget_set_sensitive (mt->load_event_list, TRUE);
  gtk_widget_set_sensitive (mt->clipedit,TRUE);
  if (mt->clip_selected>-1) gtk_widget_set_sensitive (mt->insert,TRUE);
  if (mt->video_draws!=NULL) {
    eventbox=g_list_nth_data (mt->video_draws,mt->current_track);
    gtk_widget_set_sensitive (mt->jumpto, g_object_get_data(G_OBJECT(eventbox),"blocks")!=(gulong)NULL);
  }
  if (mt->block_selected) {
    gtk_widget_set_sensitive (mt->delblock, TRUE);
    if (mt->poly_state==POLY_IN_OUT&&mt->block_selected->ordered) {
      gint64 offset_end=mt->block_selected->offset_start+(gint64)(U_SEC/mt->fps)+(get_event_timecode(mt->block_selected->end_event)-get_event_timecode(mt->block_selected->start_event));
      
      set_in_out_spin_ranges(mt,mt->block_selected->offset_start,offset_end);
    }
  }

  track_select(mt);
}


//////////////////////////////////////////////////////////////////
void multitrack_preview_clicked  (GtkWidget *button, gpointer user_data) {
  //preview during rendering
  if (mainw->playing_file==-1) multitrack_playall(button,user_data);
  else on_preview_clicked((GtkButton *)button,NULL);
}




void multitrack_playall (GtkWidget *widget, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;

  if (mainw->current_file<1) return;
  
  mt_desensitise(mt);

  if (mt->mt_frame_preview) {
    // put blank back in preview window
    gtk_widget_hide(mt->preview_button);
    gtk_widget_show(mt->sel_label);
    gtk_widget_ref (mainw->playarea);
    if (mainw->plug1!=NULL) {
      gtk_container_remove (GTK_CONTAINER(mainw->plug1),mainw->image274);
      gtk_widget_destroy (mainw->plug1);
      mainw->plug1=NULL;
    }
    
    if (mainw->playarea->parent!=NULL) gtk_container_remove (GTK_CONTAINER(mainw->playarea->parent),mainw->playarea);
    gtk_container_add (GTK_CONTAINER(mt->play_box),mt->play_blank);
    gtk_container_add (GTK_CONTAINER (mainw->playframe), mainw->playarea);
    gtk_widget_unref (mainw->playarea);
    gtk_widget_unref(mt->play_blank);
    
    mt->mt_frame_preview=FALSE;
  }

  gtk_widget_ref(mt->play_blank);
  gtk_container_remove (GTK_CONTAINER(mt->play_box),mt->play_blank);

  gtk_widget_ref (mainw->playarea);

  if (mainw->plug1!=NULL) {
    gtk_container_remove (GTK_CONTAINER(mainw->plug1),mainw->image274);
    gtk_widget_destroy (mainw->plug1);
    mainw->plug1=NULL;
  }

  if (mainw->playarea->parent!=NULL) gtk_container_remove (GTK_CONTAINER(mainw->playarea->parent),mainw->playarea);
  gtk_box_pack_start (GTK_BOX (mt->play_box), mainw->playarea, TRUE, TRUE, 0);
  gtk_widget_unref (mainw->playarea);

  mainw->xwin=0;

  gtk_widget_set_sensitive (mt->stop,TRUE);
  mt->ptr_time=GTK_RULER(mt->timeline)->position;

  if (mt->is_rendering) {
    // preview during rendering
    on_preview_clicked((GtkButton *)widget,NULL);
    mainw->cancelled=CANCEL_NONE;
  }
  else {
    if (mt->event_list!=NULL) {
      mainw->is_rendering=TRUE;
      on_preview_clicked (NULL,GINT_TO_POINTER(1));
      deinit_render_effects(); // deinit any realtime effects which are left on (e.g. if user pressed "stop")
      mainw->is_rendering=mainw->is_processing=FALSE;
      mainw->cancelled=CANCEL_NONE;
      cfile->event_list=NULL;
      cfile->next_event=NULL;
    }
    time_to_string (mt,0.,TIMECODE_LENGTH);
  }

  gtk_widget_set_sensitive (mt->stop,FALSE);
    
  gtk_widget_ref (mainw->playarea);

  if (mainw->plug1!=NULL) {
    gtk_container_remove (GTK_CONTAINER(mainw->plug1),mainw->image274);
    gtk_widget_destroy (mainw->plug1);
    mainw->plug1=NULL;
  }

  if (mainw->playarea->parent!=NULL) gtk_container_remove (GTK_CONTAINER(mainw->playarea->parent),mainw->playarea);
  gtk_container_add (GTK_CONTAINER(mt->play_box),mt->play_blank);
  gtk_container_add (GTK_CONTAINER (mainw->playframe), mainw->playarea);
  gtk_widget_unref (mainw->playarea);
  gtk_widget_unref(mt->play_blank);

  mainw->xwin=0;

  if (mt->poly_state==POLY_EFFECT) {
    if (!mainw->sep_win||prefs->sepwin_type==0) {
      gboolean sepwin=mainw->sep_win;
      // prepare internal play window for previews
      gtk_widget_hide(mt->sel_label);
      gtk_widget_show(mt->preview_button);
      gtk_widget_ref(mt->play_blank);
      gtk_container_remove (GTK_CONTAINER(mt->play_box),mt->play_blank);
      gtk_widget_ref (mainw->playarea);
      if (mainw->plug1!=NULL) {
	gtk_container_remove (GTK_CONTAINER(mainw->plug1),mainw->image274);
	gtk_widget_destroy (mainw->plug1);
	mainw->plug1=NULL;
      }
      mainw->xwin=0;
      if (mainw->playarea->parent!=NULL) gtk_container_remove (GTK_CONTAINER(mainw->playarea->parent),mainw->playarea);
      gtk_box_pack_start (GTK_BOX (mt->play_box), mainw->playarea, TRUE, TRUE, 0);
      gtk_widget_unref (mainw->playarea);
      
      mainw->sep_win=FALSE; // force showing in internal window
      add_to_playframe();
      mainw->sep_win=sepwin;
      mt->mt_frame_preview=TRUE;
    }
    on_frame_preview_clicked(NULL,mt);
  }
  
  GTK_RULER (mt->timeline)->position=mt->ptr_time;
  gtk_widget_queue_draw (mt->timeline);
  mt_sensitise(mt);
}




void multitrack_insert (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  file *sfile=mainw->files[mt->file_selected];
  gdouble secs=GTK_RULER (mt->timeline)->position/GTK_RULER (mt->timeline)->upper*mt->end_secs;
  GtkWidget *eventbox=g_list_nth_data(mt->video_draws,mt->current_track);
  weed_timecode_t ins_start=(sfile->start-1.)/sfile->fps*U_SEC;
  weed_timecode_t ins_end=(sfile->end)/sfile->fps*U_SEC;
  gboolean did_backup=mt->did_backup;

  if (!did_backup) mt_backup(mt,MT_UNDO_INSERT_BLOCK,NULL);

  if (mt->insert_start!=-1) {
    ins_start=mt->insert_start;
    ins_end=mt->insert_end;
  }

  insert_frames (mt->file_selected,ins_start,ins_end,secs*U_SECL,DIRECTION_POSITIVE,eventbox,mt,NULL);

  if (mt->snapt&&(!did_backup||mt->moving_block)) {
    gdouble oldr_start=mt->region_start;
    gdouble oldr_end=mt->region_end;
    track_rect *block=g_object_get_data(G_OBJECT(eventbox),"block_last");
    GList *tracks_sel=g_list_copy(mt->selected_tracks);
    if (mt->selected_tracks!=NULL) g_list_free(mt->selected_tracks);
    mt->selected_tracks=NULL;
    mt->selected_tracks=g_list_append(mt->selected_tracks,GINT_TO_POINTER(mt->current_track));
    if (block->prev!=NULL) mt->region_start=get_event_timecode(block->prev->end_event)/U_SEC;
    else mt->region_start=0.;
    mt->region_end=get_event_timecode(block->start_event)/U_SEC;
    remove_first_gaps(NULL,mt);
    g_list_free(mt->selected_tracks);
    mt->selected_tracks=g_list_copy(tracks_sel);
    if (tracks_sel!=NULL) g_list_free(tracks_sel);
    mt->region_start=oldr_start;
    mt->region_end=oldr_end;
  }

  mt->did_backup=did_backup;
  
  if (!resize_timeline(mt)) gtk_widget_queue_draw (eventbox); // redraw the track
}

 
void insert_frames (gint filenum, gint64 offset_start, gint64 offset_end, gint64 tc, gshort direction, GtkWidget *eventbox, lives_mt *mt, track_rect *in_block) {
  // insert the selected frames from mainw->files[filenum] from source file filenum into mt->event_list starting at timeline timecode tc
  // if in_block is non-NULL, then we extend (existing) in_block with the new frames; otherwise we create a new block and insert it into eventbox

  // this is quite complex as the framerates of the sourcefile and the timeline might not match. Therefore we resample (in memory) our source file
  // After resampling, we insert resampled frames from offset_start (inclusive) to offset_end (non-inclusive) [forwards]
  // or from offset_start (non-inclusive) to offset_end (inclusive) if going backwards 

  // if we are inserting in an existing block, we can only use this to extend the end (not shrink it)


  // TODO - handle insert before, insert after

  // TODO - handle insert with/without audio

  // TODO - handle case where frames are overwritten


  int numframes,i;
  gint render_file=mainw->current_file;
  gboolean isfirst=TRUE;
  file *sfile=mainw->files[filenum];
  gint track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(eventbox),"layer_number"));
  gint64 last_tc=0,offset_start_tc,start_tc,last_offset;
  int *clips=NULL,*frames=NULL,*rep_clips,*rep_frames,error;
  gint64 time=0;
  weed_plant_t *last_frame_event;
  weed_plant_t *event_list;
  int blank_clip=-1,blank_frame=0;
  weed_plant_t *event,*shortcut1=NULL,*shortcut2=NULL;
  int frame=(offset_start/U_SEC/mt->fps+1.4999);
  gint64 tl=q_dbl(1./mt->fps,mt->fps);
  track_rect *new_block=NULL;
  gchar *text;
  gint64 orig_st=offset_start,orig_end=offset_end;

  mt_desensitise(mt);

  last_offset=offset_start_tc=q_gint64(offset_start,mt->fps);
  offset_end=q_gint64(offset_end,mt->fps);
  start_tc=q_gint64(tc,mt->fps);
  if (direction==DIRECTION_NEGATIVE) tc-=U_SEC/mt->fps;
  last_tc=q_gint64(tc,mt->fps);

  if (direction==DIRECTION_POSITIVE) {
    // fill in blank frames in a gap
    if (mt->event_list!=NULL&&(last_frame_event=get_last_frame_event(mt->event_list))!=NULL) time=get_event_timecode(last_frame_event)+tl;

    for (;time<start_tc;time+=tl) {
      time=q_gint64(time,mt->fps);
      // TODO - memcheck
      event_list=insert_frame_event_at (mt->event_list,time,1,&blank_clip,&blank_frame,&shortcut1);
      weed_set_double_value(event_list,"fps",mainw->files[render_file]->fps);
      if (event_list!=NULL) {
	if (mt->event_list==NULL) {
	  mt->event_list=event_list;
	}
	else if (get_first_event(mt->event_list)==NULL&&get_last_event(mt->event_list)==NULL) {
	  weed_set_plantptr_value(mt->event_list,"first",get_first_event(event_list));
	  weed_set_plantptr_value(mt->event_list,"last",get_last_event(event_list));
	}
      }
    }
  }

  mainw->current_file=filenum;

  if (cfile->fps!=mt->fps&&cfile->event_list==NULL) {
    // resample clip to render fps
    cfile->undo1_dbl=mt->fps;
    on_resample_vid_ok (NULL,NULL);
  }

  mainw->current_file=render_file;

  while (((direction==DIRECTION_POSITIVE&&(offset_start=q_gint64(last_tc-start_tc+offset_start_tc,mt->fps))<offset_end)||(direction==DIRECTION_NEGATIVE&&(offset_start=q_gint64(last_tc+offset_start_tc-start_tc,mt->fps))>=offset_end))) {

    numframes=0;
    
    if ((event=get_frame_event_at (mt->event_list,last_tc,shortcut1,TRUE))!=NULL) {
      // TODO - memcheck
      numframes=weed_leaf_num_elements(event,"clips");
      clips=weed_get_int_array(event,"clips",&error);
      frames=weed_get_int_array(event,"frames",&error);
      shortcut1=event;
    }

    if (numframes<=track) {
      // TODO - memcheck
      rep_clips=(int *)g_malloc(track*sizint+sizint);
      rep_frames=(int *)g_malloc(track*sizint+sizint);

      for (i=0;i<track;i++) {
	if (i<numframes) {
	  rep_clips[i]=clips[i];
	  rep_frames[i]=frames[i];
	}
	else {
	  rep_clips[i]=-1;
	  rep_frames[i]=0;
	}
      }
      numframes=track+1;
    }
    else {
      if (mt->insert_mode==INSERT_MODE_NORMAL&&frames[track]>0) {
	if (in_block==NULL&&new_block!=NULL) {
	  if (direction==DIRECTION_POSITIVE) {
	    new_block->max_tc=last_tc;
	    shortcut1=get_prev_frame_event(shortcut1);
	  }
	  else new_block->min_tc=last_tc+tl;
	}
	g_free(clips);
	g_free(frames);
	break; // do not allow overwriting in this mode
      }
      rep_clips=clips;
      rep_frames=frames;
    }

    if (sfile->event_list!=NULL) event=get_frame_event_at(sfile->event_list,offset_start,shortcut2,TRUE);
    if (sfile->event_list!=NULL&&event==NULL) break; // insert finished: ran out of frames in resampled clip
    last_offset=offset_start;
    if (sfile->event_list!=NULL) {
      // frames were resampled, get new frame at the source file timecode
      frame=weed_get_int_value(event,"frames",&error);
      if (direction==DIRECTION_POSITIVE) shortcut2=event;
      else shortcut2=get_prev_frame_event(event); // TODO : this is not optimal for the first frame
    }
    rep_clips[track]=filenum;
    rep_frames[track]=frame;

    // TODO - memcheck
    event_list=insert_frame_event_at (mt->event_list,last_tc,numframes,rep_clips,rep_frames,&shortcut1);
    if (isfirst) {
      // TODO - memcheck
      if (in_block==NULL) new_block=add_block_start_point (eventbox,last_tc,filenum,offset_start,shortcut1,TRUE);
      isfirst=FALSE;
    }
    if (event_list!=NULL) {
      weed_set_double_value(event_list,"fps",mainw->files[render_file]->fps);
      if (mt->event_list==NULL) mt->event_list=event_list;
      else if (get_first_event(mt->event_list)==NULL&&get_last_event(mt->event_list)==NULL) {
	weed_set_plantptr_value(mt->event_list,"first",get_first_event(event_list));
	weed_set_plantptr_value(mt->event_list,"last",get_last_event(event_list));
      }
    }

    g_free(clips);
    g_free(frames);
    
    clips=rep_clips=NULL;
    frames=rep_frames=NULL;
    if (direction==DIRECTION_POSITIVE) last_tc+=tl;
    else {
      if (last_tc<tl) break;
      last_tc-=tl;
    }
    last_tc=q_gint64(last_tc,mt->fps);
    if (sfile->event_list==NULL) if ((direction==DIRECTION_POSITIVE&&(++frame>sfile->frames))||(direction==DIRECTION_NEGATIVE&&(--frame<1))) break;
  }

  if (!isfirst) {
    if (direction==DIRECTION_POSITIVE) {
      if (in_block!=NULL) g_object_set_data (G_OBJECT(eventbox),"block_last",(gpointer)in_block);
      add_block_end_point(eventbox,shortcut1);
    }
    else if (in_block!=NULL) {
      in_block->offset_start=last_offset;
      in_block->start_event=shortcut1;
    }
  }

  mt->last_direction=direction;

  if (in_block==NULL) {
    text=g_strdup_printf(_("Inserted frames %d to %d from clip %s into track %s from time %.4f to %.4f\n"),sfile->start,sfile->end,sfile->name,get_track_name(mt,mt->current_track),(orig_st+start_tc)/U_SEC,(orig_end+start_tc)/U_SEC);
    d_print(text);
    g_free(text);
  }

  mt_sensitise(mt);
}


void multitrack_view_events (GtkMenuItem *menuitem, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  GtkWidget *elist_dialog=create_event_list_dialog(mt->event_list);
  gtk_dialog_run(GTK_DIALOG(elist_dialog));
  gtk_widget_destroy(elist_dialog);
}

////////////////////////////////////////////////////////
// region functions

void draw_region (lives_mt *mt) {
  gdouble start,end;
  if (mt->region_start==mt->region_end) return;

  gtk_widget_queue_draw(mt->timeline);
  gdk_window_process_updates(mt->timeline_eb->window,FALSE);

  if (mt->region_start<mt->region_end) {
    start=mt->region_start;
    end=mt->region_end;
  }
  else {
    start=mt->region_end;
    end=mt->region_start;
  }

  gdk_draw_rectangle (GDK_DRAWABLE(mt->timeline->window), mainw->LiVES->style->black_gc, TRUE, start*mt->timeline->allocation.width/mt->end_secs, 0, (end-start)*mt->timeline->allocation.width/mt->end_secs,mt->timeline->allocation.height-1);
}

gint expose_timeline_event (GtkWidget *timeline, GdkEventExpose *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  if (event->count>0) return FALSE;
  if (mainw->playing_file>-1||mt->is_rendering) return FALSE;
  draw_region (mt);
  return FALSE;
}


////////////////////////////////////////////////////

// functions for moving and clicking on the timeline


gboolean
on_timeline_update (GtkWidget *widget, GdkEventMotion *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  gint x;
  gdouble pos;

  if (!mt->region_updating) return TRUE;
  if (mainw->playing_file>-1) return TRUE;

  gdk_window_get_pointer(GDK_WINDOW (widget->window), &x, NULL, NULL);

  if (x<0) x=0;

  pos=x*mt->end_secs/widget->allocation.width;
  if (pos>mt->end_secs) pos=mt->end_secs;
  pos=(gint)(pos*mt->fps+.5)/mt->fps; // quantise to nearest frame

  mt->region_end=pos;
  if (mt->region_end<0.) mt->region_end=0.;

  if (mt->region_end!=mt->region_start) draw_region (mt);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_start),mt->region_start);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_end),mt->region_end);

  if (mt->mouse_mode==MOUSE_MODE_SELECT&&mt->tl_selecting&&event!=NULL) mouse_select_move(widget,mt);

  return TRUE;
}



gboolean all_present (weed_plant_t *event, GList *sel) {
  int *clips;
  int *frames;
  int error,numclips,layer;
  // see if we have an actual clip/frame for each layer in sel
  if (event==NULL||sel==NULL) return FALSE;

  numclips=weed_leaf_num_elements(event,"clips");

  if (!numclips) return FALSE;

  clips=weed_get_int_array(event,"clips",&error);
  frames=weed_get_int_array(event,"frames",&error);

  while (sel!=NULL) {
    layer=GPOINTER_TO_INT(sel->data);
    if (layer>=numclips||clips[layer]<1||frames[layer]<1) {
      weed_free(clips);
      weed_free(frames);
      return FALSE;
    }
    sel=sel->next;
  }
  weed_free(clips);
  weed_free(frames);
  return TRUE;
}


void get_region_overlap(lives_mt *mt) {
  // get region which overlaps all selected tracks
  weed_plant_t *event;
  gint64 tc;

  if (mt->selected_tracks==NULL||mt->event_list==NULL) {
    mt->region_start=mt->region_end=0.;
    return;
  }

  tc=q_gint64(mt->region_start*U_SECL,mt->fps);
  event=get_frame_event_at(mt->event_list,tc,NULL,TRUE);

  while (all_present(event,mt->selected_tracks)) {
    // move start to left
    event=get_prev_frame_event(event);
  }

  if (event==NULL) {
    event=get_first_event(mt->event_list);
    if (!IS_FRAME_EVENT(event)) event=get_next_frame_event(event);
  }

  while (event!=NULL&&!all_present(event,mt->selected_tracks)) {
    event=get_next_frame_event(event);
  }

  if (event==NULL) mt->region_start=0.;
  else mt->region_start=get_event_timecode(event)/U_SEC;

  tc=q_gint64(mt->region_end*U_SECL,mt->fps);
  event=get_frame_event_at(mt->event_list,tc,NULL,TRUE);

  while (all_present(event,mt->selected_tracks)) {
    // move end to right
    event=get_next_frame_event(event);
  }

  if (event==NULL) {
     event=get_last_event(mt->event_list);
    if (!IS_FRAME_EVENT(event)) event=get_prev_frame_event(event);
  }
 
  while (event!=NULL&&!all_present(event,mt->selected_tracks)) {
    event=get_prev_frame_event(event);
  }

  if (event==NULL) mt->region_end=0.;
  mt->region_end=get_event_timecode(event)/U_SEC+1./mt->fps;

}


void do_sel_context (lives_mt *mt) {
  gchar *msg;
  if (mt->region_start==mt->region_end||mt->did_backup) return;
  clear_context(mt);
  msg=g_strdup_printf(_("Time region %.3f to %.3f selected.\n"),mt->region_start,mt->region_end);
  add_context_label(mt,msg);
  g_free(msg);
  msg=g_strdup_printf(_("%d video tracks selected.\n"),g_list_length(mt->selected_tracks));
  add_context_label (mt,msg);
  g_free(msg);
}



gboolean
on_timeline_release (GtkWidget *eventbox, GdkEventButton *event, gpointer user_data) {
  //button release
  lives_mt *mt=(lives_mt *)user_data;
  gint x;
  gdouble pos=mt->region_end;

  if (mainw->playing_file>-1) return FALSE;

  if (event!=NULL) {
    gdk_window_get_pointer(GDK_WINDOW (eventbox->window), &x, NULL, NULL);
    if (x<0) x=0;
    
    pos=x*mt->end_secs/eventbox->allocation.width;
    if (pos>mt->end_secs) pos=mt->end_secs;
    pos=(gint)(pos*mt->fps+.5)/mt->fps; // quantise to nearest frame
    
    mt->region_end=pos;

    mt->region_updating=FALSE;
  }

  if (mt->region_end!=mt->region_start) {
    if (mt->snap_over) get_region_overlap(mt);
    if (mt->region_end<mt->region_start) {
      mt->region_start-=mt->region_end;
      mt->region_end+=mt->region_start;
      mt->region_start=mt->region_end-mt->region_start;
    }
    draw_region (mt);
    if (mt->region_end>0.) gtk_widget_set_sensitive (mt->fx_region, TRUE);
    else gtk_widget_set_sensitive (mt->fx_region, FALSE);
    if (mt->region_start==mt->region_end) gtk_widget_queue_draw (mt->timeline);
  }
  else {
    GTK_RULER (mt->timeline)->position=pos;
    gtk_widget_queue_draw (mt->timeline);
    gtk_widget_set_sensitive (mt->fx_region, FALSE);
  }

  gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_start),mt->region_start);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->spinbutton_end),mt->region_end);

  if (mt->mouse_mode==MOUSE_MODE_SELECT&&event!=NULL) mouse_select_end(eventbox,mt);

  do_sel_context(mt);

  return TRUE;
}


gboolean
on_timeline_press (GtkWidget *widget, GdkEventButton *event, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  gint x;
  gdouble pos;

  if (mainw->playing_file>-1) return FALSE;

  gdk_window_get_pointer(GDK_WINDOW (widget->window), &x, NULL, NULL);
  if (x<0) x=0;

  pos=x*mt->end_secs/widget->allocation.width;
  if (pos>mt->end_secs) pos=mt->end_secs;
  pos=(gint)(pos*mt->fps+.5)/mt->fps; // quantise to nearest frame

  if (pos<0.) pos=0.;

  mt->region_start=mt->region_end=pos;

  GTK_RULER (mt->timeline)->position=pos;
  gtk_widget_queue_draw (mt->timeline);

  if (mt->mouse_mode==MOUSE_MODE_SELECT) mouse_select_start(widget,mt);

  mt->region_updating=TRUE;

  return TRUE;
}






static weed_timecode_t get_prev_node_tc(lives_mt *mt, weed_timecode_t tc) {
  int num_params=weed_leaf_num_elements(get_weed_filter(mt->current_fx),"in_parameter_templates");
  int i,error;
  weed_timecode_t prev_tc=-1;
  weed_plant_t *event;
  weed_timecode_t ev_tc;

  for (i=0;i<num_params;i++) {
    event=pchain[i];
    while (event!=NULL&&(ev_tc=get_event_timecode(event))<tc) {
      if (ev_tc>prev_tc) prev_tc=ev_tc;
      event=weed_get_voidptr_value(event,"next_change",&error);
    }
  }
  return prev_tc;
}


static weed_timecode_t get_next_node_tc(lives_mt *mt, weed_timecode_t tc) {
  int num_params=weed_leaf_num_elements(get_weed_filter(mt->current_fx),"in_parameter_templates");
  int i,error;
  weed_timecode_t next_tc=-1;
  weed_plant_t *event;
  weed_timecode_t ev_tc;

  for (i=0;i<num_params;i++) {
    event=pchain[i];
    while (event!=NULL&&(ev_tc=get_event_timecode(event))<=tc) event=weed_get_voidptr_value(event,"next_change",&error);
    if (event!=NULL) {
      if (next_tc==-1||ev_tc<next_tc) next_tc=ev_tc;
    }
  }
  return next_tc;
}


static gboolean is_node_tc(lives_mt *mt, weed_timecode_t tc) {
  int num_params=weed_leaf_num_elements(get_weed_filter(mt->current_fx),"in_parameter_templates");
  int i,error;
  weed_plant_t *event;
  weed_timecode_t ev_tc;

  for (i=0;i<num_params;i++) {
    event=pchain[i];
    ev_tc=-1;
    while (event!=NULL&&(ev_tc=get_event_timecode(event))<tc) event=weed_get_voidptr_value(event,"next_change",&error);
    if (ev_tc==tc) return TRUE;
  }
  return FALSE;
}



void
on_frame_preview_clicked (GtkButton *button, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  weed_timecode_t tc=gtk_spin_button_get_value(GTK_SPIN_BUTTON(mt->node_spinbutton))+get_event_timecode(mt->block_selected->start_event);
  show_preview(mt,tc);
}


// apply the param changes and update widgets
void
on_node_spin_value_changed           (GtkSpinButton   *spinbutton,
				      gpointer         user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  weed_timecode_t tc=q_gint64(gtk_spin_button_get_value(spinbutton)*U_SEC+get_event_timecode(mt->block_selected->start_event),mt->fps);

  interpolate_params(mt->current_rfx->source,pchain,tc);

  update_visual_params(mt->current_rfx,FALSE);

  if (get_prev_node_tc(mt,tc)>-1) gtk_widget_set_sensitive(mt->prev_node_button,TRUE);
  else gtk_widget_set_sensitive(mt->prev_node_button,FALSE);

  if (get_next_node_tc(mt,tc)>-1) gtk_widget_set_sensitive(mt->next_node_button,TRUE);
  else gtk_widget_set_sensitive(mt->next_node_button,FALSE);

  if (gtk_spin_button_get_value(spinbutton)>0.&&is_node_tc(mt,tc)) gtk_widget_set_sensitive(mt->del_node_button,TRUE);
  else gtk_widget_set_sensitive(mt->del_node_button,FALSE);

}

// node buttons
void on_next_node_clicked  (GtkWidget *button, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  weed_timecode_t tc=gtk_spin_button_get_value(GTK_SPIN_BUTTON(mt->node_spinbutton))*U_SEC+get_event_timecode(mt->block_selected->start_event);
  weed_timecode_t next_tc=get_next_node_tc(mt,tc);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->node_spinbutton),(next_tc-get_event_timecode(mt->block_selected->start_event))/U_SEC);
}


void on_prev_node_clicked  (GtkWidget *button, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  weed_timecode_t tc=gtk_spin_button_get_value(GTK_SPIN_BUTTON(mt->node_spinbutton))*U_SEC+get_event_timecode(mt->block_selected->start_event);
  weed_timecode_t prev_tc=get_prev_node_tc(mt,tc);
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(mt->node_spinbutton),(prev_tc-get_event_timecode(mt->block_selected->start_event))/U_SEC);
}


void on_del_node_clicked  (GtkWidget *button, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  int num_params=weed_leaf_num_elements(get_weed_filter(mt->current_fx),"in_parameter_templates");
  int i,error;
  weed_plant_t *event;
  weed_timecode_t ev_tc;
  weed_timecode_t tc=gtk_spin_button_get_value(GTK_SPIN_BUTTON(mt->node_spinbutton))*U_SEC+get_event_timecode(mt->block_selected->start_event);
  weed_plant_t *prev_pchange,*next_pchange;
  gchar *filter_name,*text;
  int selected_track;

  for (i=0;i<num_params;i++) {
    event=pchain[i];
    ev_tc=-1;
    while (event!=NULL&&(ev_tc=get_event_timecode(event))<tc) event=weed_get_voidptr_value(event,"next_change",&error);
    if (ev_tc==tc) {
      prev_pchange=weed_get_voidptr_value(event,"prev_change",&error);
      next_pchange=weed_get_voidptr_value(event,"next_change",&error);
      delete_event(mt->event_list,event);
      if (prev_pchange!=NULL) weed_set_voidptr_value(prev_pchange,"next_change",next_pchange);
      if (next_pchange!=NULL) weed_set_voidptr_value(next_pchange,"prev_change",prev_pchange);
    }
  }
  filter_name=weed_filter_get_name(mt->current_fx);
  selected_track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(mt->block_selected->eventbox),"layer_number"));
  text=g_strdup_printf(_("Removed parameter values for effect %s on track %s at time %.4f\n"),filter_name,get_track_name(mt,selected_track),tc);
  d_print(text);
  g_free(text);
  g_free(filter_name);
  on_node_spin_value_changed(GTK_SPIN_BUTTON(mt->node_spinbutton),(gpointer)mt);
}


static void add_to_pchain(weed_plant_t *pchange, int index, weed_timecode_t tc) {
  weed_plant_t *event=pchain[index];
  weed_plant_t *last_event=NULL;
  int error;

  while (event!=NULL&&get_event_timecode(event)<tc) {
    last_event=event;
    event=weed_get_voidptr_value(event,"next_change",&error);
  }
  if (last_event==NULL) return;
  weed_set_voidptr_value(pchange,"next_change",weed_get_voidptr_value(last_event,"next_change",&error));
  weed_set_voidptr_value(last_event,"next_change",pchange);
  weed_set_voidptr_value(pchange,"prev_change",last_event);
}





void on_set_pvals_clicked  (GtkWidget *button, gpointer user_data) {
  lives_mt *mt=(lives_mt *)user_data;
  weed_plant_t *inst=(weed_plant_t *)mt->current_rfx->source;
  weed_plant_t *param,*pchange,*at_event;
  gchar *filter_name,*text;
  gint selected_track;
  int i;
  weed_timecode_t tc=q_gint64(gtk_spin_button_get_value(GTK_SPIN_BUTTON(mt->node_spinbutton))*U_SEC,mt->fps);

  for (i=0;((param=weed_inst_in_param(inst,i,FALSE))!=NULL);i++) {
    pchange=weed_plant_new(WEED_PLANT_EVENT);
    weed_set_int_value(pchange,"hint",WEED_EVENT_HINT_PARAM_CHANGE);
    weed_set_int64_value(pchange,"timecode",tc);
    weed_set_voidptr_value(pchange,"init_event",mt->init_event);
    weed_set_int_value(pchange,"index",i);
    weed_leaf_copy(param,"value",pchange,"value");

    // set next_change, prev_change
    add_to_pchain(pchange,i,tc);

    at_event=get_frame_event_at(mt->event_list,tc,mt->init_event,TRUE);
    insert_param_change_event_at(mt->event_list,at_event,pchange);

  }
  filter_name=weed_filter_get_name(mt->current_fx);
  selected_track=GPOINTER_TO_INT(g_object_get_data(G_OBJECT(mt->block_selected->eventbox),"layer_number"));
  text=g_strdup_printf(_("Set parameter values for effect %s on track %s at time %.4f\n"),filter_name,get_track_name(mt,selected_track),tc/U_SEC);
  d_print(text);
  g_free(text);
  g_free(filter_name);
}


//////////////////////////////////////////////////////////////////////////
static int free_tt_key;
static int elist_errors;

static int *update_layout_map(weed_plant_t *event_list) {
  // update our current layout map with the current layout
  // returns an int * of maximum frame used for each clip that exists, 0 means unused
  int *used_clips;
  weed_plant_t *event=get_first_event(event_list);
  int i;

  used_clips=g_malloc((MAX_FILES+1)*sizint);
  for (i=1;i<=MAX_FILES;i++) used_clips[i]=0;

  while (event!=NULL) {
    if (WEED_EVENT_IS_FRAME(event)) {
      int numtracks=weed_leaf_num_elements(event,"clips");
      if (numtracks>0) {
	int i,error;
	int *clip_index=weed_get_int_array(event,"clips",&error);
	int *frame_index=weed_get_int_array(event,"frames",&error);
	for (i=0;i<numtracks;i++) {
	  if (clip_index[i]>0&&(frame_index[i]>used_clips[clip_index[i]])) used_clips[clip_index[i]]=frame_index[i];
	}
	weed_free(clip_index);
	weed_free(frame_index);
      }
    }
    event=get_next_event(event);
  }
  return used_clips;
}





GList *load_layout_map(void) {
  // load in a layout map for the set, create mainw->files[n]->layout_map
  // optionally, report any missing clips that are used by layouts
  int fd;
  gchar *lmap_name=g_strdup_printf("%s/%s/layouts/layout.map",prefs->tmpdir,mainw->set_name);
  GList *lmap=NULL;
  gchar *handle;
  gint64 unique_id;
  gchar *name;
  layout_map *lmap_entry;
  size_t bytes;
  int len,nm,i;
  gchar *entry;
  gchar *string;
  gchar **array;

  if (!g_file_test(lmap_name,G_FILE_TEST_EXISTS)) return NULL;

  fd=open(lmap_name,O_RDONLY);

  if (fd<0) return NULL;

  while (1) {
    bytes=read(fd,&len,sizint);
    if (bytes<sizint) {
      close(fd);
      break;
    }
    handle=g_malloc(len+1);
    bytes=read(fd,handle,len);
    if (bytes<len) {
      close(fd);
      break;
    }
    memset(handle+len,0,1);
    bytes=read(fd,&unique_id,8);
    if (bytes<8) {
      close(fd);
      break;
    }
    bytes=read(fd,&len,sizint);
    if (bytes<sizint) {
      close(fd);
      break;
    }
    name=g_malloc(len+1);
    bytes=read(fd,name,len);
    if (bytes<len) {
      close(fd);
      break;
    }
    memset(name+len,0,1);
    bytes=read(fd,&nm,sizint);
    if (bytes<sizint) {
      close(fd);
      break;
    }

    lmap_entry=(layout_map *)g_malloc(sizeof(layout_map));
    lmap_entry->handle=handle;
    lmap_entry->unique_id=unique_id;
    lmap_entry->name=name;
    lmap_entry->list=NULL;

    for (i=0;i<nm;i++) {
      bytes=read(fd,&len,sizint);
      if (bytes<sizint) {
	close(fd);
	break;
      }
      entry=g_malloc(len+1);
      bytes=read(fd,entry,len);
      if (bytes<len) {
	close(fd);
	break;
      }
      memset(entry+len,0,1);
      string=repl_tmpdir(entry,FALSE); // allow relocation of tmpdir
      lmap_entry->list=g_list_append(lmap_entry->list,string);
      array=g_strsplit(string,"|",-1);
      if (!g_list_find(mainw->current_layouts_map,array[0])) {
	mainw->current_layouts_map=g_list_append(mainw->current_layouts_map,g_strdup(array[0]));
      }
      g_strfreev(array);
      g_free(entry);
    }
    lmap=g_list_append(lmap,lmap_entry);
  }
  return lmap;
}




void save_layout_map (int *lmap, gchar *file, gchar *dir) {
  // in the file layout.map, we map each clip in the set to which layouts (if any) it is used in
  // we also record the highest frame number used, the clip number in the layout, and the current fps of the clip
  // we identify clips using the handle and unique_id
  int i;
  GList *map,*map_next;
  gchar *new_entry;
  gchar *map_name,*ldir;
  int fd;
  int len;
  gchar *string;

  if (file!=NULL&&!g_list_find(mainw->current_layouts_map,file)) mainw->current_layouts_map=g_list_append(mainw->current_layouts_map,g_strdup(file));
  if (dir==NULL) ldir=g_strdup_printf("%s/%s/layouts/",prefs->tmpdir,mainw->set_name);
  else ldir=g_strdup(dir);

  map_name=g_strdup_printf("%slayout.map",ldir);
  fd=creat(map_name,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
  g_free(map_name);
  g_free(ldir);

  for (i=1;i<=MAX_FILES;i++) {
    // add or update
    if (mainw->files[i]!=NULL) {

      if (mainw->files[i]->layout_map!=NULL) {
	map=mainw->files[i]->layout_map;
	while (map!=NULL) {
	  map_next=map->next;
	  if (map->data!=NULL) {
	    gchar **array=g_strsplit(map->data,"|",-1);
	    if ((file!=NULL&&!strcmp(array[0],file))||(file==NULL&&dir==NULL&&!g_file_test(array[0],G_FILE_TEST_EXISTS))) {
	      // remove prior entry
	      g_free(map->data);
	      if (map->prev!=NULL) map->prev->next=map_next;
	      else mainw->files[i]->layout_map=map_next;
	      if (map_next!=NULL) map_next->prev=map->prev;
	      map->next=map->prev=NULL;
	      g_list_free(map);
	      break;
	    }
	    g_strfreev(array);
	  }
	  map=map_next;
	}
      }
	
      if (file!=NULL&&lmap!=NULL&&lmap[i]!=0) {
	new_entry=g_strdup_printf("%s|%d|%d|%.8f",file,i,lmap[i],mainw->files[i]->fps);
	mainw->files[i]->layout_map=g_list_prepend(mainw->files[i]->layout_map,new_entry);
      }

      if ((map=mainw->files[i]->layout_map)!=NULL) {
	len=strlen(mainw->files[i]->handle);
	write(fd,&len,sizint);
	write(fd,mainw->files[i]->handle,len);
	write(fd,&mainw->files[i]->unique_id,8);
	len=strlen(mainw->files[i]->name);
	write(fd,&len,sizint);
	write(fd,mainw->files[i]->name,len);
	len=g_list_length(map);
	write(fd,&len,sizint);
	while (map!=NULL) {
	  string=repl_tmpdir(map->data,TRUE); // allow relocation of tmpdir
	  len=strlen(string);
	  write(fd,&len,sizint);
	  write(fd,string,len);
	  g_free(string);
	  map=map->next;
	}
      }
    }
  }
  close(fd);
}





void add_markers(lives_mt *mt, weed_plant_t *event_list) {
  GList *track_blocks=NULL;
  GList *tlist=mt->video_draws;
  GList *blist;
  track_rect *block;
  weed_timecode_t tc;
  GtkWidget *eventbox;
  weed_plant_t *event;
  gint track;

  while (tlist!=NULL) {
    eventbox=tlist->data;
    block=(track_rect *)g_object_get_data (G_OBJECT(eventbox), "blocks");
    track_blocks=g_list_append(track_blocks,(gpointer)block);
    tlist=tlist->next;
  }

  event=get_first_event(event_list);

  while (event!=NULL) {
    if (WEED_EVENT_IS_FRAME(event)) {
      tc=get_event_timecode(event);
      blist=track_blocks;
      track=0;
      while (blist!=NULL) {
	block=blist->data;
	if (block!=NULL) {
	  if (block->prev!=NULL&&(get_event_timecode(block->prev->end_event)==q_gint64(tc-U_SEC/mt->fps,mt->fps))&&(tc==get_event_timecode(block->start_event))&&(get_frame_event_clip(block->prev->end_event,track)==get_frame_event_clip(block->start_event,track))) {
	    insert_marker_event_at(mt->event_list,event,EVENT_MARKER_BLOCK_START,GINT_TO_POINTER(track));
	  }
	  if (!block->ordered) insert_marker_event_at(mt->event_list,event,EVENT_MARKER_BLOCK_UNORDERED,GINT_TO_POINTER(track));
	  if (event==block->end_event) blist->data=block->next;
	}
	track++;
	blist=blist->next;
      }
    }
    event=get_next_event(event);
  }
}






void on_save_event_list_activate (GtkMenuItem *menuitem, gpointer user_data) {
  int fd;
  lives_mt *mt=(lives_mt *)user_data;
  gchar *esave_dir;
  gchar *esave_file;
  gchar *msg;
  gboolean response;
  gchar *filt[]={"*.lay",NULL};
  gboolean was_set=mainw->was_set;
  gchar *com;
  int *layout_map;

  // update layout map
  layout_map=update_layout_map(mt->event_list);

  if (mainw->scrap_file!=-1&&layout_map[mainw->scrap_file]!=0) {
    // can't save if we have generated frames
    do_layout_scrap_file_error(GTK_WINDOW(mt->window));
    g_free(layout_map);
    mainw->cancelled=CANCEL_USER;
    return;
  }

  if (mainw->was_set) {
    weed_set_string_value(mt->event_list,"needs_set",mainw->set_name);
  }
  else {
    gchar new_set_name[256];
    do {
      // prompt for a set name, advise user to save set
      renamew=create_rename_dialog(4);
      gtk_widget_show(renamew->dialog);
      response=gtk_dialog_run(GTK_DIALOG(renamew->dialog));
      if (response==GTK_RESPONSE_CANCEL) {
	gtk_widget_destroy(renamew->dialog);
	g_free(renamew);
	mainw->cancelled=CANCEL_USER;
	return;
      }
      g_snprintf(new_set_name,256,"%s",gtk_entry_get_text (GTK_ENTRY (renamew->entry)));
      gtk_widget_destroy(renamew->dialog);
      g_free(renamew);
      while (g_main_context_iteration(NULL,FALSE));
    } while (!is_legal_set_name(new_set_name,FALSE));
    g_snprintf(mainw->set_name,256,"%s",new_set_name);
  }

  esave_dir=g_strdup_printf("%s/%s/layouts/",prefs->tmpdir,mainw->set_name);
  com=g_strdup_printf ("%s %s",capable->mkdirhier,esave_dir);
  system (com);
  g_free (com);

  if (!strlen(mt->layout_name)) esave_file=choose_file(esave_dir,NULL,filt,GTK_FILE_CHOOSER_ACTION_SAVE);
  else esave_file=choose_file(esave_dir,mt->layout_name,filt,GTK_FILE_CHOOSER_ACTION_SAVE);

  if (esave_file==NULL) {
    g_free(esave_file);
    g_free(esave_dir);
    g_free(layout_map);
    mainw->was_set=was_set;
    if (!was_set) memset(mainw->set_name,0,1);
    mainw->cancelled=CANCEL_USER;
    return;
  }

  esave_file=ensure_extension(esave_file,".lay");

  g_snprintf(mt->layout_name,256,"%s",esave_file);
  get_basename(mt->layout_name);

  add_markers(mt,mt->event_list);

  fd=creat(esave_file,S_IRUSR|S_IWUSR);

  save_event_list_inner(fd,mt->event_list,NULL);

  close(fd);
  msg=g_strdup_printf(_("Saved layout to %s\n"),esave_file);
  d_print(msg);
  g_free(msg);

  // save layout map
  save_layout_map(layout_map,esave_file,esave_dir);

  mt->changed=FALSE;

  g_free(esave_file);
  g_free(esave_dir);
  g_free(layout_map);
}


static gchar *rec_error_add(gchar *ebuf, gchar *msg, int num, weed_timecode_t tc) {
  gchar *tmp;
  gchar *new;

  elist_errors++;

  if (tc==-1) new=g_strdup(msg); // missing timecode
  else {
    if (capable->cpu_bits==32) {
      if (num==-1) new=g_strdup_printf("%s at timecode %lld\n",msg,tc);
      else new=g_strdup_printf("%s %d at timecode %lld\n",msg,num,tc);
    }
    else {
      if (num==-1) new=g_strdup_printf("%s at timecode %ld\n",msg,tc);
      else new=g_strdup_printf("%s %d at timecode %ld\n",msg,num,tc);
    }
  }
  tmp=g_strconcat (ebuf,new,NULL);
#ifndef SILENT_EVENT_LIST_LOAD // TODO
  g_printerr("Rec error: %s",new);
#endif
  g_free(ebuf);
  g_free(new);
  return tmp;
} 



static int get_next_tt_key(ttable *trans_table) {
  int i;
  for (i=free_tt_key;i<FX_KEYS_MAX-FX_KEYS_MAX_VIRTUAL;i++) {
    if (trans_table[i].in==NULL) return i;
  }
  return -1;
}

static void *find_init_event_in_ttable(ttable *trans_table, weed_plant_t *in, gboolean normal) {
  int i;
  for (i=0;i<FX_KEYS_MAX-FX_KEYS_MAX_VIRTUAL;i++) {
    if (normal&&trans_table[i].in==in) return trans_table[i].out;
    if (!normal&&trans_table[i].out==in) return trans_table[i].in; // reverse lookup for past filter_map check
    if (trans_table[i].out==NULL) return NULL;
  }
  return NULL;
}


static void **remove_nulls_from_filter_map(void **init_events, int num_events) {
  // remove NULLs from filter_map init_events
  int num_nulls=0,i,j=0;
  void **new_init_events;

  if (num_events==1) return init_events;

  for (i=0;i<num_events;i++) if (init_events[i]==NULL) num_nulls++;
  if (num_nulls==0) return init_events;

  new_init_events=g_malloc((num_events-num_nulls)*sizeof(void *));
  
  for (i=0;i<num_events;i++) if (init_events[i]==NULL) new_init_events[j++]=init_events[i];

  return new_init_events;
}


gboolean compare_filter_maps(weed_plant_t *fm1, weed_plant_t *fm2) {
  // return TRUE if the maps match exactly
  int i,error,num_events;
  void **inits1,**inits2;

  if (!weed_plant_has_leaf(fm1,"init_events")&&!weed_plant_has_leaf(fm2,"init_events")) return TRUE;
  if ((!weed_plant_has_leaf(fm1,"init_events")&&weed_get_voidptr_value(fm2,"init_events",&error)!=NULL)||(!weed_plant_has_leaf(fm2,"init_events")&&weed_get_voidptr_value(fm1,"init_events",&error)!=NULL)) return FALSE;

  num_events=weed_leaf_num_elements(fm1,"init_events");
  if (num_events!=weed_leaf_num_elements(fm2,"init_events")) return FALSE;

  inits1=weed_get_voidptr_array(fm1,"init_events",&error);
  inits2=weed_get_voidptr_array(fm2,"init_events",&error);

  for (i=0;i<num_events;i++) {
    if (inits1[i]!=inits2[i]) {
      weed_free(inits1);
      weed_free(inits2);
      return FALSE;
    }
  }
  weed_free(inits1);
  weed_free(inits2);
  return TRUE;
}

static gchar *filter_map_check(ttable *trans_table,weed_plant_t *filter_map, weed_timecode_t deinit_tc, weed_timecode_t fm_tc, gchar *ebuf) {
  int num_init_events;
  void **init_events;
  void **copy_events;
  int error,i;

  if (!weed_plant_has_leaf(filter_map,"init_events")) return ebuf;
  // check no deinited events are active
  num_init_events=weed_leaf_num_elements(filter_map,"init_events");
  if (num_init_events==1&&weed_get_voidptr_value(filter_map,"init_events",&error)==NULL) return ebuf;
  init_events=weed_get_voidptr_array(filter_map,"init_events",&error);
  copy_events=g_malloc(num_init_events*sizeof(weed_plant_t *));
  for (i=0;i<num_init_events;i++) {
    if (find_init_event_in_ttable(trans_table,init_events[i],FALSE)!=NULL) copy_events[i]=init_events[i];
    else {
      copy_events[i]=NULL;
      ebuf=rec_error_add(ebuf,"Filter_map points to invalid filter_init",-1,fm_tc);
    }
  }
  if (num_init_events>1) copy_events=remove_nulls_from_filter_map(copy_events,num_init_events);


  // TODO - create new filter_map, insert it at last_deinit_tc, it should replace any post-frame filter map

  //weed_set_plantptr_array(filter_map,"init_events",copy_events);
  g_free(copy_events);
  weed_free(init_events);
  return ebuf;
}


static gchar *add_filter_deinits(weed_plant_t *event_list, ttable *trans_table, void ***pchains, weed_timecode_t tc, gchar *ebuf) {
  // add filter deinit events for any remaining active filters
  int i,error,num_params;
  gchar *filter_hash;
  gint idx;
  weed_plant_t *filter,*init_event,*event;
  void **in_pchanges;

  for (i=0;i<FX_KEYS_MAX-FX_KEYS_MAX_VIRTUAL;i++) {
    if (trans_table[i].out==NULL) return ebuf;
    if (trans_table[i].in!=NULL) {
      append_filter_deinit_event(event_list,tc,(init_event=trans_table[i].out),pchains[i]);
      event=get_last_event(event_list);

      filter_hash=weed_get_string_value(init_event,"filter",&error);
      if ((idx=weed_get_idx_for_hashname(filter_hash))!=-1) {
	filter=get_weed_filter(idx);
	if ((num_params=num_in_params(filter,TRUE,TRUE))>0) {
	  in_pchanges=g_malloc(num_params*sizeof(void *));
	  for (i=0;i<num_params;i++) {
	    if (!WEED_EVENT_IS_FILTER_INIT(pchains[idx][i])) in_pchanges[i]=pchains[idx][i];
	    else in_pchanges[i]=NULL;
	  }
	  weed_set_voidptr_array(event,"in_parameters",num_params,in_pchanges); // set array to last param_changes
	  g_free(in_pchanges);
	  g_free(pchains[idx]);
	}
      }

      ebuf=rec_error_add(ebuf,"Added missing filter_deinit",-1,tc);
    }
  }
  return ebuf;
}


static gchar *add_null_filter_map(weed_plant_t *event_list, weed_plant_t *last_fm, weed_timecode_t tc, gchar *ebuf) {
  int num_events;
  int error;

  if (!weed_plant_has_leaf(last_fm,"init_events")) return ebuf;

  num_events=weed_leaf_num_elements(last_fm,"init_events");
  if (num_events==1&&weed_get_voidptr_value(last_fm,"init_events",&error)==NULL) return ebuf;

  append_filter_map_event(event_list,tc,NULL);

  ebuf=rec_error_add(ebuf,"Added missing empty filter_map",-1,tc);
  return ebuf;
}

static weed_plant_t *duplicate_frame_at(weed_plant_t *event_list, weed_plant_t *src_frame, weed_timecode_t tc) {
  // tc should be > src_frame tc : i.e. copy is forward in time because insert_frame_event_at searches forward
  int error;
  int *clips,*frames;
  int numframes=weed_leaf_num_elements(src_frame,"clips");

  if (!numframes) return src_frame;

  clips=weed_get_int_array(src_frame,"clips",&error);
  frames=weed_get_int_array(src_frame,"frames",&error);

  insert_frame_event_at (event_list, tc, numframes, clips, frames, &src_frame);

  weed_free(clips);
  weed_free(frames);
  return get_frame_event_at(event_list,tc,src_frame,TRUE);
}




gboolean event_list_rectify(weed_plant_t *event_list, gboolean check_clips) {
  // check and reassemble a newly loaded event_list
  // reassemply consists of matching init_event(s) to event_id's
  // we also rebuild our param_change chains ("in_parameters" in filter_init and filter_deinit, and "next_change" and "prev_change" 
  // in other param_change events)

  // The checking done is quite sophisticated, and can correct many errors in badly-formed event_lists

  weed_plant_t *event=get_first_event(event_list),*event_next;
  weed_timecode_t tc=0,last_tc=0;
  gchar *ebuf=g_strdup("");
  int hint;
  int i,error,idx,filter_idx;
  int host_tag;
  gchar *host_tag_s;
  weed_plant_t *event_id;
  void *init_event;
  void **init_events,**new_init_events;
  weed_plant_t *last_filter_map=NULL;
  gboolean check_filter_map=FALSE;
  gboolean was_deleted=FALSE;
  weed_timecode_t last_frame_tc=-1;
  gchar *filter_hash;
  weed_plant_t *filter=NULL;
  int num_ctmpls,num_inct,num_outct;
  weed_plant_t **ctmpls;
  int *inct,*outct,thisct;
  int pnum;
  weed_timecode_t last_deinit_tc=-1;
  weed_timecode_t last_filter_map_tc=-1;
  int num_init_events;
  weed_plant_t **ptmpls;
  gdouble fps=weed_get_double_value(event_list,"fps",&error);
  weed_timecode_t cur_tc=0;
  weed_plant_t *shortcut=NULL;
  int num_params;
  void **in_pchanges,**orig_pchanges;
  int pflags;
  int num_tracks;
  int *clip_index,*frame_index;
  int *new_clip_index,*new_frame_index;
  int last_valid_frame;
  weed_plant_t *last_event;
  void **pchains[FX_KEYS_MAX-FX_KEYS_MAX_VIRTUAL]; // parameter chains
  gboolean was_moved;
  weed_plant_t *last_frame_event;
  int marker_type;

  ttable trans_table[FX_KEYS_MAX-FX_KEYS_MAX_VIRTUAL]; // translation table for init_events

  for (i=0;i<FX_KEYS_MAX-FX_KEYS_MAX_VIRTUAL;i++) trans_table[i].in=trans_table[i].out=NULL;

  free_tt_key=0;

  while (event!=NULL) {
    was_deleted=FALSE;
    event_next=get_next_event(event);
    if (!weed_plant_has_leaf(event,"timecode")) {
      ebuf=rec_error_add(ebuf,"Event has no timecode",weed_get_plant_type(event),-1);
      delete_event(event_list,event);
      event=event_next;
      continue;
    }
    tc=get_event_timecode(event);
    tc=q_gint64(tc+U_SEC/(2.*fps)-1,fps);
    weed_set_int64_value(event,"timecode",tc);
    if (!weed_get_plant_type(event)==WEED_PLANT_EVENT) {
      ebuf=rec_error_add(ebuf,"Invalid plant type",weed_get_plant_type(event),tc);
      delete_event(event_list,event);
      event=event_next;
      continue;
    }
    if (tc<last_tc) {
      ebuf=rec_error_add(ebuf,"Out of order event",-1,tc);
      delete_event(event_list,event);
      event=event_next;
      continue;
    }
    if (!weed_plant_has_leaf(event,"hint")) {
      ebuf=rec_error_add(ebuf,"Event has no hint",weed_get_plant_type(event),tc);
      delete_event(event_list,event);
      event=event_next;
      continue;
    }

    hint=get_event_hint(event);
    switch (hint) {
    case WEED_EVENT_HINT_FILTER_INIT:
      // set in table
      if (!weed_plant_has_leaf(event,"event_id")) {
	ebuf=rec_error_add(ebuf,"Filter_init missing event_id",-1,tc);
	delete_event(event_list,event);
	was_deleted=TRUE;
      }
      else {
	if (!weed_plant_has_leaf(event,"filter")) {
	  ebuf=rec_error_add(ebuf,"Filter_init missing filter",-1,tc);
	  delete_event(event_list,event);
	  was_deleted=TRUE;
	}
	else {
	  filter_hash=weed_get_string_value(event,"filter",&error);
	  if ((filter_idx=weed_get_idx_for_hashname(filter_hash))!=-1) {
	    filter=get_weed_filter(filter_idx);
	    if (weed_plant_has_leaf(filter,"in_channel_templates")) {
	      if (!weed_plant_has_leaf(event,"in_count")) {
		ebuf=rec_error_add(ebuf,"Filter_init missing filter",-1,tc);
		delete_event(event_list,event);
		was_deleted=TRUE;
	      }
	      else {
		num_ctmpls=weed_leaf_num_elements(filter,"in_channel_templates");
		num_inct=weed_leaf_num_elements(event,"in_count");
		if (num_ctmpls!=num_inct) {
		  ebuf=rec_error_add(ebuf,"Filter_init has invalid in_count",-1,tc);
		  delete_event(event_list,event);
		  was_deleted=TRUE;
		}
		else {
		  inct=weed_get_int_array(event,"in_count",&error);
		  ctmpls=weed_get_plantptr_array(filter,"in_channel_templates",&error);
		  for (i=0;i<num_ctmpls;i++) {
		    thisct=inct[i];
		    if (thisct==0&&!weed_plant_has_leaf(ctmpls[i],"optional")) {
		      ebuf=rec_error_add(ebuf,"Filter_init disables a non-optional in channel",i,tc);
		      delete_event(event_list,event);
		      was_deleted=TRUE;
		    }
		    else {
		      if (thisct>1&&(!weed_plant_has_leaf(ctmpls[i],"max_repeats")||(weed_get_int_value(ctmpls[i],"max_repeats",&error)>0&&weed_get_int_value(ctmpls[i],"max_repeats",&error)<thisct))) {
			ebuf=rec_error_add(ebuf,"Filter_init has too many repeats of in channel",i,tc);
			delete_event(event_list,event);
			was_deleted=TRUE;
		      }
		    }
		  }
		  weed_free(inct);
		  weed_free(ctmpls);
		  if (!was_deleted) {
		    num_ctmpls=weed_leaf_num_elements(filter,"out_channel_templates");
		    num_outct=weed_leaf_num_elements(event,"out_count");
		    if (num_ctmpls!=num_outct) {
		      ebuf=rec_error_add(ebuf,"Filter_init has invalid out_count",-1,tc);
		      delete_event(event_list,event);
		      was_deleted=TRUE;
		    }
		    else {
		      outct=weed_get_int_array(event,"out_count",&error);
		      ctmpls=weed_get_plantptr_array(filter,"out_channel_templates",&error);
		      for (i=0;i<num_ctmpls;i++) {
			thisct=outct[i];
			if (thisct==0&&!weed_plant_has_leaf(ctmpls[i],"optional")) {
			  ebuf=rec_error_add(ebuf,"Filter_init disables a non-optional out channel",i,tc);
			  delete_event(event_list,event);
			  was_deleted=TRUE;
			}
			else {
			  if (thisct>1&&(!weed_plant_has_leaf(ctmpls[i],"max_repeats")||(weed_get_int_value(ctmpls[i],"max_repeats",&error)>0&&weed_get_int_value(ctmpls[i],"max_repeats",&error)<thisct))) {
			    ebuf=rec_error_add(ebuf,"Filter_init has too many repeats of out channel",i,tc);
			    delete_event(event_list,event);
			    was_deleted=TRUE;
			  }
			}
		      }
		      weed_free(outct);
		      weed_free(ctmpls);
		    }}}}}}
	  weed_free(filter_hash);
	  if (!was_deleted) {
	    host_tag=get_next_tt_key(trans_table)+FX_KEYS_MAX_VIRTUAL+1;
	    if (host_tag==-1) {
	      ebuf=rec_error_add(ebuf,"Fatal: too many active effects",FX_KEYS_MAX-FX_KEYS_MAX_VIRTUAL,tc);
	      return FALSE;
	    }
	    host_tag_s=g_strdup_printf("%d",host_tag);
	    weed_set_string_value(event,"host_tag",host_tag_s);
	    g_free(host_tag_s);
	    event_id=weed_get_voidptr_value(event,"event_id",&error);
	    trans_table[(idx=host_tag-FX_KEYS_MAX_VIRTUAL-1)].in=event_id;
	    trans_table[idx].out=event;
	    
	    // use pchain array
	    if ((num_params=num_in_params(filter,TRUE,TRUE))>0) {
	      pchains[idx]=g_malloc(num_params*sizeof(void *));
	      in_pchanges=g_malloc(num_params*sizeof(void *));
	      for (i=0;i<num_params;i++) {
		pchains[idx][i]=event;
		in_pchanges[i]=NULL;
	      }
	      weed_set_voidptr_array(event,"in_parameters",num_params,in_pchanges); // set all to NULL, we will re-fill as we go along
	      g_free(in_pchanges);
	    }
	  }
	}
      }
      break;
    case WEED_EVENT_HINT_FILTER_DEINIT:

      // update "init_event" from table, remove entry; we will check filter_map at end of tc
      if (!weed_plant_has_leaf(event,"init_event")) {
	ebuf=rec_error_add(ebuf,"Filter_deinit missing init_event",-1,tc);
	delete_event(event_list,event);
	was_deleted=TRUE;
      }
      else {
	event_id=weed_get_voidptr_value(event,"init_event",&error);
	init_event=find_init_event_in_ttable(trans_table,event_id,TRUE);
	if (init_event==NULL) {
	  ebuf=rec_error_add(ebuf,"Filter_deinit has invalid init_event",-1,tc);
	  delete_event(event_list,event);
	  was_deleted=TRUE;
	}
	else {
	  weed_set_voidptr_value(init_event,"deinit_event",event);

	  host_tag_s=weed_get_string_value(init_event,"host_tag",&error);
	  host_tag=atoi(host_tag_s);
	  weed_free(host_tag_s);
	  trans_table[(idx=host_tag-FX_KEYS_MAX_VIRTUAL-1)].in=NULL;
	  if (idx<free_tt_key) free_tt_key=idx;
	  weed_set_voidptr_value(event,"init_event",init_event);
	  check_filter_map=TRUE;
	  last_deinit_tc=tc;

	  filter_hash=weed_get_string_value(init_event,"filter",&error);
	  if ((filter_idx=weed_get_idx_for_hashname(filter_hash))!=-1) {
	    filter=get_weed_filter(filter_idx);
	    if ((num_params=num_in_params(filter,TRUE,TRUE))>0) {
	      in_pchanges=g_malloc(num_params*sizeof(void *));
	      for (i=0;i<num_params;i++) {
		if (!WEED_EVENT_IS_FILTER_INIT(pchains[idx][i])) in_pchanges[i]=pchains[idx][i];
		else in_pchanges[i]=NULL;
	      }
	      weed_set_voidptr_array(event,"in_parameters",num_params,in_pchanges); // set array to last param_changes
	      g_free(in_pchanges);
	      g_free(pchains[idx]);
	    }
	  }
	}
      }
      break;
    case WEED_EVENT_HINT_FILTER_MAP:
	// update "init_events" from table
      if (weed_plant_has_leaf(event,"init_events")) {
	num_init_events=weed_leaf_num_elements(event,"init_events");
	init_events=weed_get_voidptr_array(event,"init_events",&error);
	new_init_events=(void **)g_malloc(num_init_events*sizeof(void *));
	for (i=0;i<num_init_events;i++) {
	  event_id=init_events[i];
	  if (event_id!=NULL) {
	    init_event=find_init_event_in_ttable(trans_table,event_id,TRUE);
	    if (init_event==NULL) {
	      ebuf=rec_error_add(ebuf,"Filter_map has invalid init_event",-1,tc);
	      new_init_events[i]=NULL;
	    }
	    else new_init_events[i]=init_event;
	  }
	  else new_init_events[i]=NULL;
	}
	if (num_init_events>1) new_init_events=remove_nulls_from_filter_map(new_init_events,num_init_events);
	weed_set_voidptr_array(event,"init_events",num_init_events,new_init_events);
	weed_free(init_events);
	g_free(new_init_events);
      }
      if (last_filter_map!=NULL) {
	if (compare_filter_maps(last_filter_map,event)) {
	  // filter map is identical to prior one, we can remove this one
	  delete_event(event_list,event);
	  was_deleted=TRUE;
	}
      }
      if (!was_deleted) last_filter_map=event;

      break;
    case WEED_EVENT_HINT_PARAM_CHANGE:
      if (!weed_plant_has_leaf(event,"index")) {
	ebuf=rec_error_add(ebuf,"Param_change has no index",-1,tc);
	delete_event(event_list,event);
	was_deleted=TRUE;
      }
      else {
	if (!weed_plant_has_leaf(event,"value")) {
	  ebuf=rec_error_add(ebuf,"Param_change has no value",-1,tc);
	  delete_event(event_list,event);
	  was_deleted=TRUE;
	}
	else {
	  if (!weed_plant_has_leaf(event,"init_event")) {
	    ebuf=rec_error_add(ebuf,"Param_change has no init_event",-1,tc);
	    delete_event(event_list,event);
	    was_deleted=TRUE;
	  }
	  else {
	    event_id=weed_get_voidptr_value(event,"init_event",&error);
	    if ((init_event=find_init_event_in_ttable(trans_table,event_id,TRUE))==NULL) {
	      ebuf=rec_error_add(ebuf,"Param_change has invalid init_event",-1,tc);
	      delete_event(event_list,event);
	      was_deleted=TRUE;
	    }
	    else {
	      filter_hash=weed_get_string_value(init_event,"filter",&error);
	      if ((filter_idx=weed_get_idx_for_hashname(filter_hash))!=-1) {
		filter=get_weed_filter(filter_idx);
		pnum=weed_get_int_value(event,"index",&error);
		if (pnum<0||pnum>=(num_params=num_in_params(filter,TRUE,TRUE))) {
		  ebuf=rec_error_add(ebuf,"Param_change has invalid index",pnum,tc);
		  delete_event(event_list,event);
		  was_deleted=TRUE;
		}
		else {
		  ptmpls=weed_get_plantptr_array(filter,"in_parameter_templates",&error);
		  if (!weed_plant_has_leaf(event,"value")) {
		    ebuf=rec_error_add(ebuf,"Param_change has no value with index",pnum,tc);
		    delete_event(event_list,event);
		    was_deleted=TRUE;
		  }
		  else {
		    if (weed_leaf_seed_type(event,"value")!=weed_leaf_seed_type(ptmpls[pnum],"default")) {
		      ebuf=rec_error_add(ebuf,"Param_change has invalid seed type with index",pnum,tc);
		      delete_event(event_list,event);
		      was_deleted=TRUE;
		    }
		    else {
		      pflags=weed_get_int_value(ptmpls[pnum],"flags",&error);
		      if (pflags&WEED_PARAMETER_REINIT_ON_VALUE_CHANGE&&!is_init_pchange(init_event,event)) {
			// check we are not changing a reinit param, unless we immediately follow the filter_init event
			ebuf=rec_error_add(ebuf,"Param_change sets a reinit parameter",pnum,tc);
			delete_event(event_list,event);
			was_deleted=TRUE;
		      }
		      else {
			// all checks passed
			host_tag_s=weed_get_string_value(init_event,"host_tag",&error);
			host_tag=atoi(host_tag_s);
			weed_free(host_tag_s);
			idx=host_tag-FX_KEYS_MAX_VIRTUAL-1;
			if (pchains[idx][pnum]==init_event) {
			  orig_pchanges=weed_get_voidptr_array(init_event,"in_parameters",&error);
			  in_pchanges=g_malloc(num_params*sizeof(void *));
			  for (i=0;i<num_params;i++) {
			    if (orig_pchanges[i]==NULL&&i==pnum) in_pchanges[i]=(void *)event;
			    else in_pchanges[i]=orig_pchanges[i];
			  }
			  weed_set_voidptr_array(init_event,"in_parameters",num_params,in_pchanges);
			  g_free(in_pchanges);
			  g_free(orig_pchanges);
			  weed_set_voidptr_value(event,"prev_change",NULL);
			}
			else {
			  weed_set_voidptr_value(pchains[idx][pnum],"next_change",event);
			  weed_set_voidptr_value(event,"prev_change",pchains[idx][pnum]);
			}
			weed_set_voidptr_value(event,"next_change",NULL);
			weed_set_voidptr_value(event,"init_event",init_event);
			pchains[idx][pnum]=event;
		      }}}
		  weed_free(ptmpls);
		}
		weed_free(filter_hash);
	      }}}}}
      break;
    case WEED_EVENT_HINT_FRAME:
      if (tc==last_frame_tc) {
	ebuf=rec_error_add(ebuf,"Duplicate frame event",-1,tc);
	delete_event(event_list,event);
	was_deleted=TRUE;
      }
      else {
	last_frame_tc=tc;

	num_tracks=weed_leaf_num_elements (event,"clips");
	clip_index=weed_get_int_array(event,"clips",&error);
	frame_index=weed_get_int_array(event,"frames",&error);

	new_clip_index=g_malloc(num_tracks*sizint);
	new_frame_index=g_malloc(num_tracks*sizint);
	last_valid_frame=0;

	for (i=0;i<num_tracks;i++) {
	  if (clip_index[i]!=-1&&renumbered_clips[clip_index[i]]==0) {
	    // clip has probably been closed, so we remove its frames
	    new_clip_index[i]=-1;
	    new_frame_index[i]=0;
	    ebuf=rec_error_add(ebuf,"Invalid clip number",clip_index[i],tc);
	  }
	  else {
	    // take into account the fact that clip could have been resampled since layout was saved
	    if (clip_index[i]>0&&frame_index[i]>0) {
	      new_frame_index[i]=count_resampled_frames(frame_index[i],lfps[clip_index[i]],mainw->files[clip_index[i]]->fps);
	      if (new_frame_index[i]>mainw->files[renumbered_clips[clip_index[i]]]->frames) {
		ebuf=rec_error_add(ebuf,"Invalid frame number",new_frame_index[i],tc);
		new_clip_index[i]=-1;
		new_frame_index[i]=0;
	      }
	      else last_valid_frame=i+1;
	    }
	    else last_valid_frame=i+1;
	  }
	}

	if (last_valid_frame==0) {
	  g_free(new_clip_index);
	  g_free(new_frame_index);
	  new_clip_index=g_malloc(sizint);
	  new_frame_index=g_malloc(sizint);
	  *new_clip_index=-1;
	  *new_frame_index=0;
	  num_tracks=1;
	}
	else {
	  if (last_valid_frame<num_tracks) {
	    weed_free(clip_index);
	    weed_free(frame_index);
	    clip_index=weed_malloc(last_valid_frame*sizint);
	    frame_index=weed_malloc(last_valid_frame*sizint);
	    for (i=0;i<last_valid_frame;i++) {
	      clip_index[i]=new_clip_index[i];
	      frame_index[i]=new_frame_index[i];
	    }
	    num_tracks=last_valid_frame;
	    weed_set_int_array(event,"clips",num_tracks,clip_index);
	    weed_set_int_array(event,"frames",num_tracks,frame_index);
	  }
	  else {
	    g_free(new_clip_index);
	    g_free(new_frame_index);
	    new_clip_index=clip_index;
	    new_frame_index=frame_index;
	  }
	}

	weed_set_int_array(event,"clips",num_tracks,new_clip_index);
	weed_set_int_array(event,"frames",num_tracks,new_frame_index);

	if (new_clip_index!=clip_index) g_free(new_clip_index);
	weed_free(clip_index);
	if (new_frame_index!=frame_index) g_free(new_frame_index);
	weed_free(frame_index);

      }
      break;
    case WEED_EVENT_HINT_MARKER:
      // check marker values
      if (!weed_plant_has_leaf(event,"lives_type")) {
	ebuf=rec_error_add(ebuf,"Unknown marker type",-1,tc);
	delete_event(event_list,event);
	was_deleted=TRUE;
      }
      else {
	marker_type=weed_get_int_value(event,"lives_type",&error);
	if (marker_type!=EVENT_MARKER_BLOCK_START&&marker_type!=EVENT_MARKER_BLOCK_UNORDERED) {
	  ebuf=rec_error_add(ebuf,"Unknown marker type",marker_type,tc);
	  delete_event(event_list,event);
	  was_deleted=TRUE;
	}
	if (!weed_plant_has_leaf(event,"tracks")) {
	  ebuf=rec_error_add(ebuf,"Block start marker has no tracks",-1,tc);
	  delete_event(event_list,event);
	  was_deleted=TRUE;
	}
      }
      break;
    default:
      ebuf=rec_error_add(ebuf,"Invalid hint",hint,tc);
      delete_event(event_list,event);
      was_deleted=TRUE;
    }
    if (!was_deleted&&check_filter_map&&last_filter_map!=NULL&&\
	(event_next==NULL||get_event_timecode(event_next)>last_deinit_tc)) {
      // if our last filter_map refers to filter instances which were deinited, we must add another filter_map here
      ebuf=filter_map_check(trans_table,last_filter_map,last_deinit_tc,tc,ebuf);
      check_filter_map=FALSE;
    }
    event=event_next;

    if (!was_deleted) {
      while (cur_tc<last_frame_tc) {
	// add blank frames
	if (!has_frame_event_at(event_list,cur_tc,&shortcut)) {
	  if (shortcut!=NULL) {
	    shortcut=duplicate_frame_at(event_list,shortcut,cur_tc);
	    ebuf=rec_error_add(ebuf,"Duplicated frame at",-1,cur_tc);
	  }
	  else {
	    insert_blank_frame_event_at(event_list,cur_tc,&shortcut);
	    ebuf=rec_error_add(ebuf,"Inserted missing blank frame",-1,cur_tc);
	  }
	}
	cur_tc+=U_SEC/fps;
	cur_tc=q_gint64(cur_tc,fps);
      }
      last_tc=tc;
    }
  }

  last_event=get_last_event(event_list);
  remove_end_blank_frames(event_list);

  if (get_last_event(event_list)!=last_event) {
    last_event=get_last_event(event_list);
    last_tc=get_event_timecode(last_event);
    ebuf=rec_error_add(ebuf,"Removed final blank frames",-1,last_tc);
  }

  // add any missing filter_deinit events
  ebuf=add_filter_deinits(event_list,trans_table,pchains,last_tc,ebuf);

  // check the last filter map
  if (last_filter_map!=NULL) ebuf=add_null_filter_map(event_list,last_filter_map,last_tc,ebuf);

  // pass 2 - move left any FILTER_DEINITS before the FRAME, move right any FILTER_INITS or PARAM_CHANGES after the FRAME
  // ensure we have at most 1 FILTER_MAP before each FRAME, and 1 FILTER_MAP after a FRAME

  // we do this as a second pass since we may have inserted blank frames
  last_frame_tc=last_filter_map_tc=-1;
  last_frame_event=NULL;

  event=get_first_event(event_list);
  while (event!=NULL) {
    was_moved=FALSE;
    event_next=get_next_event(event);
    tc=get_event_timecode(event);
    hint=get_event_hint(event);
    switch (hint) {
    case WEED_EVENT_HINT_FILTER_INIT:
      if (!move_event_right(event_list,event,last_frame_tc!=tc,fps)) was_moved=TRUE;
      break;
    case WEED_EVENT_HINT_PARAM_CHANGE:
      if (last_frame_tc==tc) if (!move_event_right(event_list,event,FALSE,fps)) was_moved=TRUE;
      break;
    case WEED_EVENT_HINT_FILTER_DEINIT:
      if (!move_event_left(event_list,event,last_frame_tc==tc,fps)) was_moved=TRUE;
      break;
    case WEED_EVENT_HINT_FILTER_MAP:
      if (last_filter_map_tc==tc) {
	// remove last filter_map
	ebuf=rec_error_add(ebuf,"Duplicate filter maps",-1,tc);
	delete_event(event_list,last_filter_map);
      }
      last_filter_map_tc=tc;
      last_filter_map=event;
      break;
    case WEED_EVENT_HINT_FRAME:
      last_frame_tc=tc;
      last_filter_map_tc=-1;
      last_frame_event=event;
      break;
      }
    if (was_moved) {
      if (last_frame_event!=NULL) event=last_frame_event;
      else event=get_first_event(event_list);
    }
    else event=event_next;
  }
  return TRUE;
}


static weed_plant_t *load_event_list(lives_mt *mt) {
  // here we assume mt->event_list is NULL and create it
  int fd;
  int error;

  gchar *filt[]={"*.lay",NULL};
  gchar *startdir=NULL;
  gchar *eload_dir=g_strdup_printf("%s/%s/layouts/",prefs->tmpdir,mainw->set_name);
  gchar *eload_file;
  gchar *msg;
  gchar *com;

  weed_plant_t *event_list=NULL;

  int num_events=0;

  com=g_strdup_printf ("%s %s",capable->mkdirhier,eload_dir);
  system (com);
  g_free (com);

  if (!g_file_test(eload_dir,G_FILE_TEST_IS_DIR)) return NULL;
  
  if (mainw->was_set) startdir=g_strdup_printf("%s%s/",eload_dir,mainw->set_name);
  if (!g_file_test(startdir,G_FILE_TEST_IS_DIR)||!mainw->was_set) startdir=g_strdup(eload_dir);

  eload_file=choose_file(startdir,NULL,filt,GTK_FILE_CHOOSER_ACTION_OPEN);
  g_free(startdir);

  if ((fd=open(eload_file,O_RDONLY))<0) return NULL;

  g_snprintf(mt->layout_name,256,"%s",eload_file);
  get_basename(mt->layout_name);

  msg=g_strdup_printf(_("Loading layout from %s..."),eload_file);
  d_print(msg);
  g_free(msg);

  mt_desensitise(mt);

  if ((event_list=load_event_list_inner(fd,TRUE,&num_events,NULL,NULL))==NULL) {
    close(fd);
    mt_sensitise(mt);
    return NULL;
  }

  close(fd);

  d_print(_("done.\n"));

  msg=g_strdup_printf(_("Got %d events...processing..."),num_events);
  d_print(msg);
  g_free(msg);
  g_free(eload_dir);

  mt->changed=FALSE;

  // event list loaded, now we set the pointers for filter_map (init_events), param_change (init_events and param chains), filter_deinit (init_events)
  do_threaded_dialog(_("Checking and rebuilding event list"));

  elist_errors=0;

  rerenumber_clips(eload_file); // re-map clips so our loaded event_list refers to the correct clips and frames
  g_free(eload_file);

  if (!event_list_rectify(event_list,FALSE)) {
    event_list_free(event_list);
    event_list=NULL;
  }

  if (get_first_event(event_list)==NULL) {
    event_list_free(event_list);
    event_list=NULL;
  }

  end_threaded_dialog();
  if (event_list!=NULL) {
    msg=g_strdup_printf(_("%d errors detected.\n"),elist_errors);
    d_print(msg);
    g_free(msg);

    // resave with corrections/updates
    fd=creat(eload_file,S_IRUSR|S_IWUSR);
    save_event_list_inner(fd,event_list,NULL);
    close(fd);
  }
  else d_print(_("failed.\n"));
  mt_sensitise(mt);

  if (event_list!=NULL) {
    msg=g_strdup_printf(_("Multitrack fps set to %.3f\n"),weed_get_double_value(event_list,"fps",&error));
    d_print(msg);
    g_free(msg);
    gtk_window_set_title (GTK_WINDOW (mt->window), g_strdup_printf(_("LiVES-%s: Multitrack %dx%d : %d bpp %.3f fps"),LiVES_VERSION,cfile->hsize,cfile->vsize,cfile->bpp,cfile->fps));
  }

  return (event_list);
}


void remove_markers(weed_plant_t *event_list) {
  weed_plant_t *event=get_first_event(event_list);
  weed_plant_t *event_next;
  int marker_type;
  int error;

  while (event!=NULL) {
    event_next=get_next_event(event);
    if (WEED_EVENT_IS_MARKER(event)) {
      marker_type=weed_get_int_value(event,"lives_type",&error);
      if (marker_type==EVENT_MARKER_BLOCK_START||marker_type==EVENT_MARKER_BLOCK_UNORDERED) {
	delete_event(event_list,event);
      }
    }
    event=event_next;
  }
}




void on_load_event_list_activate (GtkMenuItem *menuitem, gpointer user_data) {
  int i;
  lives_mt *mt=(lives_mt *)user_data;
  gdouble end_secs;
  GList *slist,*slist_next;

  if (mt->event_list!=NULL) event_list_free(mt->event_list);

  for (i=0;i<mt->num_video_tracks;i++) {
    delete_video_track(mt,i,FALSE);
  }
  g_list_free(mt->video_draws);
  mt->video_draws=NULL;
  mt->num_video_tracks=0;

  end_secs=mt->end_secs;
  mt->event_list=NULL;

  mt->event_list=load_event_list(mt);

  slist=mt->selected_tracks;
  while (slist!=NULL) {
    slist_next=slist->next;
    g_free(slist->data);
    slist=slist_next;
  }
  if (slist!=NULL) g_free(slist);
  mt->selected_tracks=NULL;

  gtk_widget_destroy(mt->timeline_eb);
  init_tracks(mt);
  remove_markers(mt->event_list);
  gtk_widget_show_all(mt->timeline_eb);
  mt_sensitise(mt);
  mt->end_secs=end_secs;
  set_timeline_end_secs(mt,end_secs);
}



void migrate_layouts (gchar *old_set_name, gchar *new_set_name) {
  // load each event_list in mainw->layout_map_list
  GList *map=mainw->current_layouts_map;
  int fd;
  weed_plant_t *event_list;
  gchar *tmp;
  int i;

  gchar *changefrom=g_strdup_printf("%s/%s/layouts/",prefs->tmpdir,old_set_name);
  size_t chlen=strlen(changefrom);

  while (map!=NULL) {
    if ((fd=open(map->data,O_RDONLY))>-1) {
      if ((event_list=load_event_list_inner(fd,FALSE,NULL,NULL,NULL))!=NULL) {
	close (fd);
	// adjust the value of "needs_set" to new_set_name
	weed_set_string_value(event_list,"needs_set",new_set_name);
	// save the event_list with the same name
	unlink(map->data);
	fd=creat(map->data,S_IRUSR|S_IWUSR);
	save_event_list_inner(fd,event_list,NULL);
	event_list_free(event_list);
      }
      close(fd);
    }
    if (!strncmp(map->data,changefrom,chlen)) {
      // update mainw->current_layouts_map
      tmp=g_strdup_printf("%s/%s/layouts/%s",prefs->tmpdir,new_set_name,(char *)map->data+chlen);
      if (g_file_test(tmp,G_FILE_TEST_EXISTS)) {
	gchar *com;
	// prevent duplication of layouts
	g_free(tmp);
	tmp=g_strdup_printf("%s/%s/layouts/%s-%s",prefs->tmpdir,new_set_name,old_set_name,(char *)map->data+chlen);
	com=g_strdup_printf("/bin/mv %s %s",(gchar *)map->data,tmp);
	system(com);
	g_free(com);
      }
      g_free(map->data);
      map->data=tmp;
    }
    map=map->next;
  }

  // update layout_map's in mainw->files
  for (i=1;i<=MAX_FILES;i++) {
    if (mainw->files[i]!=NULL) {
      if (mainw->files[i]->layout_map!=NULL) {
	map=mainw->files[i]->layout_map;
	while (map!=NULL) {
	  if (map->data!=NULL) {
	    if (!strncmp(map->data,changefrom,chlen)) {
	      tmp=g_strdup_printf("%s/%s/layouts/%s",prefs->tmpdir,new_set_name,(char *)map->data+chlen);
	      g_free(map->data);
	      map->data=tmp;
	    }
	    map=map->next;
	  }
	}
      }
    }
  }

  // update mainw->affected_layouts_map
  map=mainw->affected_layouts_map; 
  while (map!=NULL) {
    if (!strncmp(map->data,changefrom,chlen)) {
      tmp=g_strdup_printf("%s/%s/layouts/%s",prefs->tmpdir,new_set_name,(char *)map->data+chlen);
      g_free(map->data);
      map->data=tmp;
    }
    map=map->next;
  }
}



gboolean layout_frame_is_affected(gint clipno, gint frame) {
  gchar **array;
  GList *lmap=mainw->files[clipno]->layout_map;
  gdouble orig_fps;
  gint resampled_frame;

  while (lmap!=NULL) {
    array=g_strsplit(lmap->data,"|",-1);
    orig_fps=strtod(array[3],NULL);
    resampled_frame=count_resampled_frames(frame,orig_fps,mainw->files[clipno]->fps);
    if (resampled_frame<=atoi(array[1])) {
      g_strfreev(array);
      return TRUE;
    }
    lmap=lmap->next;
  }
  return FALSE;
}





///////////////////////////////////////////////

void show_frame_events_activate (GtkMenuItem *menuitem, gpointer user_data) {
  prefs->event_window_show_frame_events=!prefs->event_window_show_frame_events;
}


guint event_list_get_byte_size(weed_plant_t *event_list,int *num_events) {
  // return serialisation size
  int i,j;
  guint tot=0;
  weed_plant_t *event=get_first_event(event_list);
  gchar **leaves;
  int ne;
  int tot_events=0;

  while (event!=NULL) {
    if (WEED_EVENT_IS_FILTER_INIT(event)) weed_set_voidptr_value(event,"event_id",(void *)event);
    tot_events++;
    leaves=weed_plant_list_leaves(event);
    tot+=sizint;
    for (i=0;leaves[i]!=NULL;i++) {
      tot+=sizint*3+strlen(leaves[i]);
      ne=weed_leaf_num_elements(event,leaves[i]);
      for (j=0;j<ne;j++) tot+=sizint+weed_leaf_element_size(event,leaves[i],j);
      weed_free(leaves[i]);
    }
    weed_free(leaves);
    event=get_next_event(event);
  }

  event=event_list;
  leaves=weed_plant_list_leaves(event);
  tot+=sizint;
  for (i=0;leaves[i]!=NULL;i++) {
    tot+=sizint*3+strlen(leaves[i]);
    ne=weed_leaf_num_elements(event,leaves[i]);
    for (j=0;j<ne;j++) tot+=sizint+weed_leaf_element_size(event,leaves[i],j);
    weed_free(leaves[i]);
  }
  weed_free(leaves);

  if (num_events!=NULL) *num_events=tot_events;
  return tot;
}
